@@ -39,6 +39,77 @@ static void be_daemon(int close_stderr)
 	setsid();
 }
 
+
+static const char *corosync_exit_reason(signed char status)
+{
+	static char reason[256];
+	switch (status) {
+	case -2:
+		return "Could not determine UID to run as";
+		break;
+	case -3:
+		return "Could not determine GID to run as";
+		break;
+	case -4:
+		return "Error initialising memory pool";
+		break;
+	case -5:
+		return "Could not fork";
+		break;
+	case -6:
+		return "Could not bind to libais socket";
+		break;
+	case -7:
+		return "Could not bind to network socket";
+		break;
+	case -8:
+		return "Could not read security key for communications";
+		break;
+	case -9:
+		return "Could not read cluster configuration";
+		break;
+	case -10:
+		return "Could not set up logging";
+		break;
+	case -11:
+		return "Could not dynamically load modules";
+		break;
+	case -12:
+		return "Could not load and initialise object database";
+		break;
+	case -13:
+		return "Could not initialise all required services";
+		break;
+	case -14:
+		return "Out of memory";
+		break;
+	default:
+		sprintf(reason, "Error, reason code is %d", status);
+		return reason;
+		break;
+	}
+}
+
+static int check_corosync_status(pid_t pid)
+{
+	int status;
+	int pidstatus;
+
+	status = waitpid(pid, &pidstatus, WNOHANG);
+	if (status == -1 && errno == ECHILD) {
+
+		return 0;
+	}
+	if ((status == 0 || status == pid) && pidstatus != 0) {
+		if (WIFEXITED(pidstatus))
+			fprintf(stderr, "corosync died: %s\n", corosync_exit_reason(WEXITSTATUS(pidstatus)));
+		if (WIFSIGNALED(pidstatus))
+			fprintf(stderr, "corosync died with signal: %d\n", WTERMSIG(pidstatus));
+		exit(1);
+	}
+	return status;
+}
+
 int join(commandline_t *comline, char *main_envp[])
 {
 	int i, err;
@@ -95,18 +166,18 @@ int join(commandline_t *comline, char *main_envp[])
 		envp[envptr++] = strdup(scratch);
 	}
 	if (comline->verbose ^ DEBUG_STARTUP_ONLY) {
-		snprintf(scratch, sizeof(scratch), "CMAN_DEBUGLOG=%d", comline->verbose);
+		snprintf(scratch, sizeof(scratch), "CMAN_DEBUG=%d", comline->verbose);
 		envp[envptr++] = strdup(scratch);
 	}
 	if (comline->noconfig_opt) {
 		envp[envptr++] = strdup("CMAN_NOCONFIG=true");
 		snprintf(scratch, sizeof(scratch), "COROSYNC_DEFAULT_CONFIG_IFACE=cmanpreconfig%s",
-			 comline->noopenais_opt?"":":openaisserviceenable");
+			 comline->noopenais_opt?"":":openaisserviceenablestable");
 		envp[envptr++] = strdup(scratch);
 	}
 	else {
 		snprintf(scratch, sizeof(scratch), "COROSYNC_DEFAULT_CONFIG_IFACE=%s:cmanpreconfig%s", comline->config_lcrso,
-			 comline->noopenais_opt?"":":openaisserviceenable");
+			 comline->noopenais_opt?"":":openaisserviceenablestable");
 		envp[envptr++] = strdup(scratch);
 	}
 
@@ -127,11 +198,14 @@ int join(commandline_t *comline, char *main_envp[])
 	envp[envptr++] = strdup(scratch);
 	envp[envptr++] = NULL;
 
-	argv[0] = "corosync";
-	if (comline->verbose & ~DEBUG_STARTUP_ONLY)
-		argv[++argvptr] = "-f";
+	/* Always run corosync -f because we have already forked twice anyway, and
+	   we want to return any exit code that might happen */
+	/* also strdup strings because it's otherwise virtually impossible to fix
+	 * build warnings due to the way argv C implementation is done */
+	argv[0] = strdup("corosync");
+	argv[++argvptr] = strdup("-f");
 	if (comline->nosetpri_opt)
-		argv[++argvptr] = "-p";
+		argv[++argvptr] = strdup("-p");
 	argv[++argvptr] = NULL;
 
 	/* Fork/exec cman */
@@ -153,6 +227,10 @@ int join(commandline_t *comline, char *main_envp[])
 			}
 		}
 		be_daemon(!(comline->verbose & ~DEBUG_STARTUP_ONLY));
+
+		sprintf(scratch, "FORKED: %d\n", getpid());
+		err = write(p[1], scratch, strlen(scratch));
+
 		execve(COROSYNCBIN, argv, envp);
 
 		/* exec failed - tell the parent process */
@@ -176,49 +254,44 @@ int join(commandline_t *comline, char *main_envp[])
 		fd_set fds;
 		struct timeval tv={1, 0};
 		char message[1024];
+		char *messageptr = message;
 
 		FD_ZERO(&fds);
 		FD_SET(p[0], &fds);
 
 		status = select(p[0]+1, &fds, NULL, NULL, &tv);
 
-		/* Did we get an error? */
+		/* Did we get a cman-reported error? */
 		if (status == 1) {
 			int len;
 			if ((len = read(p[0], message, sizeof(message)) > 0)) {
 
+				/* Forked OK - get the real corosync pid */
+				if (sscanf(messageptr, "FORKED: %d", &corosync_pid) == 1) {
+					if (comline->verbose & DEBUG_STARTUP_ONLY)
+						fprintf(stderr, "forked process ID is %d\n", corosync_pid);
+					status = 1;
+
+					/* There might be a SUCCESS message in the pipe too. */
+					messageptr = strchr(messageptr, '\n');
+					if (messageptr)
+						messageptr++;
+					else
+						continue;
+				}
 				/* Success! get the new PID of double-forked corosync */
-				if (sscanf(message, "SUCCESS: %d", &corosync_pid) == 1) {
+				if (sscanf(messageptr, "SUCCESS: %d", &corosync_pid) == 1) {
 					if (comline->verbose & DEBUG_STARTUP_ONLY)
 						fprintf(stderr, "corosync running, process ID is %d\n", corosync_pid);
 					status = 0;
+					break;
 				}
-				else {
-					fprintf(stderr, "cman not started: %s\n", message);
-				}
-				break;
 			}
 			else if (len < 0 && errno == EINTR) {
 				continue;
 			}
 			else { /* Error or EOF - check the child status */
-				int pidstatus;
-				status = waitpid(corosync_pid, &pidstatus, WNOHANG);
-				if (status == -1 && errno == ECHILD) {
-					fprintf(stderr, "cman not started\n");
-					break;
-				}
-				if (status == 0 && pidstatus != 0) {
-					if (WIFEXITED(pidstatus))
-						fprintf(stderr, "corosync died with status: %d\n", WEXITSTATUS(pidstatus));
-					if (WIFSIGNALED(pidstatus))
-						fprintf(stderr, "corosync died with signal: %d\n", WTERMSIG(pidstatus));
-					status = -1;
-					break;
-				}
-				else {
-					status = 0; /* Try to connect */
-				}
+				status = check_corosync_status(corosync_pid);
 			}
 		}
 
@@ -230,13 +303,15 @@ int join(commandline_t *comline, char *main_envp[])
 		do {
 			if (status == 0) {
 				if (kill(corosync_pid, 0) < 0) {
+					status = check_corosync_status(corosync_pid);
 					die("corosync died during startup\n");
 				}
 
 				h = cman_admin_init(NULL);
 				if (!h && comline->verbose & DEBUG_STARTUP_ONLY)
 				{
-					fprintf(stderr, "waiting for corosync to start\n");
+					fprintf(stderr, "waiting for cman to start\n");
+					status = check_corosync_status(corosync_pid);
 				}
 			}
 			sleep (1);
