@@ -13,10 +13,15 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
+#include <inttypes.h>
 
 /* corosync headers */
 #include <corosync/corotypes.h>
-#include <corosync/ipc_gen.h>
+#include <corosync/corotypes.h>
+#include <corosync/coroipc_types.h>
+#include <corosync/coroipcc.h>
+#include <corosync/corodefs.h>
+#include <corosync/mar_gen.h>
 #include <corosync/engine/coroapi.h>
 #include <corosync/engine/logsys.h>
 #include <corosync/engine/quorum.h>
@@ -26,7 +31,6 @@
 #include "cnxman-socket.h"
 #include "cnxman-private.h"
 #include "commands.h"
-#include "logging.h"
 
 #include "ais.h"
 #include "cman.h"
@@ -35,7 +39,6 @@
 #include "cmanconfig.h"
 #include "daemon.h"
 
-extern int our_nodeid();
 extern char cluster_name[MAX_CLUSTER_NAME_LEN+1];
 extern unsigned int quorumdev_poll;
 extern unsigned int ccsd_poll_interval;
@@ -50,19 +53,18 @@ int num_ais_nodes;
 quorum_set_quorate_fn_t corosync_set_quorum;
 struct memb_ring_id cman_ring_id;
 extern unsigned int config_version;
-static unsigned int cluster_parent_handle;
+static hdb_handle_t cluster_parent_handle;
 
 static int startup_pipe;
-static unsigned int debug_mask;
 static int first_trans = 1;
 struct corosync_api_v1 *corosync;
 
-static cs_tpg_handle group_handle;
+static hdb_handle_t group_handle;
 static struct corosync_tpg_group cman_group[1] = {
         { .group          = "CMAN", .group_len      = 4},
 };
 
-LOGSYS_DECLARE_SUBSYS (CMAN_NAME, LOG_INFO);
+LOGSYS_DECLARE_SUBSYS (CMAN_NAME);
 
 /* This structure is tacked onto the start of a cluster message packet for our
  * own nefarious purposes. */
@@ -82,11 +84,11 @@ struct cl_protheader {
 static int cman_exit_fn(void *conn_info);
 static int cman_exec_init_fn(struct corosync_api_v1 *api);
 static void cman_confchg_fn(enum totem_configuration_type configuration_type,
-			    unsigned int *member_list, int member_list_entries,
-			    unsigned int *left_list, int left_list_entries,
-			    unsigned int *joined_list, int joined_list_entries,
-			    struct memb_ring_id *ring_id);
-static void cman_deliver_fn(unsigned int nodeid, struct iovec *iovec, int iov_len,
+			    const unsigned int *member_list, size_t member_list_entries,
+			    const unsigned int *left_list, size_t left_list_entries,
+			    const unsigned int *joined_list, size_t joined_list_entries,
+			    const struct memb_ring_id *ring_id);
+static void cman_deliver_fn(unsigned int nodeid, const void *msg, unsigned int msg_len,
 			    int endian_conversion_required);
 static void cman_quorum_init(struct corosync_api_v1 *api, quorum_set_quorate_fn_t report);
 
@@ -100,6 +102,7 @@ static struct corosync_service_engine cman_service_handler = {
 	.lib_exit_fn		       	= cman_exit_fn,
 	.exec_init_fn		       	= cman_exec_init_fn,
 	.config_init_fn                 = NULL,
+	.sync_mode			= CS_SYNC_V1,
 };
 
 static struct corosync_service_engine *cman_get_handler_ver0(void)
@@ -161,8 +164,8 @@ static void cman_quorum_init(struct corosync_api_v1 *api, quorum_set_quorate_fn_
 
 static int cman_exec_init_fn(struct corosync_api_v1 *api)
 {
-	unsigned int object_handle;
-	unsigned int find_handle;
+	hdb_handle_t object_handle;
+	hdb_handle_t find_handle;
 	char pipe_msg[256];
 
 	corosync = api;
@@ -181,20 +184,17 @@ static int cman_exec_init_fn(struct corosync_api_v1 *api)
 		objdb_get_int(api, object_handle, "quorum_dev_poll", &quorumdev_poll, DEFAULT_QUORUMDEV_POLL);
 		objdb_get_int(api, object_handle, "shutdown_timeout", &shutdown_timeout, DEFAULT_SHUTDOWN_TIMEOUT);
 		objdb_get_int(api, object_handle, "ccsd_poll", &ccsd_poll_interval, DEFAULT_CCSD_POLL);
-		objdb_get_int(api, object_handle, "debug_mask", &debug_mask, 0);
 
-		/* All other debugging options should already have been set in preconfig */
-		set_debuglog(debug_mask);
 	}
 	corosync->object_find_destroy(find_handle);
-	P_DAEMON(CMAN_NAME " starting");
+	log_printf(LOGSYS_LEVEL_DEBUG, CMAN_NAME " starting");
 
 	/* Open local sockets and initialise I/O queues */
 	if (read_cman_config(api, &config_version)) {
 		/* An error message will have been written to cman_pipe */
 		exit(9);
 	}
-	cman_init();
+	cman_init(api);
 
 	/* Let cman_tool know we are running and our PID */
 	sprintf(pipe_msg,"SUCCESS: %d", getpid());
@@ -227,7 +227,7 @@ int comms_send_message(void *buf, int len,
 	struct cl_protheader header;
 	int totem_flags = TOTEM_AGREED;
 
-	P_AIS("comms send message %p len = %d\n", buf,len);
+	log_printf(LOGSYS_LEVEL_DEBUG, "ais: comms send message %p len = %d\n", buf,len);
 	header.tgtport = toport;
 	header.srcport = fromport;
 	header.flags   = flags;
@@ -245,45 +245,50 @@ int comms_send_message(void *buf, int len,
 	return corosync->tpg_joined_mcast(group_handle, iov, 2, totem_flags);
 }
 
-// This assumes the iovec has only one element ... is it true ??
-static void cman_deliver_fn(unsigned int nodeid, struct iovec *iovec, int iov_len,
+static void cman_deliver_fn(unsigned int nodeid, const void *msg, unsigned int msg_len,
 			    int endian_conversion_required)
 {
-	struct cl_protheader *header = iovec->iov_base;
-	char *buf = iovec->iov_base;
+	const struct cl_protheader *original_header = msg;
+	struct cl_protheader header;
+	const char *buf = msg;
 
-	P_AIS("deliver_fn source nodeid = %d, len=%d, endian_conv=%d\n",
-	      nodeid, (int)iovec->iov_len, endian_conversion_required);
+	log_printf(LOGSYS_LEVEL_DEBUG, "ais: deliver_fn source nodeid = %d, len=%d, endian_conv=%d\n",
+	      nodeid, msg_len, endian_conversion_required);
 
 	if (endian_conversion_required) {
-		header->srcid = swab32(header->srcid);
-		header->tgtid = swab32(header->tgtid);
-		header->flags = swab32(header->flags);
+		header.srcid = swab32(original_header->srcid);
+		header.tgtid = swab32(original_header->tgtid);
+		header.flags = swab32(original_header->flags);
+		header.srcport = original_header->srcport;
+		header.tgtport = original_header->tgtport;
+	} 
+	else {
+	        memcpy(&header, original_header, sizeof(header));
 	}
 
 	/* Only pass on messages for us or everyone */
-	if (header->tgtid == our_nodeid() ||
-	    header->tgtid == 0) {
-		send_to_userport(header->srcport, header->tgtport,
-				 header->srcid, header->tgtid,
-				 buf + sizeof(struct cl_protheader), iovec->iov_len - sizeof(struct cl_protheader),
+	if (header.tgtid == our_nodeid() ||
+	    header.tgtid == 0) {
+		send_to_userport(header.srcport, header.tgtport,
+				 header.srcid, header.tgtid,
+				 buf + sizeof(struct cl_protheader), msg_len - sizeof(struct cl_protheader),
 				 endian_conversion_required);
 	}
 }
 
 static void cman_confchg_fn(enum totem_configuration_type configuration_type,
-			    unsigned int *member_list, int member_list_entries,
-			    unsigned int *left_list, int left_list_entries,
-			    unsigned int *joined_list, int joined_list_entries,
-			    struct memb_ring_id *ring_id)
+			    const unsigned int *member_list, size_t member_list_entries,
+			    const unsigned int *left_list, size_t left_list_entries,
+			    const unsigned int *joined_list, size_t joined_list_entries,
+			    const struct memb_ring_id *ring_id)
 {
 	int i;
 	static int last_memb_count = 0;
-	static int saved_left_list_entries;
-	static int saved_left_list_size;
+	static size_t saved_left_list_entries;
+	static size_t saved_left_list_size;
 	static unsigned int *saved_left_list = NULL;
 
-	P_AIS("confchg_fn called type = %d, seq=%lld\n", configuration_type, ring_id->seq);
+	log_printf(LOGSYS_LEVEL_DEBUG, "ais: confchg_fn called type = %d, seq=%lld\n", configuration_type, ring_id->seq);
 
 	memcpy(&cman_ring_id, ring_id, sizeof(*ring_id));
 	incarnation = ring_id->seq;
@@ -303,7 +308,7 @@ static void cman_confchg_fn(enum totem_configuration_type configuration_type,
 			saved_left_list_size = left_list_entries*2;
 			saved_left_list = malloc(sizeof(int) * saved_left_list_size);
 			if (!saved_left_list) {
-				log_printf(LOG_LEVEL_CRIT, "cannot allocate memory for confchg message");
+				log_printf(LOGSYS_LEVEL_CRIT, "cannot allocate memory for confchg message");
 				exit(3);
 			}
 		}
@@ -311,7 +316,7 @@ static void cman_confchg_fn(enum totem_configuration_type configuration_type,
 			saved_left_list_size = left_list_entries*2;
 			saved_left_list = realloc(saved_left_list, sizeof(int) * saved_left_list_size);
 			if (!saved_left_list) {
-				log_printf(LOG_LEVEL_CRIT, "cannot reallocate memory for confchg message");
+				log_printf(LOGSYS_LEVEL_CRIT, "cannot reallocate memory for confchg message");
 				exit(3);
 			}
 		}
@@ -320,7 +325,7 @@ static void cman_confchg_fn(enum totem_configuration_type configuration_type,
 	}
 
 	if (configuration_type == TOTEM_CONFIGURATION_REGULAR) {
-		P_AIS("last memb_count = %d, current = %d\n", last_memb_count, member_list_entries);
+		log_printf(LOGSYS_LEVEL_DEBUG, "ais: last memb_count = %d, current = %"PRIuFAST32"\n", last_memb_count, member_list_entries);
 		send_transition_msg(last_memb_count, first_trans);
 		last_memb_count = member_list_entries;
 		if (member_list_entries > 1)
@@ -332,9 +337,14 @@ static void cman_confchg_fn(enum totem_configuration_type configuration_type,
 	}
 }
 
+void corosync_shutdown(void)
+{
+	corosync->shutdown_request();
+}
+
 /* Write an error message down the CMAN startup pipe so
    that cman_tool can display it */
-int write_cman_pipe(char *message)
+int write_cman_pipe(const char *message)
 {
 	if (startup_pipe)
 		return write(startup_pipe, message, strlen(message)+1);
