@@ -24,13 +24,12 @@
 #define OBJDB_API struct objdb_iface_ver0
 #include "cnxman-socket.h"
 #include "nodelist.h"
-#include "logging.h"
 
 #define MAX_PATH_LEN PATH_MAX
 
-static unsigned int debug_mask;
-static int cmanpre_readconfig(struct objdb_iface_ver0 *objdb, char **error_string);
-static int cmanpre_reloadconfig(struct objdb_iface_ver0 *objdb, int flush, char **error_string);
+static unsigned int debug;
+static int cmanpre_readconfig(struct objdb_iface_ver0 *objdb, const char **error_string);
+static int cmanpre_reloadconfig(struct objdb_iface_ver0 *objdb, int flush, const char **error_string);
 
 static char *nodename_env;
 static int expected_votes;
@@ -48,7 +47,7 @@ static char *key_filename=NULL;
 static char *mcast_name;
 static char *cluster_name;
 static char error_reason[1024] = { '\0' };
-static unsigned int cluster_parent_handle;
+static hdb_handle_t cluster_parent_handle;
 
 /*
  * Exports the interface for the service
@@ -116,21 +115,19 @@ static char *priority_name_get (unsigned int priority)
 static int ipaddr_equal(struct sockaddr_storage *addr1, struct sockaddr_storage *addr2)
 {
 	int addrlen = 0;
-	struct sockaddr *saddr1 = (struct sockaddr *)addr1;
-	struct sockaddr *saddr2 = (struct sockaddr *)addr2;
 
-	if (saddr1->sa_family != saddr2->sa_family)
+	if (addr1->ss_family != addr2->ss_family)
 		return 0;
 
-	if (saddr1->sa_family == AF_INET) {
+	if (addr1->ss_family == AF_INET) {
 		addrlen = sizeof(struct sockaddr_in);
 	}
-	if (saddr1->sa_family == AF_INET6) {
+	if (addr1->ss_family == AF_INET6) {
 		addrlen = sizeof(struct sockaddr_in6);
 	}
 	assert(addrlen);
 
-	if (memcmp(saddr1, saddr2, addrlen) == 0)
+	if (memcmp(addr1, addr2, addrlen) == 0)
 		return 1;
 	else
 		return 0;
@@ -140,7 +137,7 @@ static int ipaddr_equal(struct sockaddr_storage *addr1, struct sockaddr_storage
 /* Build a localhost ip_address */
 static int get_localhost(int family, struct sockaddr_storage *localhost)
 {
-	char *addr_text;
+	const char *addr_text;
 	struct addrinfo *ainfo;
 	struct addrinfo ahints;
 	int ret;
@@ -169,7 +166,7 @@ static int get_localhost(int family, struct sockaddr_storage *localhost)
 }
 
 /* Return the address family of an IP[46] name */
-static int address_family(char *addr, struct sockaddr_storage *ssaddr)
+static int address_family(char *addr, struct sockaddr_storage *ssaddr, int family_hint)
 {
 	struct addrinfo *ainfo;
 	struct addrinfo ahints;
@@ -179,6 +176,7 @@ static int address_family(char *addr, struct sockaddr_storage *ssaddr)
 	memset(&ahints, 0, sizeof(ahints));
 	ahints.ai_socktype = SOCK_DGRAM;
 	ahints.ai_protocol = IPPROTO_UDP;
+	ahints.ai_family = family_hint;
 
 	/* Lookup the nodename address */
 	ret = getaddrinfo(addr, NULL, &ahints, &ainfo);
@@ -197,48 +195,40 @@ static int address_family(char *addr, struct sockaddr_storage *ssaddr)
 /* Find the "CMAN" logger_subsys object. Or create one if it does not
    exist
 */
-static unsigned int find_cman_logger(struct objdb_iface_ver0 *objdb, unsigned int object_handle)
+static hdb_handle_t find_cman_logger(struct objdb_iface_ver0 *objdb, hdb_handle_t object_handle)
 {
-	unsigned int subsys_handle;
-	unsigned int find_handle;
+	hdb_handle_t subsys_handle;
+	hdb_handle_t find_handle;
 	char *str;
 
 	objdb->object_find_create(object_handle, "logger_subsys", strlen("logger_subsys"), &find_handle);
-	while (!objdb->object_find_next(object_handle, &subsys_handle)) {
-
-		if (objdb_get_string(objdb, subsys_handle, "subsys", &str)) {
-			continue;
+	while (!objdb->object_find_next(find_handle, &subsys_handle)) {
+		if (!objdb_get_string(objdb, subsys_handle, "subsys", &str)) {
+			if (strncmp(str, CMAN_NAME, 4) == 0) {
+				objdb->object_find_destroy(find_handle);
+				return subsys_handle;
+			}
 		}
-		if (strcmp(str, CMAN_NAME) == 0)
-			return subsys_handle;
 	}
 	objdb->object_find_destroy(find_handle);
 
-	/* We can't find it ... create one */
-	if (objdb->object_create(object_handle, &subsys_handle,
-				    "logger_subsys", strlen("logger_subsys")) == 0) {
-
-		objdb->object_key_create(subsys_handle, "subsys", strlen("subsys"),
-					    CMAN_NAME, strlen(CMAN_NAME)+1);
-	}
-
-	return subsys_handle;
+	return -1;
 
 }
 
 
-static int add_ifaddr(struct objdb_iface_ver0 *objdb, char *mcast, char *ifaddr, int portnum)
+static int add_ifaddr(struct objdb_iface_ver0 *objdb, char *mcast, char *ifaddr, int port, int broadcast)
 {
-	unsigned int totem_object_handle;
-	unsigned int find_handle;
-	unsigned int interface_object_handle;
+	hdb_handle_t totem_object_handle;
+	hdb_handle_t find_handle;
+	hdb_handle_t interface_object_handle;
 	struct sockaddr_storage if_addr, localhost, mcast_addr;
 	char tmp[132];
 	int ret = 0;
 
 	/* Check the families match */
-	if (address_family(mcast, &mcast_addr) !=
-	    address_family(ifaddr, &if_addr)) {
+	if (address_family(mcast, &mcast_addr, 0) !=
+	    address_family(ifaddr, &if_addr, mcast_addr.ss_family)) {
 		sprintf(error_reason, "Node address family does not match multicast address family");
 		return -1;
 	}
@@ -260,18 +250,30 @@ static int add_ifaddr(struct objdb_iface_ver0 *objdb, char *mcast, char *ifaddr,
 
 	if (objdb->object_create(totem_object_handle, &interface_object_handle,
 				 "interface", strlen("interface")) == 0) {
+		struct sockaddr_in *in = (struct sockaddr_in *)&if_addr;
+		struct sockaddr_in6 *in6= (struct sockaddr_in6 *)&if_addr;
+		void *addrptr;
 
 		sprintf(tmp, "%d", num_interfaces);
 		objdb->object_key_create(interface_object_handle, "ringnumber", strlen("ringnumber"),
 					 tmp, strlen(tmp)+1);
 
+		if (if_addr.ss_family == AF_INET)
+			addrptr = &in->sin_addr;
+		else
+			addrptr = &in6->sin6_addr;
+		inet_ntop(if_addr.ss_family, addrptr, tmp, sizeof(tmp));
 		objdb->object_key_create(interface_object_handle, "bindnetaddr", strlen("bindnetaddr"),
-					 ifaddr, strlen(ifaddr)+1);
+					 tmp, strlen(tmp)+1);
 
-		objdb->object_key_create(interface_object_handle, "mcastaddr", strlen("mcastaddr"),
-					 mcast, strlen(mcast)+1);
+		if (broadcast)
+			objdb->object_key_create(interface_object_handle, "broadcast", strlen("broadcast"),
+						 "yes", strlen("yes")+1);
+		else
+			objdb->object_key_create(interface_object_handle, "mcastaddr", strlen("mcastaddr"),
+						 mcast, strlen(mcast)+1);
 
-		sprintf(tmp, "%d", portnum);
+		sprintf(tmp, "%d", port);
 		objdb->object_key_create(interface_object_handle, "mcastport", strlen("mcastport"),
 					 tmp, strlen(tmp)+1);
 
@@ -293,7 +295,7 @@ static uint16_t generate_cluster_id(char *name)
 	return value & 0xFFFF;
 }
 
-static char *default_mcast(char *nodename, uint16_t cluster_id)
+static char *default_mcast(char *node, uint16_t clusterid)
 {
         struct addrinfo *ainfo;
         struct addrinfo ahints;
@@ -305,9 +307,9 @@ static char *default_mcast(char *nodename, uint16_t cluster_id)
 
         /* Lookup the the nodename address and use it's IP type to
 	   default a multicast address */
-        ret = getaddrinfo(nodename, NULL, &ahints, &ainfo);
+        ret = getaddrinfo(node, NULL, &ahints, &ainfo);
 	if (ret) {
-		sprintf(error_reason, "Can't determine address family of nodename %s\n", nodename);
+		sprintf(error_reason, "Can't determine address family of nodename %s\n", node);
 		write_cman_pipe("Can't determine address family of nodename");
 		return NULL;
 	}
@@ -316,40 +318,40 @@ static char *default_mcast(char *nodename, uint16_t cluster_id)
 	freeaddrinfo(ainfo);
 
 	if (family == AF_INET) {
-		snprintf(addr, sizeof(addr), "239.192.%d.%d", cluster_id >> 8, cluster_id % 0xFF);
+		snprintf(addr, sizeof(addr), "239.192.%d.%d", clusterid >> 8, clusterid % 0xFF);
 		return addr;
 	}
 	if (family == AF_INET6) {
-		snprintf(addr, sizeof(addr), "ff15::%x", cluster_id);
+		snprintf(addr, sizeof(addr), "ff15::%x", clusterid);
 		return addr;
 	}
 
 	return NULL;
 }
 
-static int verify_nodename(struct objdb_iface_ver0 *objdb, char *nodename)
+static int verify_nodename(struct objdb_iface_ver0 *objdb, char *node)
 {
 	char nodename2[MAX_CLUSTER_MEMBER_NAME_LEN+1];
 	char nodename3[MAX_CLUSTER_MEMBER_NAME_LEN+1];
 	char *str, *dot = NULL;
 	struct ifaddrs *ifa, *ifa_list;
 	struct sockaddr *sa;
-	unsigned int nodes_handle;
-	unsigned int find_handle = 0;
+	hdb_handle_t nodes_handle;
+	hdb_handle_t find_handle = 0;
 	int error;
 
 	/* nodename is either from commandline or from uname */
-	if (nodelist_byname(objdb, cluster_parent_handle, nodename))
+	if (nodelist_byname(objdb, cluster_parent_handle, node))
 		return 0;
 
 	/* If nodename was from uname, try a domain-less version of it */
-	strcpy(nodename2, nodename);
+	strcpy(nodename2, node);
 	dot = strchr(nodename2, '.');
 	if (dot) {
 		*dot = '\0';
 
 		if (nodelist_byname(objdb, cluster_parent_handle, nodename2)) {
-			strcpy(nodename, nodename2);
+			strcpy(node, nodename2);
 			return 0;
 		}
 	}
@@ -375,7 +377,7 @@ static int verify_nodename(struct objdb_iface_ver0 *objdb, char *nodename)
 
 		if (strlen(nodename2) == len &&
 		    !strncmp(nodename2, nodename3, len)) {
-			strcpy(nodename, str);
+			strcpy(node, str);
 			return 0;
 		}
 		nodes_handle = nodeslist_next(objdb, find_handle);
@@ -391,42 +393,50 @@ static int verify_nodename(struct objdb_iface_ver0 *objdb, char *nodename)
 		return -1;
 
 	for (ifa = ifa_list; ifa; ifa = ifa->ifa_next) {
+		socklen_t salen = 0;
+
 		/* Restore this */
-		strcpy(nodename2, nodename);
+		strcpy(nodename2, node);
 		sa = ifa->ifa_addr;
-		if (!sa || sa->sa_family != AF_INET)
+		if (!sa)
+			continue;
+		if (sa->sa_family != AF_INET && sa->sa_family != AF_INET6)
 			continue;
 
-		error = getnameinfo(sa, sizeof(*sa), nodename2,
+		if (sa->sa_family == AF_INET)
+			salen = sizeof(struct sockaddr_in);
+		if (sa->sa_family == AF_INET6)
+			salen = sizeof(struct sockaddr_in6);
+
+		error = getnameinfo(sa, salen, nodename2,
 				    sizeof(nodename2), NULL, 0, 0);
-		if (error)
-			goto out;
+		if (!error) {
 
-		if (nodelist_byname(objdb, cluster_parent_handle, nodename2)) {
-			strcpy(nodename, nodename2);
-			goto out;
-		}
+			if (nodelist_byname(objdb, cluster_parent_handle, nodename2)) {
+				strcpy(node, nodename2);
+				goto out;
+			}
 
-		/* truncate this name and try again */
+			/* Truncate this name and try again */
+			dot = strchr(nodename2, '.');
+			if (dot) {
+				*dot = '\0';
 
-		dot = strchr(nodename2, '.');
-		if (!dot)
-			continue;
-		*dot = '\0';
-
-		if (nodelist_byname(objdb, cluster_parent_handle, nodename2)) {
-			strcpy(nodename, nodename2);
-			goto out;
+				if (nodelist_byname(objdb, cluster_parent_handle, nodename2)) {
+					strcpy(node, nodename2);
+					goto out;
+				}
+			}
 		}
 
 		/* See if it's the IP address that's in cluster.conf */
 		error = getnameinfo(sa, sizeof(*sa), nodename2,
 				    sizeof(nodename2), NULL, 0, NI_NUMERICHOST);
 		if (error)
-			goto out;
+			continue;
 
 		if (nodelist_byname(objdb, cluster_parent_handle, nodename2)) {
-			strcpy(nodename, nodename2);
+			strcpy(node, nodename2);
 			goto out;
 		}
 	}
@@ -438,7 +448,7 @@ static int verify_nodename(struct objdb_iface_ver0 *objdb, char *nodename)
 }
 
 /* Get any environment variable overrides */
-static int get_env_overrides()
+static int get_env_overrides(void)
 {
 	if (getenv("CMAN_CLUSTER_NAME")) {
 		cluster_name = strdup(getenv("CMAN_CLUSTER_NAME"));
@@ -487,8 +497,10 @@ static int get_env_overrides()
 		expected_votes = 1;
 		votes = 1;
 	}
-	if (getenv("CMAN_DEBUGLOG")) {
-		debug_mask = atoi(getenv("CMAN_DEBUGLOG"));
+	if (getenv("CMAN_DEBUG")) {
+		debug = atoi(getenv("CMAN_DEBUG"));
+		if (debug > 0)
+			debug = 1;
 	}
 
 	return 0;
@@ -498,10 +510,12 @@ static int get_env_overrides()
 static int get_nodename(struct objdb_iface_ver0 *objdb)
 {
 	char *nodeid_str = NULL;
-	unsigned int object_handle;
-	unsigned int find_handle;
-	unsigned int node_object_handle;
-	unsigned int alt_object;
+	hdb_handle_t object_handle;
+	hdb_handle_t find_handle;
+	hdb_handle_t node_object_handle;
+	hdb_handle_t alt_object;
+	int broadcast = 0;
+	char *str;
 	int error;
 
 	if (!getenv("CMAN_NOCONFIG")) {
@@ -561,8 +575,8 @@ static int get_nodename(struct objdb_iface_ver0 *objdb)
 
 	if (objdb->object_find_next(find_handle, &object_handle) == 0) {
 
-		unsigned int mcast_handle;
-		unsigned int find_handle2;
+		hdb_handle_t mcast_handle;
+		hdb_handle_t find_handle2;
 
 		if (!mcast_name) {
 
@@ -576,7 +590,10 @@ static int get_nodename(struct objdb_iface_ver0 *objdb)
 
 		if (!mcast_name) {
 			mcast_name = default_mcast(nodename, cluster_id);
+
 		}
+		if (!mcast_name)
+			return -1;
 
 		/* See if the user wants our default set of openais services (default=yes) */
 		objdb_get_int(objdb, object_handle, "disable_openais", &disable_openais, 0);
@@ -594,7 +611,18 @@ static int get_nodename(struct objdb_iface_ver0 *objdb)
 		objdb_get_int(objdb, object_handle, "port", &portnum, DEFAULT_PORT);
 	}
 
-	if (add_ifaddr(objdb, mcast_name, nodename, portnum))
+	/* Check for broadcast */
+	if (!objdb_get_string(objdb, object_handle, "broadcast", &str)) {
+		if (strcmp(str, "yes") == 0) {
+			mcast_name = strdup("255.255.255.255");
+			if (!mcast_name)
+				return -1;
+			broadcast = 1;
+		}
+		free(str);
+	}
+
+	if (add_ifaddr(objdb, mcast_name, nodename, portnum, broadcast))
 		return -1;
 
 	/* Get all alternative node names */
@@ -602,10 +630,10 @@ static int get_nodename(struct objdb_iface_ver0 *objdb)
 	objdb->object_find_create(node_object_handle,"altname", strlen("altname"), &find_handle);
 	while (objdb->object_find_next(find_handle, &alt_object) == 0) {
 		unsigned int port;
-		char *nodename;
+		char *node;
 		char *mcast;
 
-		if (objdb_get_string(objdb, alt_object, "name", &nodename)) {
+		if (objdb_get_string(objdb, alt_object, "name", &node)) {
 			continue;
 		}
 
@@ -615,7 +643,7 @@ static int get_nodename(struct objdb_iface_ver0 *objdb)
 			mcast = mcast_name;
 		}
 
-		if (add_ifaddr(objdb, mcast, nodename, portnum))
+		if (add_ifaddr(objdb, mcast, node, portnum, broadcast))
 			return -1;
 
 		num_nodenames++;
@@ -629,11 +657,11 @@ out:
 /* These are basically cman overrides to the totem config bits */
 static void add_cman_overrides(struct objdb_iface_ver0 *objdb)
 {
-	unsigned int logger_object_handle;
 	char *logstr;
 	char *logfacility;
-	unsigned int object_handle;
-	unsigned int find_handle;
+	char *loglevel;
+	hdb_handle_t object_handle;
+	hdb_handle_t find_handle;
 	char tmp[256];
 
 	/* "totem" key already exists, because we have added the interfaces by now */
@@ -730,53 +758,84 @@ static void add_cman_overrides(struct objdb_iface_ver0 *objdb)
 	objdb->object_find_destroy(find_handle);
 
 	logfacility = facility_name_get(SYSLOGFACILITY);
+	loglevel = priority_name_get(SYSLOGLEVEL);
 
-	logger_object_handle = find_cman_logger(objdb, object_handle);
+	/* enable timestamps on logging */
+	if (objdb_get_string(objdb, object_handle, "timestamp", &logstr)) {
+		objdb->object_key_create(object_handle, "timestamp", strlen("timestamp"),
+					    "on", strlen("on")+1);
+	}
+
+	/* configure logfile */
+	if (objdb_get_string(objdb, object_handle, "to_logfile", &logstr)) {
+		objdb->object_key_create(object_handle, "to_logfile", strlen("to_logfile"),
+					    "yes", strlen("yes")+1);
+	}
+
+	if (objdb_get_string(objdb, object_handle, "logfile", &logstr)) {
+		objdb->object_key_create(object_handle, "logfile", strlen("logfile"),
+					    LOGDIR "/corosync.log", strlen(LOGDIR "/corosync.log")+1);
+	}
+
+	if (objdb_get_string(objdb, object_handle, "logfile_priority", &logstr)) {
+		objdb->object_key_create(object_handle, "logfile_priority", strlen("logfile_priority"),
+					    loglevel, strlen(loglevel)+1);
+	}
+
+	/* syslog */
+	if (objdb_get_string(objdb, object_handle, "to_syslog", &logstr)) {
+		objdb->object_key_create(object_handle, "to_syslog", strlen("to_syslog"),
+					    "yes", strlen("yes")+1);
+	}
 
 	if (objdb_get_string(objdb, object_handle, "syslog_facility", &logstr)) {
 		objdb->object_key_create(object_handle, "syslog_facility", strlen("syslog_facility"),
 					    logfacility, strlen(logfacility)+1);
 	}
 
-	if (objdb_get_string(objdb, object_handle, "to_file", &logstr)) {
-		objdb->object_key_create(object_handle, "to_file", strlen("to_file"),
-					    "yes", strlen("yes")+1);
-	}
-
-	if (objdb_get_string(objdb, object_handle, "logfile", &logstr)) {
-		objdb->object_key_create(object_handle, "logfile", strlen("logfile"),
-					    LOGDIR "/cman.log", strlen(LOGDIR "/cman.log")+1);
-	}
-
-	if (debug_mask) {
-		objdb->object_key_create(object_handle, "to_stderr", strlen("to_stderr"),
-					    "yes", strlen("yes")+1);
-		objdb->object_key_create(logger_object_handle, "debug", strlen("debug"),
-					    "on", strlen("on")+1);
-		objdb->object_key_create(logger_object_handle, "syslog_level", strlen("syslog_level"),
-					    "debug", strlen("debug")+1);
-
-	}
-	else {
-		char *loglevel;
-		loglevel = priority_name_get(SYSLOGLEVEL);
-		objdb->object_key_create(logger_object_handle, "syslog_level", strlen("syslog_level"),
+	if (objdb_get_string(objdb, object_handle, "syslog_priority", &logstr)) {
+		objdb->object_key_create(object_handle, "syslog_priority", strlen("syslog_priority"),
 					    loglevel, strlen(loglevel)+1);
 	}
 
+	if (!debug) {
+		hdb_handle_t logger_object_handle;
 
-	/* Don't run under user "ais" */
+		if (!objdb_get_string(objdb, object_handle, "debug", &logstr)) {
+			if (!strncmp(logstr, "on", 2)) {
+				debug=1;
+			}
+		}
+
+		logger_object_handle = find_cman_logger(objdb, object_handle);
+		if (logger_object_handle > -1) {
+			if (!objdb_get_string(objdb, logger_object_handle, "debug", &logstr)) {
+				if (!strncmp(logstr, "on", 2)) {
+					debug=1;
+				}
+				if (!strncmp(logstr, "off", 3)) {
+					debug=0;
+				}
+			}
+		}
+	}
+
+	if (debug) {
+		objdb->object_key_create(object_handle, "to_stderr", strlen("to_stderr"),
+					    "yes", strlen("yes")+1);
+	}
+
+	/* Make sure we allow connections from user/group "ais" */
 	objdb->object_find_create(OBJECT_PARENT_HANDLE, "aisexec", strlen("aisexec"), &find_handle);
 	if (objdb->object_find_next(find_handle, &object_handle) != 0) {
 		objdb->object_create(OBJECT_PARENT_HANDLE, &object_handle,
 					"aisexec", strlen("aisexec"));
-
 	}
 	objdb->object_find_destroy(find_handle);
 	objdb->object_key_create(object_handle, "user", strlen("user"),
-				    "root", strlen("root") + 1);
+				    "ais", strlen("ais") + 1);
 	objdb->object_key_create(object_handle, "group", strlen("group"),
-				    "root", strlen("root") + 1);
+				    "ais", strlen("ais") + 1);
 
 	objdb->object_find_create(cluster_parent_handle, "cman", strlen("cman"), &find_handle);
 	if (objdb->object_find_next(find_handle, &object_handle) == 0)
@@ -793,14 +852,17 @@ static void add_cman_overrides(struct objdb_iface_ver0 *objdb)
 			objdb->object_key_create(object_handle, "two_node", strlen("two_node"),
 						 str, strlen(str) + 1);
 		}
-		if (debug_mask) {
-			sprintf(str, "%d", debug_mask);
-			objdb->object_key_create(object_handle, "debug_mask", strlen("debug_mask"),
-						 str, strlen(str) + 1);
-		}
 	}
 	objdb->object_find_destroy(find_handle);
 
+	/* Load the quorum service */
+	objdb->object_create(OBJECT_PARENT_HANDLE, &object_handle,
+			     "service", strlen("service"));
+	objdb->object_key_create(object_handle, "name", strlen("name"),
+				 "corosync_quorum", strlen("corosync_quorum") + 1);
+	objdb->object_key_create(object_handle, "ver", strlen("ver"),
+				 "0", 2);
+
 	/* Make sure we load our alter-ego - the main cman module */
 	objdb->object_create(OBJECT_PARENT_HANDLE, &object_handle,
 			     "service", strlen("service"));
@@ -825,14 +887,25 @@ static void add_cman_overrides(struct objdb_iface_ver0 *objdb)
 static int set_noccs_defaults(struct objdb_iface_ver0 *objdb)
 {
 	char tmp[255];
-	unsigned int object_handle;
-	unsigned int find_handle;
+	hdb_handle_t object_handle;
+	hdb_handle_t find_handle;
 
 	/* Enforce key */
-	key_filename = NOCCS_KEY_FILENAME;
+	key_filename = strdup(NOCCS_KEY_FILENAME);
+	if (!key_filename) {
+		sprintf(error_reason, "cannot allocate memory for key file name");
+		write_cman_pipe("cannot allocate memory for key file name");
+		return -1;
+	}
 
 	if (!cluster_name)
-		cluster_name = DEFAULT_CLUSTER_NAME;
+		cluster_name = strdup(DEFAULT_CLUSTER_NAME);
+
+	if (!cluster_name) {
+		sprintf(error_reason, "cannot allocate memory for cluster_name");
+		write_cman_pipe("cannot allocate memory for cluster_name");
+		return -1;
+	}
 
 	if (!cluster_id)
 		cluster_id = generate_cluster_id(cluster_name);
@@ -931,24 +1004,24 @@ static int set_noccs_defaults(struct objdb_iface_ver0 *objdb)
 }
 
 /* Move an object/key tree */
-static int copy_config_tree(struct objdb_iface_ver0 *objdb, unsigned int source_object, unsigned int target_parent_object, int always_create)
+static int copy_config_tree(struct objdb_iface_ver0 *objdb, hdb_handle_t source_object, hdb_handle_t target_parent_object, int always_create)
 {
-	unsigned int object_handle;
-	unsigned int new_object;
-	unsigned int find_handle;
+	hdb_handle_t object_handle;
+	hdb_handle_t new_object;
+	hdb_handle_t find_handle;
 	char object_name[1024];
-	int object_name_len;
+	size_t object_name_len;
 	void *key_name;
-	int key_name_len;
+	size_t key_name_len;
 	void *key_value;
-	int key_value_len;
+	size_t key_value_len;
 	int res;
 
 	/* Create new parent object if necessary */
 	objdb->object_name_get(source_object, object_name, &object_name_len);
 
 	objdb->object_find_create(target_parent_object, object_name, strlen(object_name), &find_handle);
-	if (always_create || objdb->object_find_next(find_handle, &object_handle))
+	if (always_create || objdb->object_find_next(find_handle, &new_object))
 			objdb->object_create(target_parent_object, &new_object, object_name, object_name_len);
 	objdb->object_find_destroy(find_handle);
 
@@ -965,7 +1038,7 @@ static int copy_config_tree(struct objdb_iface_ver0 *objdb, unsigned int source_
 	/* Create sub-objects */
 	res = objdb->object_find_create(source_object, NULL, 0, &find_handle);
 	if (res) {
-		sprintf(error_reason, "error resetting object iterator for object %d: %d\n", source_object, res);
+		sprintf(error_reason, "error resetting object iterator for object %ud: %d\n", (unsigned int)source_object, res);
 		return -1;
 	}
 
@@ -983,10 +1056,10 @@ static int copy_config_tree(struct objdb_iface_ver0 *objdb, unsigned int source_
  * Copy trees from /cluster where they live in cluster.conf, into the root
  * of the config tree where corosync expects to find them.
  */
-static int copy_tree_to_root(struct objdb_iface_ver0 *objdb, char *name, int always_create)
+static int copy_tree_to_root(struct objdb_iface_ver0 *objdb, const char *name, int always_create)
 {
-	unsigned int find_handle;
-	unsigned int object_handle;
+	hdb_handle_t find_handle;
+	hdb_handle_t object_handle;
 	int res=0;
 
 	objdb->object_find_create(cluster_parent_handle, name, strlen(name), &find_handle);
@@ -1000,8 +1073,8 @@ static int copy_tree_to_root(struct objdb_iface_ver0 *objdb, char *name, int alw
 
 static int get_cman_globals(struct objdb_iface_ver0 *objdb)
 {
-	unsigned int object_handle;
-	unsigned int find_handle;
+	hdb_handle_t object_handle;
+	hdb_handle_t find_handle;
 
 	objdb_get_string(objdb, cluster_parent_handle, "name", &cluster_name);
 
@@ -1024,12 +1097,12 @@ static int get_cman_globals(struct objdb_iface_ver0 *objdb)
 	return 0;
 }
 
-static int cmanpre_reloadconfig(struct objdb_iface_ver0 *objdb, int flush, char **error_string)
+static int cmanpre_reloadconfig(struct objdb_iface_ver0 *objdb, int flush, const char **error_string)
 {
 	int ret = -1;
-	unsigned int object_handle;
-	unsigned int find_handle;
-	unsigned int cluster_parent_handle_new;
+	hdb_handle_t object_handle;
+	hdb_handle_t find_handle;
+	hdb_handle_t cluster_parent_handle_new;
 
 	/* don't reload if we've been told to run configless */
 	if (getenv("CMAN_NOCONFIG")) {
@@ -1060,22 +1133,17 @@ static int cmanpre_reloadconfig(struct objdb_iface_ver0 *objdb, int flush, char
 
 	/* destroy top level /logging */
 	objdb->object_find_create(OBJECT_PARENT_HANDLE, "logging", strlen("logging"), &find_handle);
-	objdb->object_find_next(find_handle, &object_handle);
+	ret = objdb->object_find_next(find_handle, &object_handle);
 	objdb->object_find_destroy(find_handle);
-	if (object_handle) {
+	if (ret) {
 		objdb->object_destroy(object_handle);
 	}
 
 	/* copy /cluster/logging to /logging */
 	ret = copy_tree_to_root(objdb, "logging", 0);
 
-	/* destroy top level /totem */
-	objdb->object_find_create(OBJECT_PARENT_HANDLE, "totem", strlen("totem"), &find_handle);
-	objdb->object_find_next(find_handle, &object_handle);
-	objdb->object_find_destroy(find_handle);
-	if (object_handle) {
-		objdb->object_destroy(object_handle);
-	}
+	/* Note: we do NOT delete /totem as corosync stores other things in there that
+	   it needs! */
 
 	/* copy /cluster/totem to /totem */
 	ret = copy_tree_to_root(objdb, "totem", 0);
@@ -1087,11 +1155,11 @@ err:
 	return ret;
 }
 
-static int cmanpre_readconfig(struct objdb_iface_ver0 *objdb, char **error_string)
+static int cmanpre_readconfig(struct objdb_iface_ver0 *objdb, const char **error_string)
 {
 	int ret = 0;
-	unsigned int object_handle;
-	unsigned int find_handle;
+	hdb_handle_t object_handle;
+	hdb_handle_t find_handle;
 
 	if (getenv("CMAN_PIPE"))
                 startup_pipe = atoi(getenv("CMAN_PIPE"));
@@ -1163,7 +1231,7 @@ static int cmanpre_readconfig(struct objdb_iface_ver0 *objdb, char **error_strin
 
 /* Write an error message down the CMAN startup pipe so
    that cman_tool can display it */
-int write_cman_pipe(char *message)
+int write_cman_pipe(const char *message)
 {
 	if (startup_pipe)
 		return write(startup_pipe, message, strlen(message)+1);
