@@ -20,7 +20,11 @@
 #include <sys/errno.h>
 #include <dlfcn.h>
 
-#include <corosync/ipc_gen.h>
+#include <corosync/corotypes.h>
+#include <corosync/coroipc_types.h>
+#include <corosync/coroipcc.h>
+#include <corosync/corodefs.h>
+#include <corosync/mar_gen.h>
 #include <corosync/engine/coroapi.h>
 #include <corosync/engine/logsys.h>
 #include "list.h"
@@ -29,7 +33,6 @@
 #include "cnxman-private.h"
 #include "daemon.h"
 #include "barrier.h"
-#include "logging.h"
 #define OBJDB_API struct corosync_api_v1
 #include "cmanconfig.h"
 #include "nodelist.h"
@@ -37,7 +40,7 @@
 #include "ais.h"
 
 #define max(a,b) (((a) > (b)) ? (a) : (b))
-LOGSYS_DECLARE_SUBSYS (CMAN_NAME, LOG_INFO);
+LOGSYS_DECLARE_SUBSYS (CMAN_NAME);
 
 /* Reference counting for cluster applications */
 static int use_count;
@@ -89,7 +92,9 @@ static int send_port_enquire(int nodeid);
 static void process_internal_message(char *data, int nodeid, int byteswap);
 static void recalculate_quorum(int allow_decrease, int by_current_nodes);
 static void send_kill(int nodeid, uint16_t reason);
-static char *killmsg_reason(int reason);
+static const char *killmsg_reason(int reason);
+static void ccsd_timer_fn(void *arg);
+static int reread_config(int new_version);
 
 static void set_port_bit(struct cluster_node *node, uint8_t port)
 {
@@ -174,6 +179,10 @@ static int totemip_to_sockaddr(struct totem_ip_address *ip_addr,
 static void set_quorate(int total_votes)
 {
 	int quorate;
+	unsigned int nodelist[PROCESSOR_COUNT_MAX];
+	int nodecount = 0;
+	struct cluster_node *node = NULL;
+	struct list *tmp;
 
 	if (quorum > total_votes || config_error) {
 		quorate = 0;
@@ -189,8 +198,6 @@ static void set_quorate(int total_votes)
 
 	/* If we are newly quorate, then kill any AISONLY nodes */
 	if (!cluster_is_quorate && quorate) {
-		struct cluster_node *node = NULL;
-		struct list *tmp;
 
 		list_iterate(tmp, &cluster_members_list) {
 			node = list_item(tmp, struct cluster_node);
@@ -201,6 +208,15 @@ static void set_quorate(int total_votes)
 
 	cluster_is_quorate = quorate;
 
+	/* Inform corosync subsystems */
+	list_iterate(tmp, &cluster_members_list) {
+		node = list_item(tmp, struct cluster_node);
+		if (node->state == NODESTATE_MEMBER) {
+			nodelist[nodecount++] = node->node_id;
+		}
+	}
+
+	corosync_set_quorum(nodelist, nodecount, quorate, &cman_ring_id);
 }
 
 static void node_add_ordered(struct cluster_node *newnode)
@@ -270,7 +286,7 @@ static struct cluster_node *add_new_node(char *name, int nodeid, int votes, int
 
 		newname = strdup(name);
 		if (newname) {
-			P_MEMB("replacing old node name %s with %s\n", newnode->name, name);
+			log_printf(LOGSYS_LEVEL_DEBUG, "memb: replacing old node name %s with %s\n", newnode->name, name);
 			free(newnode->name);
 			newnode->name = newname;
 		}
@@ -281,7 +297,7 @@ static struct cluster_node *add_new_node(char *name, int nodeid, int votes, int
 
 	newnode->flags |= NODE_FLAGS_REREAD;
 
-	P_MEMB("add_new_node: %s, (id=%d, votes=%d) newalloc=%d\n",
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: add_new_node: %s, (id=%d, votes=%d) newalloc=%d\n",
 	       name, nodeid, votes, newalloc);
 
 	return newnode;
@@ -464,6 +480,18 @@ static int do_cmd_set_version(char *cmdbuf, int *retlen)
 	if (config_version == version->config)
 		return 0;
 
+	/* If the passed-in version number is 0 then read the file now, then
+	 * tell the other nodes to look for that version number.
+	 * That means we also have to send the notification here, because it will
+	 * beskipped when we get our own RECONFIGURE message back as the version
+	 * number will match.
+	 */
+	if (!version->config) {
+		if (!reread_config(0))
+			notify_listeners(NULL, EVENT_REASON_CONFIG_UPDATE, config_version);
+		version->config = config_version;
+	}
+
 	/* We will re-read CCS when we get our own message back */
 	send_reconfigure(us->node_id, RECONFIG_PARAM_CONFIG_VERSION, version->config);
 	return 0;
@@ -479,10 +507,10 @@ static int do_cmd_get_extrainfo(char *cmdbuf, char **retbuf, int retsize, int *r
 	int addrlen;
 	int uncounted = 0;
 	unsigned int num_interfaces;
-	unsigned int totem_object_handle;
-	unsigned int object_handle;
-	unsigned int totem_find_handle;
-	unsigned int iface_find_handle;
+	hdb_handle_t totem_object_handle;
+	hdb_handle_t object_handle;
+	hdb_handle_t totem_find_handle;
+	hdb_handle_t iface_find_handle;
 	char **status;
 	struct cluster_node *node;
 	struct sockaddr_storage *ss;
@@ -510,10 +538,12 @@ static int do_cmd_get_extrainfo(char *cmdbuf, char **retbuf, int retsize, int *r
 		       sizeof(struct sockaddr_storage) * (MAX_INTERFACES*2))) {
 
 		*retbuf = malloc(sizeof(struct cl_extra_info) + sizeof(struct sockaddr_storage) * (MAX_INTERFACES*2));
+		if (*retbuf == NULL)
+			return -ENOMEM;
 		outbuf = *retbuf + offset;
 		einfo = (struct cl_extra_info *)outbuf;
 
-		P_MEMB("get_extrainfo: allocated new buffer\n");
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: get_extrainfo: allocated new buffer\n");
 	}
 
 	einfo->node_state = us->state;
@@ -551,6 +581,9 @@ static int do_cmd_get_extrainfo(char *cmdbuf, char **retbuf, int retsize, int *r
 			struct sockaddr_in6 *saddr6;
 
 			objdb_get_string(corosync, object_handle, "mcastaddr", &mcast);
+			/* If this fails, it must be using broadcast*/
+			if (!mcast)
+				mcast = (char*)"255.255.255.255";
 			memset(ptr, 0, sizeof(struct sockaddr_storage));
 
 			saddr4 = (struct sockaddr_in *)ptr;
@@ -604,8 +637,10 @@ static int do_cmd_get_all_members(char *cmdbuf, char **retbuf, int retsize, int
 		/* If there is not enough space in the default buffer, allocate some more. */
 		if ((retsize / sizeof(struct cl_cluster_node)) < total_nodes) {
 			*retbuf = malloc(sizeof(struct cl_cluster_node) * total_nodes + offset);
+			if (!*retbuf)
+				return -ENOMEM;
 			outbuf = *retbuf + offset;
-			P_MEMB("get_all_members: allocated new buffer (retsize=%d)\n", retsize);
+			log_printf(LOGSYS_LEVEL_DEBUG, "memb: get_all_members: allocated new buffer (retsize=%d)\n", retsize);
 		}
 	}
 	user_node = (struct cl_cluster_node *)outbuf;
@@ -627,7 +662,7 @@ static int do_cmd_get_all_members(char *cmdbuf, char **retbuf, int retsize, int
 	}
 
 	*retlen = sizeof(struct cl_cluster_node) * num_nodes;
-	P_MEMB("get_all_members: retlen = %d\n", *retlen);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: get_all_members: retlen = %d\n", *retlen);
 	return num_nodes;
 }
 
@@ -669,7 +704,7 @@ static int do_cmd_get_node(char *cmdbuf, char *retbuf, int *retlen)
 			node = find_node_by_name(u_node->name);
 
 		if (!node) {
-			P_MEMB("cmd_get_node failed: id=%d, name='%s'\n", u_node->node_id, u_node->name);
+			log_printf(LOGSYS_LEVEL_DEBUG, "memb: cmd_get_node failed: id=%d, name='%s'\n", u_node->node_id, u_node->name);
 			return -ENOENT;
 		}
 	}
@@ -680,6 +715,37 @@ static int do_cmd_get_node(char *cmdbuf, char *retbuf, int *retlen)
 	return 0;
 }
 
+static int do_cmd_get_node_extra(char *cmdbuf, char *retbuf, int *retlen)
+{
+	struct cluster_node *node;
+	struct cl_node_extra *r_node = (struct cl_node_extra *)retbuf;
+	int nodeid;
+
+	if (!we_are_a_cluster_member)
+		return -ENOENT;
+
+	memcpy(&nodeid, cmdbuf, sizeof(int));
+
+	if (nodeid == CLUSTER_GETNODE_QUORUMDEV) {
+		return -EINVAL;
+	}
+	node = find_node_by_nodeid(nodeid);
+	if (nodeid == 0)
+		node = us;
+	if (!node)
+		return -EINVAL;
+
+	r_node->votes = node->votes;
+	r_node->expected_votes = node->expected_votes;
+	r_node->state = node->state;
+	r_node->leave_reason = node->leave_reason;
+	r_node->nodeid = nodeid;
+
+	*retlen = sizeof(struct cl_node_extra);
+
+	return 0;
+}
+
 static int do_cmd_set_expected(char *cmdbuf, int *retlen)
 {
 	unsigned int total_votes;
@@ -720,7 +786,7 @@ static void send_kill(int nodeid, uint16_t reason)
 {
 	struct cl_killmsg msg;
 
-	P_MEMB("Sending KILL to node %d\n", nodeid);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: Sending KILL to node %d\n", nodeid);
 
 	msg.cmd = CLUSTER_MSG_KILLNODE;
 	msg.reason = reason;
@@ -736,7 +802,7 @@ static void send_leave(uint16_t reason)
 {
 	struct cl_leavemsg msg;
 
-	P_MEMB("Sending LEAVE, reason %d\n", reason);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: Sending LEAVE, reason %d\n", reason);
 
 	msg.cmd = CLUSTER_MSG_LEAVE;
 	msg.reason = reason;
@@ -808,14 +874,14 @@ static int do_cmd_islistening(struct connection *con, char *cmdbuf, int *retlen)
 	   a cluster that has been running for a while
 	*/
        	if (!get_port_bit(rem_node, 0)) {
-		P_MEMB("islistening, no data for node %d, sending PORTENQ\n", nodeid);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: islistening, no data for node %d, sending PORTENQ\n", nodeid);
 		send_port_enquire(rem_node->node_id);
 
 		/* Admit our ignorance */
 		return -EBUSY;
 	}
 	else {
-		P_MEMB("islistening, for node %d, low bytes are %x %x\n", nodeid, rem_node->port_bits[0], rem_node->port_bits[1]);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: islistening, for node %d, low bytes are %x %x\n", nodeid, rem_node->port_bits[0], rem_node->port_bits[1]);
 		return get_port_bit(rem_node, rq.port);
 	}
 }
@@ -837,7 +903,7 @@ static int do_cmd_set_votes(char *cmdbuf, int *retlen)
 	if (!arg.nodeid)
 		arg.nodeid = us->node_id;
 
-	P_MEMB("Setting votes for node %d to %d\n", arg.nodeid, arg.newvotes);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: Setting votes for node %d to %d\n", arg.nodeid, arg.newvotes);
 
 	node = find_node_by_nodeid(arg.nodeid);
 	if (!node)
@@ -868,7 +934,7 @@ static int do_cmd_bind(struct connection *con, char *cmdbuf)
 
 	memcpy(&port, cmdbuf, sizeof(int));
 
-	P_MEMB("requested bind to port %d, (us=%p)\n", port, con);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: requested bind to port %d, (us=%p)\n", port, con);
 
 	if (port == 0 || port > 255)
 		return -EINVAL;
@@ -906,8 +972,10 @@ static int do_cmd_leave_cluster(char *cmdbuf, int *retlen)
 	quit_threads = 1;
 
 	/* No messaging available yet, just die */
-	if (!we_are_a_cluster_member)
+	if (!we_are_a_cluster_member) {
+		cman_finish();
 		exit(0);
+	}
 
 	send_leave(leave_flags);
 	use_count = 0;
@@ -916,7 +984,7 @@ static int do_cmd_leave_cluster(char *cmdbuf, int *retlen)
 	return 0;
 }
 
-static void check_shutdown_status()
+static void check_shutdown_status(void)
 {
 	int reply;
 	int leaveflags = CLUSTER_LEAVEFLAG_DOWN;
@@ -942,14 +1010,14 @@ static void check_shutdown_status()
 			shutdown_con = NULL;
 		}
 
-		P_MEMB("shutdown decision is: %d (yes=%d, no=%d) flags=%x\n", reply, shutdown_yes, shutdown_no, shutdown_flags);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: shutdown decision is: %d (yes=%d, no=%d) flags=%x\n", reply, shutdown_yes, shutdown_no, shutdown_flags);
 	}
 }
 
 /* Not all nodes responded to the shutdown */
 static void shutdown_timer_fn(void *arg)
 {
-	P_MEMB("Shutdown timer fired. flags = %x\n", shutdown_flags);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: Shutdown timer fired. flags = %x\n", shutdown_flags);
 
 	/* Mark undecideds as "NO" */
 	shutdown_no = shutdown_expected;
@@ -965,7 +1033,7 @@ static int do_cmd_shutdown_reply(struct connection *con, char *cmdbuf)
 	if (!shutdown_con)
 		return -EWOULDBLOCK;
 
-	P_MEMB("Shutdown reply is %d\n", response);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: Shutdown reply is %d\n", response);
 
 	/* We only need to keep a track of a client's response in
 	   case it pulls the connection before the shutdown process
@@ -991,7 +1059,7 @@ static int do_cmd_try_shutdown(struct connection *con, char *cmdbuf)
 	int flags = *(int *)cmdbuf;
 
 	/* Are we already in shutdown ? */
-	if (shutdown_con)
+	if (shutdown_con || quit_threads)
 		return -EALREADY;
 
 	shutdown_con = con;
@@ -1002,6 +1070,7 @@ static int do_cmd_try_shutdown(struct connection *con, char *cmdbuf)
 
 	/* If no-one is listening for events then we can just go down now */
 	if (shutdown_expected == 0) {
+		quit_threads = 1;
 		send_leave(CLUSTER_LEAVEFLAG_DOWN);
 		return 0;
 	}
@@ -1081,21 +1150,49 @@ static int do_cmd_unregister_quorum_device(char *cmdbuf, int *retlen)
         return 0;
 }
 
-static void ccsd_timer_fn(void *arg)
+static int reread_config(int new_version)
 {
-	int ccs_err;
+	int read_err;
+	const char *reload_err = NULL;
 
-	log_printf(LOG_DEBUG, "Polling ccsd for updated information\n");
-	ccs_err = read_cman_nodes(corosync, &config_version, 0);
-	if (ccs_err || config_version < wanted_config_version) {
-		log_printf(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
-				wanted_config_version);
+	wanted_config_version = new_version;
 
+	/* Tell objdb to reload */
+	read_err = corosync->object_reload_config(1, &reload_err);
+
+	/* Now get our bits */
+	if (!read_err)
+		read_err = read_cman_nodes(corosync, &config_version, 0);
+
+	if (read_err) {
+		config_error = 1;
+		log_printf(LOG_ERR, "Can't get updated config version %d: %s. Activity suspended on this node\n",
+			   wanted_config_version, reload_err?reload_err:"version mismatch on this node");
+	}
+
+	/* Still too old?? */
+	if (new_version && config_version < wanted_config_version) {
+		log_printf(LOG_ERR, "Can't get updated config version %d, config file is version %d.\n",
+			   wanted_config_version, config_version);
+	}
+
+	/* Keep looking */
+	if (read_err || config_version < wanted_config_version) {
 		corosync->timer_add_duration((unsigned long long)ccsd_poll_interval*1000000, NULL,
-					   ccsd_timer_fn, &ccsd_timer);
+					     ccsd_timer_fn, &ccsd_timer);
 	}
 	else {
-		log_printf(LOG_ERR, "Now got CCS information version %d, continuing\n", config_version);
+		send_transition_msg(0,0);
+	}
+
+	return read_err;
+}
+
+static void ccsd_timer_fn(void *arg)
+{
+	log_printf(LOG_DEBUG, "Polling configuration for updated information\n");
+	if (!reread_config(wanted_config_version) && config_version >= wanted_config_version) {
+		log_printf(LOG_ERR, "Now got config information version %d, continuing\n", config_version);
 		config_error = 0;
 		recalculate_quorum(0, 0);
 		notify_listeners(NULL, EVENT_REASON_CONFIG_UPDATE, config_version);
@@ -1109,7 +1206,7 @@ static void quorum_device_timer_fn(void *arg)
 	if (!quorum_device || quorum_device->state == NODESTATE_DEAD)
 		return;
 
-	P_MEMB("quorum_device_timer_fn\n");
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: quorum_device_timer_fn\n");
 	gettimeofday(&now, NULL);
 	if (quorum_device->last_hello.tv_sec + quorumdev_poll/1000 < now.tv_sec) {
 		quorum_device->state = NODESTATE_DEAD;
@@ -1179,7 +1276,7 @@ static int do_cmd_update_fence_info(char *cmdbuf)
 	strcpy(fence_msg->agent, f->fence_agent);
 	comms_send_message(msg, sizeof(msg), 0,0, 0, 0);
 
-	P_MEMB("node %d fenced by %s\n", f->nodeid, f->fence_agent);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: node %d fenced by %s\n", f->nodeid, f->fence_agent);
 	return 0;
 }
 
@@ -1231,8 +1328,11 @@ static int do_cmd_get_node_addrs(char *cmdbuf, char **retbuf, int retsize, int *
 	memset(outbuf, 0, retsize - offset);
 
 	/* AIS doesn't know about nodes that are not members */
-	if (node->state != NODESTATE_MEMBER)
+	if (node->state != NODESTATE_MEMBER)  {
+		addrs->numaddrs = 0;
+		*retlen = sizeof(struct cl_get_node_addrs);
 		return 0;
+	}
 
 	if (corosync->totem_ifaces_get(nodeid, node_ifs, &status, (unsigned int *)&addrs->numaddrs))
 		return -errno;
@@ -1255,11 +1355,8 @@ int process_command(struct connection *con, int cmd, char *cmdbuf,
 	struct cl_version cnxman_version;
 	char *outbuf = *retbuf;
 	int value;
-#ifdef DEBUG
-	FILE *dumpfile;
-#endif
 
-	P_MEMB("command to process is %x\n", cmd);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: command to process is %x\n", cmd);
 
 	switch (cmd) {
 
@@ -1285,22 +1382,13 @@ int process_command(struct connection *con, int cmd, char *cmdbuf,
 
 	case CMAN_CMD_SET_DEBUGLOG:
 		memcpy(&value, cmdbuf, sizeof(int));
-		set_debuglog(value);
+		/* sanitize input value */
+		if (value > 0)
+			value = 1;
+
+		logsys_config_debug_set(CMAN_NAME, value);
 		err = 0;
 		break;
-#ifdef DEBUG
-	case CMAN_CMD_DUMP_OBJDB:
-		dumpfile = fopen(cmdbuf, "w+");
-		if (dumpfile)  {
-			corosync->object_dump(OBJECT_PARENT_HANDLE, dumpfile);
-			fclose(dumpfile);
-			err = 0;
-		}
-		else {
-			err = -errno;
-		}
-		break;
-#endif
 	case CMAN_CMD_START_CONFCHG:
 		con->confchg = 1;
 		err = 0;
@@ -1345,6 +1433,10 @@ int process_command(struct connection *con, int cmd, char *cmdbuf,
 		err = do_cmd_get_node(cmdbuf, outbuf+offset, retlen);
 		break;
 
+	case CMAN_CMD_GETNODE_EXTRA:
+		err = do_cmd_get_node_extra(cmdbuf, outbuf+offset, retlen);
+		break;
+
 	case CMAN_CMD_GETCLUSTER:
 		err = do_cmd_get_cluster(cmdbuf, outbuf+offset, retlen);
 		break;
@@ -1422,20 +1514,28 @@ int process_command(struct connection *con, int cmd, char *cmdbuf,
 		err = do_cmd_get_node_addrs(cmdbuf, retbuf, retsize, retlen, offset);
 		break;
 	}
-	P_MEMB("command return code is %d\n", err);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: command return code is %d\n", err);
 	return err;
 }
 
 
 int send_to_userport(unsigned char fromport, unsigned char toport,
 		     int nodeid, int tgtid,
-		     char *recv_buf, int len,
+		     const char *recv_buf, int len,
 		     int endian_conv)
 {
 	int ret = -1;
 
 	if (toport == 0) {
-		process_internal_message(recv_buf, nodeid, endian_conv);
+
+	  /* We need to make a private copy here so that the internal command
+	   * processors can do byteswapping.
+	   */
+	        char *newmsg = alloca(len);
+		if (!newmsg)
+		    return -1;
+		memcpy(newmsg, recv_buf, len);
+		process_internal_message(newmsg, nodeid, endian_conv);
 		ret = 0;
 	}
 	else {
@@ -1443,7 +1543,7 @@ int send_to_userport(unsigned char fromport, unsigned char toport,
 		if (port_array[toport]) {
 			struct connection *c = port_array[toport];
 
-			P_MEMB("send_to_userport. cmd=%d,  endian_conv=%d\n", recv_buf[0],endian_conv);
+			log_printf(LOGSYS_LEVEL_DEBUG, "memb: send_to_userport. cmd=%d,  endian_conv=%d\n", recv_buf[0],endian_conv);
 
 			send_data_reply(c, nodeid, fromport, recv_buf, len);
 			ret = 0;
@@ -1452,9 +1552,9 @@ int send_to_userport(unsigned char fromport, unsigned char toport,
 	return ret;
 }
 
-void cman_send_confchg(unsigned int *member_list, int member_list_entries,
-		       unsigned int *left_list, int left_list_entries,
-		       unsigned int *joined_list, int joined_list_entries)
+void cman_send_confchg(const unsigned int *member_list, size_t member_list_entries,
+		       const unsigned int *left_list, size_t left_list_entries,
+		       const unsigned int *joined_list, size_t joined_list_entries)
 {
 	char buf[sizeof(struct sock_confchg_message) +
 		 (member_list_entries+left_list_entries+joined_list_entries) * sizeof(int)];
@@ -1514,7 +1614,7 @@ static int send_port_open_msg(unsigned char port)
 void unbind_con(struct connection *con)
 {
 	if (con->port) {
-		P_MEMB("Unbinding con for port %d\n", con->port);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: Unbinding con for port %d\n", con->port);
 		port_array[con->port] = NULL;
 		send_port_close_msg(con->port);
 		clear_port_bit(us, con->port);
@@ -1552,7 +1652,7 @@ static void post_port_event(int reason, unsigned char port, int nodeid)
 		notify_listeners(con, reason, nodeid);
 }
 
-int our_nodeid()
+int our_nodeid(void)
 {
 	if (us)
 		return us->node_id;
@@ -1584,24 +1684,16 @@ static int valid_transition_msg(int nodeid, struct cl_transmsg *msg)
 
 	/* New config version - try to read new file */
 	if (msg->config_version > config_version) {
-		int ccs_err;
 
-		ccs_err = read_cman_nodes(corosync, &config_version, 0);
-		if (ccs_err || config_version < msg->config_version) {
-			config_error = 1;
-			log_printf(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
-				msg->config_version);
+		if (!reread_config(msg->config_version)) {
 
-			wanted_config_version = msg->config_version;
-			corosync->timer_add_duration((unsigned long long)ccsd_poll_interval*1000000, NULL,
-						   ccsd_timer_fn, &ccsd_timer);
+			if (config_version > msg->config_version) {
+				/* Tell everyone else to update */
+				send_reconfigure(us->node_id, RECONFIG_PARAM_CONFIG_VERSION, config_version);
+			}
+			recalculate_quorum(0, 0);
+			notify_listeners(NULL, EVENT_REASON_CONFIG_UPDATE, config_version);
 		}
-		if (config_version > msg->config_version) {
-			/* Tell everyone else to update */
-			send_reconfigure(us->node_id, RECONFIG_PARAM_CONFIG_VERSION, config_version);
-		}
-		recalculate_quorum(0, 0);
-		notify_listeners(NULL, EVENT_REASON_CONFIG_UPDATE, config_version);
 	}
 
 
@@ -1623,10 +1715,10 @@ void send_transition_msg(int last_memb_count, int first_trans)
 
 	we_are_a_cluster_member = 1;
 
-	P_MEMB("sending TRANSITION message. cluster_name = %s\n", cluster_name);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: sending TRANSITION message. cluster_name = %s\n", cluster_name);
 	msg->cmd = CLUSTER_MSG_TRANSITION;
 	msg->first_trans = first_trans;
-	msg->high_nodeid = get_highest_nodeid();
+	msg->votes = us->votes;
 	msg->expected_votes = us->expected_votes;
 	msg->cluster_id = cluster_id;
 	msg->major_version = CNXMAN_MAJOR_VERSION;
@@ -1676,7 +1768,7 @@ static void byteswap_internal_message(char *data)
 	case CLUSTER_MSG_TRANSITION:
 		transmsg = (struct cl_transmsg *)data;
 		transmsg->cluster_id = swab16(transmsg->cluster_id);
-		transmsg->high_nodeid = swab32(transmsg->high_nodeid);
+		transmsg->votes = swab32(transmsg->votes);
 		transmsg->expected_votes = swab32(transmsg->expected_votes);
 		transmsg->major_version = swab32(transmsg->major_version);
 		transmsg->minor_version = swab32(transmsg->minor_version);
@@ -1747,19 +1839,10 @@ static void do_reconfigure_msg(void *data)
 		break;
 
 	case RECONFIG_PARAM_CONFIG_VERSION:
-		if (config_version != msg->value &&
-		    read_cman_nodes(corosync, &config_version, 0)) {
-			log_printf(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
-				   msg->value);
-
-			config_error = 1;
-			recalculate_quorum(0, 0);
-
-			wanted_config_version = config_version;
-			corosync->timer_add_duration((unsigned long long)ccsd_poll_interval*1000000, NULL,
-						   ccsd_timer_fn, &ccsd_timer);
+		if (config_version != msg->value) {
+			if (!reread_config(msg->value))
+				notify_listeners(NULL, EVENT_REASON_CONFIG_UPDATE, config_version);
 		}
-		notify_listeners(NULL, EVENT_REASON_CONFIG_UPDATE, config_version);
 		break;
 	}
 }
@@ -1769,7 +1852,7 @@ static void do_fence_msg(void *data)
 	struct cl_fencemsg *msg = data;
 	struct cluster_node *node;
 
-	P_DAEMON("got FENCE message, node %d fenced by %s\n", msg->nodeid, msg->agent);
+	log_printf(LOGSYS_LEVEL_DEBUG, "daemon: got FENCE message, node %d fenced by %s\n", msg->nodeid, msg->agent);
 
 	node = find_node_by_nodeid(msg->nodeid);
 	if (!node)
@@ -1792,9 +1875,11 @@ static void do_process_transition(int nodeid, char *data)
 {
 	struct cl_transmsg *msg = (struct cl_transmsg *)data;
 	struct cluster_node *node;
+	unsigned int old_expected;
+	nodestate_t old_state;
 
 	if (valid_transition_msg(nodeid, msg) != 0) {
-		P_MEMB("Transition message from %d does not match current config - should quit ?\n", nodeid);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: Transition message from %d does not match current config - should quit ?\n", nodeid);
 		// Now what ??
 		return;
 	}
@@ -1804,6 +1889,7 @@ static void do_process_transition(int nodeid, char *data)
 	if (msg->flags & NODE_FLAGS_SEESDISALLOWED && !have_disallowed()) {
 		/* Must use syslog directly here or the message will never arrive */
 		syslog(LOG_CRIT, "CMAN: Joined a cluster with disallowed nodes. must die");
+		cman_finish();
 		exit(2);
 	}
 	msg->flags &= ~NODE_FLAGS_SEESDISALLOWED;
@@ -1814,24 +1900,24 @@ static void do_process_transition(int nodeid, char *data)
 		node = find_node_by_nodeid(nodeid);
 	}
 	assert(node);
+	old_expected = node->expected_votes;
+	old_state = node->state;
 
-	P_MEMB("Got TRANSITION message. msg->flags=%x, node->flags=%x, first_trans=%d\n",
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: Got TRANSITION message. msg->flags=%x, node->flags=%x, first_trans=%d\n",
 	       msg->flags, node->flags, msg->first_trans);
 
 	/* Newer nodes 6.1.0 onwards, set the DIRTY flag if they have state. If the new node has been down
 	   and has state then we mark it disallowed because we cannot merge stateful nodes */
 	if ( (msg->flags & NODE_FLAGS_DIRTY && (node->flags & NODE_FLAGS_BEENDOWN)) ||
-	     (msg->flags & NODE_FLAGS_DIRTY && msg->first_trans && !node->us)) {
+	     (msg->flags & NODE_FLAGS_DIRTY && msg->first_trans && !node->us && (us->flags & NODE_FLAGS_DIRTY))) {
 		/* Don't duplicate messages */
 		if (node->state != NODESTATE_AISONLY) {
 			if (cluster_is_quorate) {
-				P_MEMB("Killing node %s because it has rejoined the cluster with existing state", node->name);
 				log_printf(LOG_CRIT, "Killing node %s because it has rejoined the cluster with existing state", node->name);
 				node->state = NODESTATE_AISONLY;
 				send_kill(nodeid, CLUSTER_KILL_REJOIN);
 			}
 			else {
-				P_MEMB("Node %s not joined to cman because it has existing state", node->name);
 				log_printf(LOG_CRIT, "Node %s not joined to cman because it has existing state", node->name);
 				node->state = NODESTATE_AISONLY;
 			}
@@ -1846,13 +1932,11 @@ static void do_process_transition(int nodeid, char *data)
 		/* Don't duplicate messages */
 		if (node->state != NODESTATE_AISONLY) {
 			if (cluster_is_quorate) {
-				P_MEMB("Killing node %s because it has rejoined the cluster without cman_tool join", node->name);
 				log_printf(LOG_CRIT, "Killing node %s because it has rejoined the cluster without cman_tool join", node->name);
 				node->state = NODESTATE_AISONLY;
 				send_kill(nodeid, CLUSTER_KILL_REJOIN);
 			}
 			else {
-				P_MEMB("Node %s not joined to cman because it has rejoined an inquorate cluster", node->name);
 				log_printf(LOG_CRIT, "Node %s not joined to cman because it has rejoined an inquorate cluster", node->name);
 				node->state = NODESTATE_AISONLY;
 			}
@@ -1874,7 +1958,9 @@ static void do_process_transition(int nodeid, char *data)
 
 	/* Take into account any new expected_votes value that the new node has */
 	node->expected_votes = msg->expected_votes;
-	recalculate_quorum(0, 0);
+
+	if (old_state != node->state || old_expected != node->expected_votes)
+		recalculate_quorum(0, 0);
 
 	if (node->fence_agent && msg->fence_agent[0] && strcmp(node->fence_agent, msg->fence_agent))
 	{
@@ -1890,15 +1976,15 @@ static void do_process_transition(int nodeid, char *data)
 	if (node->fence_time && !msg->fence_time &&
 	    node->fence_agent && !msg->fence_agent[0])
 	{
-		char msg[sizeof(struct cl_fencemsg)+strlen(node->fence_agent)+1];
-		struct cl_fencemsg *fence_msg = (struct cl_fencemsg *)msg;
+		char fencemsg[sizeof(struct cl_fencemsg)+strlen(node->fence_agent)+1];
+		struct cl_fencemsg *fence_msg = (struct cl_fencemsg *)fencemsg;
 
 		fence_msg->cmd = CLUSTER_MSG_FENCESTATUS;
 		fence_msg->nodeid = nodeid;
 		fence_msg->timesec = node->fence_time;
 		fence_msg->fenced = 0;
 		strcpy(fence_msg->agent, node->fence_agent);
-		comms_send_message(msg, sizeof(msg), 0,0, nodeid, 0);
+		comms_send_message(fencemsg, sizeof(fencemsg), 0,0, nodeid, 0);
 	}
 }
 
@@ -1912,7 +1998,7 @@ static void process_internal_message(char *data, int nodeid, int need_byteswap)
 	struct cluster_node *node = find_node_by_nodeid(nodeid);
 	unsigned char portresult[PORT_BITS_SIZE+1];
 
-	P_MEMB("Message on port 0 is %d\n", msg->cmd);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: Message on port 0 is %d\n", msg->cmd);
 
 	/* Byteswap messages if needed */
 	if (need_byteswap)
@@ -1936,7 +2022,7 @@ static void process_internal_message(char *data, int nodeid, int need_byteswap)
 	case CLUSTER_MSG_PORTENQ:
 		portresult[0] = CLUSTER_MSG_PORTSTATUS;
 		memcpy(portresult+1, us->port_bits, PORT_BITS_SIZE);
-		P_MEMB("Sending PORTRESULT, low bytes = %x %x\n", us->port_bits[0], us->port_bits[1]);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: Sending PORTRESULT, low bytes = %x %x\n", us->port_bits[0], us->port_bits[1]);
 
 		/* Broadcast reply as other new nodes may be interested */
 		comms_send_message(portresult, PORT_BITS_SIZE+1, 0,0, 0, 0);
@@ -1944,38 +2030,40 @@ static void process_internal_message(char *data, int nodeid, int need_byteswap)
 
 	case CLUSTER_MSG_PORTSTATUS:
 		if (nodeid != us->node_id) {
-			P_MEMB("got PORTRESULT from %d, low bytes = %x %x\n", nodeid, data[1], data[2]);
+			log_printf(LOGSYS_LEVEL_DEBUG, "memb: got PORTRESULT from %d, low bytes = %x %x\n", nodeid, data[1], data[2]);
 			if (node)
 				memcpy(node->port_bits, data+1, PORT_BITS_SIZE);
 		}
 		break;
 
 	case CLUSTER_MSG_TRANSITION:
-		P_MEMB("got TRANSITION from node %d\n", nodeid);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: got TRANSITION from node %d\n", nodeid);
 		do_process_transition(nodeid, data);
 		break;
 
 	case CLUSTER_MSG_KILLNODE:
 		killmsg = (struct cl_killmsg *)data;
-		P_MEMB("got KILL for node %d\n", killmsg->nodeid);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: got KILL for node %d\n", killmsg->nodeid);
 		if (killmsg->nodeid == wanted_nodeid) {
 			/* Must use syslog directly here or the message will never arrive */
 			syslog(LOG_CRIT, "cman killed by node %d because %s\n", nodeid,
 			       killmsg_reason(killmsg->reason));
+			cman_finish();
 			exit(1);
 		}
 		break;
 
 	case CLUSTER_MSG_LEAVE:
 		leavemsg = (struct cl_leavemsg *)data;
-		P_MEMB("got LEAVE from node %d, reason = %d\n", nodeid, leavemsg->reason);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: got LEAVE from node %d, reason = %d\n", nodeid, leavemsg->reason);
 
 		/* We got our own leave message back. now quit */
 		if (node && node->node_id == us->node_id) {
 			/* Tell whomever asked us to leave that we are now going down */
 			if (shutdown_con)
 				send_status_return(shutdown_con, CMAN_CMD_TRY_SHUTDOWN, 0);
-			exit(0);
+			cman_finish();
+			corosync_shutdown();
 		}
 
 		/* Someone else, make a note of the reason for leaving */
@@ -2050,28 +2138,28 @@ void remove_unread_nodes()
 }
 
 /* Add a node from CCS, note that it may already exist if user has simply updated the config file */
-void add_ccs_node(char *nodename, int nodeid, int votes, int expected_votes)
+void add_ccs_node(char *node, int nodeid, int votes, int expected_votes)
 {
 	/* Update node entry */
-	add_new_node(nodename, nodeid, votes, expected_votes, NODESTATE_DEAD);
+	add_new_node(node, nodeid, votes, expected_votes, NODESTATE_DEAD);
 }
 
-void add_ais_node(int nodeid, uint64_t incarnation, int total_members)
+void add_ais_node(int nodeid, uint64_t incar, int total_members)
 {
 	struct cluster_node *node;
 
-	P_MEMB("add_ais_node ID=%d, incarnation = %" PRIu64 "\n",nodeid, incarnation);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: add_ais_node ID=%d, incarnation = %" PRIu64 "\n",nodeid, incar);
 
 	node = find_node_by_nodeid(nodeid);
 	if (!node && total_members == 1) {
 		node = us;
-		P_MEMB("Adding AIS node for 'us'\n");
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: Adding AIS node for 'us'\n");
 	}
 
  	/* This really should exist!! */
 	if (!node) {
 		char tempname[256];
-		log_printf(LOG_ERR, "Got node from AIS id %d with no CCS entry\n", nodeid);
+		log_printf(LOG_ERR, "Got node from AIS id %d with no config entry\n", nodeid);
 
 		/* Emergency nodename */
 		sprintf(tempname, "Node%d", nodeid);
@@ -2080,7 +2168,7 @@ void add_ais_node(int nodeid, uint64_t incarnation, int total_members)
 
 	if (node->state == NODESTATE_DEAD || node->state == NODESTATE_LEAVING) {
 		gettimeofday(&node->join_time, NULL);
-		node->incarnation = incarnation;
+		node->incarnation = incar;
 		node->state = NODESTATE_MEMBER;
 		cluster_members++;
 		recalculate_quorum(0, 0);
@@ -2091,7 +2179,7 @@ void del_ais_node(int nodeid)
 {
 	struct cluster_node *node;
 	time_t t;
-	P_MEMB("del_ais_node %d\n", nodeid);
+	log_printf(LOGSYS_LEVEL_DEBUG, "memb: del_ais_node %d\n", nodeid);
 
 	node = find_node_by_nodeid(nodeid);
 	if (!node)
@@ -2124,9 +2212,10 @@ void del_ais_node(int nodeid)
 
 	case NODESTATE_LEAVING:
 		node->state = NODESTATE_DEAD;
+		memset(&node->port_bits, 0, sizeof(node->port_bits));
 		cluster_members--;
 
-		P_MEMB("del_ais_node %s, leave_reason=%x\n", node->name, node->leave_reason);
+		log_printf(LOGSYS_LEVEL_DEBUG, "memb: del_ais_node %s, leave_reason=%x\n", node->name, node->leave_reason);
 		if ((node->leave_reason & 0xF) == CLUSTER_LEAVEFLAG_REMOVED)
 			recalculate_quorum(1, 1);
 		else
@@ -2186,7 +2275,7 @@ static struct cluster_node *find_node_by_name(char *name)
 	return NULL;
 }
 
-static char *killmsg_reason(int reason)
+static const char *killmsg_reason(int reason)
 {
 	static char msg[1024];
 
