@@ -28,13 +28,19 @@
 
 #define LOG_DAEMON_NAME  "qdiskd"
 #define LOG_MODE_DEFAULT LOG_MODE_OUTPUT_SYSLOG|LOG_MODE_OUTPUT_FILE
+#include "iostate.h"
+
+
+/* from main.c */
+void set_priority(int queue, int prio);
 
 /* from daemon_init.c */
-int daemon_init(char *);
-int check_process_running(char *, pid_t *);
+extern int daemon_init(char *);
+extern void daemon_cleanup(void);
+extern int check_process_running(char *, pid_t *);
 
 /* from proc.c */
-char *state_str(disk_node_state_t s);
+extern const char *state_str(disk_node_state_t s);
 
 /*
   TODO:
@@ -45,14 +51,16 @@ char *state_str(disk_node_state_t s);
  */
 
 /* From bitmap.c */
-int clear_bit(uint8_t *mask, uint32_t bitidx, uint32_t masklen);
-int set_bit(uint8_t *mask, uint32_t bitidx, uint32_t masklen);
-int is_bit_set(uint8_t *mask, uint32_t bitidx, uint32_t masklen);
-inline int get_time(struct timeval *tv, int use_uptime);
-inline void _diff_tv(struct timeval *dest, struct timeval *start,
+extern int clear_bit(uint8_t *mask, uint32_t bitidx, uint32_t masklen);
+extern int set_bit(uint8_t *mask, uint32_t bitidx, uint32_t masklen);
+extern int is_bit_set(uint8_t *mask, uint32_t bitidx, uint32_t masklen);
+
+/* From disk_utils.c */
+extern inline int get_time(struct timeval *tv, int use_uptime);
+extern inline void _diff_tv(struct timeval *dest, struct timeval *start,
 		     struct timeval *end);
 
-static int _running = 1, _reconfig = 0;
+static int _running = 1, _reconfig = 0, _cman_shutdown = 0;
 static int _debug = 0, _foreground = 0;
 
 /* */
@@ -62,6 +70,7 @@ static int _debug = 0, _foreground = 0;
 static void update_local_status(qd_ctx *ctx, node_info_t *ni, int max, int score,
 		    	 int score_req, int score_max);
 static int get_config_data(qd_ctx *ctx, struct h_data *h, int maxh, int *cfh);
+static int cman_wait(cman_handle_t ch, struct timeval *_tv);
 
 
 static void
@@ -148,7 +157,7 @@ check_self(qd_ctx *ctx, status_block_t *sb)
   or has not updated their timestamp recently.  See check_transitions as
   well.
  */
-int
+static int
 read_node_blocks(qd_ctx *ctx, node_info_t *ni, int max)
 {
 	int x, errors = 0;
@@ -281,7 +290,7 @@ check_transitions(qd_ctx *ctx, node_info_t *ni, int max, memb_mask_t mask)
 			ni[x].ni_state = S_EVICT;
 			ni[x].ni_status.ps_state = S_EVICT;
 			ni[x].ni_evil_incarnation = 
-				ni[x].ni_status.ps_incarnation;
+				ni[x].ni_incarnation;
 			
 			/*
 			   Write eviction notice if we're the master.
@@ -316,12 +325,14 @@ check_transitions(qd_ctx *ctx, node_info_t *ni, int max, memb_mask_t mask)
 		 */
 		if (ni[x].ni_evil_incarnation &&
                     (ni[x].ni_evil_incarnation == 
-		     ni[x].ni_status.ps_incarnation)) {
+ 		     ni[x].ni_status.ps_incarnation) &&
+ 		    (ni[x].ni_status.ps_updatenode ==
+ 		     ni[x].ni_status.ps_nodeid)) {
 			logt_print(LOG_CRIT, "Node %d is undead.\n",
 			       ni[x].ni_status.ps_nodeid);
 
 			logt_print(LOG_ALERT,
-			       "Writing eviction notice for node %d\n",
+ 			       "Writing eviction notice (again) for node %d\n",
 			       ni[x].ni_status.ps_nodeid);
 			qd_write_status(ctx, ni[x].ni_status.ps_nodeid,
 					S_EVICT, NULL, NULL, NULL);
@@ -356,6 +367,8 @@ check_transitions(qd_ctx *ctx, node_info_t *ni, int max, memb_mask_t mask)
 			       ni[x].ni_status.ps_nodeid);
 			ni[x].ni_incarnation =
 			    ni[x].ni_status.ps_incarnation;
+			ni[x].ni_evil_incarnation = 0;
+
 			if (mask)
 				set_bit(mask, (ni[x].ni_status.ps_nodeid-1),
 					sizeof(memb_mask_t));
@@ -378,6 +391,23 @@ check_transitions(qd_ctx *ctx, node_info_t *ni, int max, memb_mask_t mask)
 			continue;
 		}
 
+  		/*
+ 		   Case 6:  Check for a node which has changed its
+ 		   incarnation # quickly (e.g. killall -9 qdiskd;
+ 		   qdiskd).  Not a transition.
+ 		 */
+ 		if (state_run(ni[x].ni_state) &&
+ 		    ni[x].ni_incarnation != ni[x].ni_status.ps_incarnation) {
+ 
+ 			logt_print(LOG_DEBUG, "Node %d incarnation # changed\n",
+ 			       ni[x].ni_status.ps_nodeid);
+ 			ni[x].ni_incarnation =
+ 			    ni[x].ni_status.ps_incarnation;
+ 			ni[x].ni_evil_incarnation = 0;
+ 
+ 			continue;
+ 		}
+ 
 		/*
 		   All other cases: Believe the node's reported state
 		 */
@@ -467,6 +497,7 @@ static int
 quorum_init(qd_ctx *ctx, node_info_t *ni, int max, struct h_data *h, int maxh)
 {
 	int x = 0, score, maxscore, score_req = 0;
+	struct timeval tv;
 
 	logt_print(LOG_INFO, "Quorum Daemon Initializing\n");
 	
@@ -517,7 +548,13 @@ quorum_init(qd_ctx *ctx, node_info_t *ni, int max, struct h_data *h, int maxh)
 			score_req = (maxscore/2 + 1);
 		update_local_status(ctx, ni, max, score, score_req, maxscore);
 
-		sleep(ctx->qc_interval);
+		tv.tv_sec = ctx->qc_interval;
+		tv.tv_usec = 0;
+		cman_wait(ctx->qc_cman_user, &tv);
+	}
+
+	if (!_running) {
+		return 1;
 	}
 
 	get_my_score(&score, &maxscore);
@@ -642,13 +679,19 @@ check_votes(qd_ctx *ctx, node_info_t *ni, int max, disk_msg_t *msg)
 static void
 print_node_info(FILE *fp, node_info_t *ni)
 {
+	uint64_t incarnation;
+
 	fprintf(fp, "node_info_t [node %d] {\n", ni->ni_status.ps_nodeid);
+
+	incarnation = be_swap64(ni->ni_incarnation);
 	fprintf(fp, "    ni_incarnation = 0x%08x%08x\n",
-		((int)(ni->ni_incarnation>>32))&0xffffffff,
-		((int)(ni->ni_incarnation)&0xffffffff));
+		((int)(incarnation>>32))&0xffffffff,
+		((int)(incarnation)&0xffffffff));
+
+	incarnation = be_swap64(ni->ni_evil_incarnation);
 	fprintf(fp, "    ni_evil_incarnation = 0x%08x%08x\n",
-		((int)(ni->ni_evil_incarnation>>32))&0xffffffff,
-		((int)(ni->ni_evil_incarnation)&0xffffffff));
+		((int)(incarnation>>32))&0xffffffff,
+		((int)(incarnation)&0xffffffff));
 	fprintf(fp, "    ni_last_seen = %s", ctime(&ni->ni_last_seen));
 	fprintf(fp, "    ni_misses = %d\n", ni->ni_misses);
 	fprintf(fp, "    ni_seen = %d\n", ni->ni_seen);
@@ -785,7 +828,7 @@ set_priority(int queue, int prio)
 {
 	struct sched_param s;
 	int ret;
-	char *func = "nice";
+	const char *func = "nice";
 	
 	if (queue == SCHED_OTHER) {
 		s.sched_priority = 0;
@@ -843,6 +886,7 @@ process_cman_event(cman_handle_t handle, void *private, int reason, int arg)
 		break;
 	case CMAN_REASON_TRY_SHUTDOWN:
 		_running = 0;
+		_cman_shutdown = 1;
 		break;
 	case CMAN_REASON_CONFIG_UPDATE:
 		get_config_data(ctx, NULL, 0, NULL);
@@ -1144,7 +1188,7 @@ quorum_logout(qd_ctx *ctx)
 }
 
 
-void
+static void
 conf_logging(int debug, int logmode, int facility, int loglevel,
 	     int filelevel, char *fname)
 {
@@ -1172,7 +1216,7 @@ conf_logging(int debug, int logmode, int facility, int loglevel,
 }
 
 
-int
+static int
 ccs_read_old_logging(int ccsfd, int *facility, int *priority)
 {
 	char query[256];
@@ -1235,7 +1279,7 @@ get_log_config_data(int ccsfd)
 		need_close = 1;
 	}
 
-	snprintf(fname, sizeof(fname)-1, LOGDIR "/qdisk.log");
+	snprintf(fname, sizeof(fname)-1, LOGDIR "/qdiskd.log");
 	if (ccs_read_old_logging(ccsfd, &facility, &loglevel))
 		filelevel = loglevel;
 
@@ -1611,7 +1655,7 @@ main(int argc, char **argv)
 		}
 	}
 
-	if(getenv("QDISK_DEBUGLOG"))
+	if(getenv("QDISK_DEBUG"))
 		_debug = 1;
 
 	if (!_foreground && daemon_init(argv[0]) < 0) {
@@ -1682,7 +1726,11 @@ main(int argc, char **argv)
 					 ctx.qc_label);
 			check_stop_cman(&ctx);
 			goto out;
-		}
+		} else if (ret > 0) {
+ 			logt_print(LOG_WARNING, "%d matches found for "
+ 					 "label '%s'; please use 'device=' "
+ 					 "instead!\n", ret, ctx.qc_label);
+  		}
 
 		if (ctx.qc_device)
 			free(ctx.qc_device);
@@ -1712,10 +1760,17 @@ main(int argc, char **argv)
 		}
 	}
 
-	if (quorum_init(&ctx, ni, MAX_NODES_DISK, h, cfh) < 0) {
+	ret = quorum_init(&ctx, ni, MAX_NODES_DISK, h, cfh);
+	if (ret < 0) {
 		logt_print(LOG_CRIT, "Initialization failed\n");
 		check_stop_cman(&ctx);
 		goto out;
+	} else if (ret > 0) {
+		/* ret > 0 means we received a shutdown during initialization */
+		/* Write our 'clean down' state to disk and get out of here */
+		logt_print(LOG_INFO, "Shutdown request received during initialization\n");
+		quorum_logout(&ctx);
+		goto out;
 	}
 
 	ret = 0;
@@ -1740,16 +1795,25 @@ main(int argc, char **argv)
 		goto out;
 	}
 	*/
+
+	io_nanny_start(ctx.qc_tko * ctx.qc_interval);
+
 	if (quorum_loop(&ctx, ni, MAX_NODES_DISK) == 0)
 		cman_unregister_quorum_device(ctx.qc_cman_admin);
 
+	io_nanny_stop();
+
 	quorum_logout(&ctx);
-	/* free cman handle to avoid leak in cman */
 out:
+	/* free cman handle to avoid leak in cman */
 	cman_finish(ch_admin);
-	cman_finish(ch_user);
+	if (_cman_shutdown) {
+		cman_replyto_shutdown(ch_user, 1);
+		cman_finish(ch_user);
+	}
 	qd_destroy(&ctx);
 	logt_exit();
+	daemon_cleanup();
 	return ret;
 }
 
