@@ -16,7 +16,7 @@ static int do_read(int fd, void *buf, size_t count)
 	int rv, off = 0;
 
 	while (off < count) {
-		rv = read(fd, buf + off, count - off);
+		rv = read(fd, (char *)buf + off, count - off);
 		if (rv == 0)
 			return -1;
 		if (rv == -1 && errno == EINTR)
@@ -33,7 +33,7 @@ static int do_write(int fd, void *buf, size_t count)
 	int rv, off = 0;
 
  retry:
-	rv = write(fd, buf + off, count);
+	rv = write(fd, (char *)buf + off, count);
 	if (rv == -1 && errno == EINTR)
 		goto retry;
 	if (rv < 0) {
@@ -48,7 +48,7 @@ static int do_write(int fd, void *buf, size_t count)
 	return 0;
 }
 
-static int do_connect(char *sock_path)
+static int do_connect(const char *sock_path)
 {
 	struct sockaddr_un sun;
 	socklen_t addrlen;
@@ -86,7 +86,7 @@ static void init_header(struct dlmc_header *h, int cmd, char *name,
 		strncpy(h->name, name, DLM_LOCKSPACE_LEN);
 }
 
-int do_dump(int cmd, char *name, char *buf)
+static int do_dump(int cmd, char *name, char *buf)
 {
 	struct dlmc_header h, *rh;
 	char *reply;
@@ -376,9 +376,19 @@ int dlmc_fs_result(int fd, char *name, int *type, int *nodeid, int *result)
 		goto out;
 
 	strncpy(name, h.name, DLM_LOCKSPACE_LEN);
-	*type = h.command;
 	*nodeid = h.option;
 	*result = h.data;
+
+	switch (h.command) {
+	case DLMC_CMD_FS_REGISTER:
+		*type = DLMC_RESULT_REGISTER;
+		break;
+	case DLMC_CMD_FS_NOTIFIED:
+		*type = DLMC_RESULT_NOTIFIED;
+		break;
+	default:
+		*type = 0;
+	}
  out:
 	return rv;
 }
