@@ -61,6 +61,7 @@ struct id_info {
 	int fence_external_node;
 	int fence_master;
 	int fence_how;
+	int pad;
 	uint64_t fence_time;
 	uint64_t fence_external_time;
 };
@@ -107,9 +108,11 @@ static void ids_in(struct fd_info *fi, struct id_info *ids)
         }
 }
 
-static char *msg_name(int type)
+static const char *msg_name(int type)
 {
 	switch (type) {
+	case FD_MSG_PROTOCOL:
+		return "protocol";
 	case FD_MSG_START:
 		return "start";
 	case FD_MSG_VICTIM_DONE:
@@ -233,7 +236,7 @@ static struct node_history *get_node_history(struct fd *fd, int nodeid)
 	return NULL;
 }
 
-static void node_history_init(struct fd *fd, int nodeid)
+void node_history_init(struct fd *fd, int nodeid)
 {
 	struct node_history *node;
 
@@ -1077,9 +1080,12 @@ void process_fd_changes(void)
 }
 
 static int add_change(struct fd *fd,
-		      struct cpg_address *member_list, int member_list_entries,
-		      struct cpg_address *left_list, int left_list_entries,
-		      struct cpg_address *joined_list, int joined_list_entries,
+		      const struct cpg_address *member_list,
+		      size_t member_list_entries,
+		      const struct cpg_address *left_list,
+		      size_t left_list_entries,
+		      const struct cpg_address *joined_list,
+		      size_t joined_list_entries,
 		      struct change **cg_out)
 {
 	struct change *cg;
@@ -1193,7 +1199,7 @@ static void add_victims_init(struct fd *fd, struct change *cg)
 	}
 }
 
-static int we_left(struct cpg_address *left_list, int left_list_entries)
+static int we_left(const struct cpg_address *left_list, size_t left_list_entries)
 {
 	int i;
 
@@ -1204,10 +1210,14 @@ static int we_left(struct cpg_address *left_list, int left_list_entries)
 	return 0;
 }
 
-static void confchg_cb(cpg_handle_t handle, struct cpg_name *group_name,
-		       struct cpg_address *member_list, int member_list_entries,
-		       struct cpg_address *left_list, int left_list_entries,
-		       struct cpg_address *joined_list, int joined_list_entries)
+static void confchg_cb(cpg_handle_t handle,
+		       const struct cpg_name *group_name,
+		       const struct cpg_address *member_list,
+		       size_t member_list_entries,
+		       const struct cpg_address *left_list,
+		       size_t left_list_entries,
+		       const struct cpg_address *joined_list,
+		       size_t joined_list_entries)
 {
 	struct fd *fd;
 	struct change *cg;
@@ -1265,8 +1275,10 @@ static void fd_header_in(struct fd_header *hd)
 	hd->msgdata     = le32_to_cpu(hd->msgdata);
 }
 
-static void deliver_cb(cpg_handle_t handle, struct cpg_name *group_name,
-		       uint32_t nodeid, uint32_t pid, void *data, int len)
+static void deliver_cb(cpg_handle_t handle,
+		       const struct cpg_name *group_name,
+		       uint32_t nodeid, uint32_t pid,
+		       void *data, size_t len)
 {
 	struct fd *fd;
 	struct fd_header *hd;
@@ -1278,7 +1290,7 @@ static void deliver_cb(cpg_handle_t handle, struct cpg_name *group_name,
 	}
 
 	if (len < sizeof(*hd)) {
-		log_error("deliver_cb short message %d", len);
+		log_error("deliver_cb short message %zd", len);
 		return;
 	}
 
@@ -1740,13 +1752,15 @@ int set_protocol(void)
    kernel state and they can skip fencing us if we're a victim.  (We have
    to check for that uncontrolled state before calling setup_cpg, obviously.) */
 
-static void deliver_cb_daemon(cpg_handle_t handle, struct cpg_name *group_name,
-		uint32_t nodeid, uint32_t pid, void *data, int len)
+static void deliver_cb_daemon(cpg_handle_t handle,
+			      const struct cpg_name *group_name,
+			      uint32_t nodeid, uint32_t pid,
+			      void *data, size_t len)
 {
 	struct fd_header *hd;
 
 	if (len < sizeof(*hd)) {
-		log_error("deliver_cb short message %d", len);
+		log_error("deliver_cb short message %zd", len);
 		return;
 	}
 
@@ -1762,10 +1776,14 @@ static void deliver_cb_daemon(cpg_handle_t handle, struct cpg_name *group_name,
 	}
 }
 
-static void confchg_cb_daemon(cpg_handle_t handle, struct cpg_name *group_name,
-		struct cpg_address *member_list, int member_list_entries,
-		struct cpg_address *left_list, int left_list_entries,
-		struct cpg_address *joined_list, int joined_list_entries)
+static void confchg_cb_daemon(cpg_handle_t handle,
+			      const struct cpg_name *group_name,
+			      const struct cpg_address *member_list,
+			      size_t member_list_entries,
+			      const struct cpg_address *left_list,
+			      size_t left_list_entries,
+			      const struct cpg_address *joined_list,
+			      size_t joined_list_entries)
 {
 	int i;
 
@@ -1824,7 +1842,7 @@ int setup_cpg(void)
 	error = cpg_initialize(&cpg_handle_daemon, &cpg_callbacks_daemon);
 	if (error != CPG_OK) {
 		log_error("daemon cpg_initialize error %d", error);
-		goto fail;
+		goto ret;
 	}
 
 	cpg_fd_get(cpg_handle_daemon, &daemon_cpg_fd);
@@ -1851,17 +1869,21 @@ int setup_cpg(void)
 
  fail:
 	cpg_finalize(cpg_handle_daemon);
+ ret:
 	return -1;
 }
 
 void close_cpg(void)
 {
+	struct fd *fd;
 	cpg_error_t error;
 	struct cpg_name name;
 	int i = 0;
 
-	if (!cpg_handle_daemon || cluster_down)
+	if (!cpg_handle_daemon)
 		return;
+	if (cluster_down)
+		goto fin;
 
 	memset(&name, 0, sizeof(name));
 	sprintf(name.value, "fenced:daemon");
@@ -1877,6 +1899,12 @@ void close_cpg(void)
 	}
 	if (error != CPG_OK)
 		log_error("daemon cpg_leave error %d", error);
+ fin:
+	list_for_each_entry(fd, &domains, list) {
+		if (fd->cpg_handle)
+			cpg_finalize(fd->cpg_handle);
+	}
+	cpg_finalize(cpg_handle_daemon);
 }
 
 int set_node_info(struct fd *fd, int nodeid, struct fenced_node *nodeinfo)
