@@ -98,11 +98,24 @@ void gfs_sys_fs_del(struct gfs_sbd *sdp)
 	kobject_put(&sdp->sd_kobj);
 }
 
+static int gfs_uevent(struct kset *kset, struct kobject *kobj,
+		       struct kobj_uevent_env *env)
+{
+	struct gfs_sbd *sdp = container_of(kobj, struct gfs_sbd, sd_kobj);
+	add_uevent_var(env, "LOCKTABLE=%s", sdp->sd_table_name);
+	add_uevent_var(env, "LOCKPROTO=%s", sdp->sd_proto_name);
+	return 0;
+}
+
+static struct kset_uevent_ops gfs_uevent_ops = {
+	.uevent = gfs_uevent,
+};
+
 int gfs_sys_init(void)
 {
 	gfs_sys_margs = NULL;
 	spin_lock_init(&gfs_sys_margs_lock);
-	gfs_kset = kset_create_and_add("gfs", NULL, fs_kobj);
+	gfs_kset = kset_create_and_add("gfs", &gfs_uevent_ops, fs_kobj);
 	if (!gfs_kset)
 		return -ENOMEM;
 	return 0;
