@@ -1,15 +1,3 @@
-/*****************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) 2005-2007 Red Hat, Inc.  All rights reserved.
-**
-**  This copyrighted material is made available to anyone wishing to use,
-**  modify, copy, or redistribute it subject to the terms and conditions
-**  of the GNU General Public License v.2.
-**
-*******************************************************************************
-******************************************************************************/
-
 #include "fsck_incore.h"
 #include "fsck.h"
 #include "bio.h"
@@ -20,7 +8,7 @@
 
 #include "metawalk.h"
 
-int check_entries(struct fsck_inode *ip, osi_buf_t *bh, int index,
+static int check_entries(struct fsck_inode *ip, osi_buf_t *bh, int lindex,
 		  int type, int *update, uint16_t *count,
 		  struct metawalk_fxns *pass)
 {
@@ -126,8 +114,8 @@ int check_entries(struct fsck_inode *ip, osi_buf_t *bh, int index,
 /* so that they replace the bad ones.  We have to hack up the old    */
 /* leaf a bit, but it's better than deleting the whole directory,    */
 /* which is what used to happen before.                              */
-void warn_and_patch(struct fsck_inode *ip, uint64_t *leaf_no, 
-		    uint64_t *bad_leaf, uint64_t old_leaf, int index,
+static void warn_and_patch(struct fsck_inode *ip, uint64_t *leaf_no, 
+		    uint64_t *bad_leaf, uint64_t old_leaf, int lindex,
 		    const char *msg)
 {
 	if (*bad_leaf != *leaf_no) {
@@ -137,7 +125,7 @@ void warn_and_patch(struct fsck_inode *ip, uint64_t *leaf_no,
 	}
 	if (*leaf_no == *bad_leaf ||
 	    query(ip->i_sbd, "Attempt to patch around it? (y/n) ")) {
-		put_leaf_nr(ip, index, old_leaf);
+		put_leaf_nr(ip, lindex, old_leaf);
 	}
 	else
 		log_err("Bad leaf left in place.\n");
@@ -146,28 +134,28 @@ void warn_and_patch(struct fsck_inode *ip, uint64_t *leaf_no,
 }
 
 /* Checks exthash directory entries */
-int check_leaf(struct fsck_inode *ip, int *update, struct metawalk_fxns *pass)
+static int check_leaf(struct fsck_inode *ip, int *update, struct metawalk_fxns *pass)
 {
 	int error;
 	struct gfs_leaf leaf, oldleaf;
 	uint64_t leaf_no, old_leaf, bad_leaf = -1;
 	osi_buf_t *lbh;
-	int index;
+	int lindex;
 	struct fsck_sb *sbp = ip->i_sbd;
 	uint16_t count;
 	int ref_count = 0, exp_count = 0;
 
 	old_leaf = 0;
 	memset(&oldleaf, 0, sizeof(oldleaf));
-	for(index = 0; index < (1 << ip->i_di.di_depth); index++) {
-		if(get_leaf_nr(ip, index, &leaf_no)) {
+	for(lindex = 0; lindex < (1 << ip->i_di.di_depth); lindex++) {
+		if(get_leaf_nr(ip, lindex, &leaf_no)) {
 			log_err("Unable to get leaf block number in dir %"
 				PRIu64"\n"
 				"\tDepth = %u\n"
-				"\tindex = %u\n",
+				"\tlindex = %u\n",
 				ip->i_num.no_addr,
 				ip->i_di.di_depth,
-				index);
+				lindex);
 			return -1;
 		}
 
@@ -175,7 +163,7 @@ int check_leaf(struct fsck_inode *ip, int *update, struct metawalk_fxns *pass)
 		 * until those extra pointers are needed, so skip the
 		 * dups */
 		if (leaf_no == bad_leaf) {
-			put_leaf_nr(ip, index, old_leaf); /* fill w/old leaf */
+			put_leaf_nr(ip, lindex, old_leaf); /* fill w/old leaf */
 			ref_count++;
 			continue;
 		}
@@ -220,7 +208,7 @@ int check_leaf(struct fsck_inode *ip, int *update, struct metawalk_fxns *pass)
 					"range for directory #%"PRIu64".\n",
 					leaf_no, ip->i_di.di_num.no_addr);
 				warn_and_patch(ip, &leaf_no, &bad_leaf,
-					       old_leaf, index,
+					       old_leaf, lindex,
 					       "that is out of range");
 				memcpy(&leaf, &oldleaf, sizeof(oldleaf));
 				break;
@@ -232,7 +220,7 @@ int check_leaf(struct fsck_inode *ip, int *update, struct metawalk_fxns *pass)
 					"directory #%"PRIu64".\n",
 					leaf_no, ip->i_di.di_num.no_addr);
 				warn_and_patch(ip, &leaf_no, &bad_leaf,
-					       old_leaf, index,
+					       old_leaf, lindex,
 					       "that cannot be read");
 				memcpy(&leaf, &oldleaf, sizeof(oldleaf));
 				relse_buf(sbp, lbh);
@@ -241,7 +229,7 @@ int check_leaf(struct fsck_inode *ip, int *update, struct metawalk_fxns *pass)
 			/* Make sure it's really a valid leaf block. */
 			if (check_meta(lbh, GFS_METATYPE_LF)) {
 				warn_and_patch(ip, &leaf_no, &bad_leaf,
-					       old_leaf, index,
+					       old_leaf, lindex,
 					       "that is not really a leaf");
 				memcpy(&leaf, &oldleaf, sizeof(oldleaf));
 				relse_buf(sbp, lbh);
@@ -258,7 +246,7 @@ int check_leaf(struct fsck_inode *ip, int *update, struct metawalk_fxns *pass)
 				  ip->i_di.di_depth, leaf.lf_depth);
 			if(pass->check_dentry && 
 			   ip->i_di.di_type == GFS_FILE_DIR) {
-				error = check_entries(ip, lbh, index,
+				error = check_entries(ip, lbh, lindex,
 						      DIR_EXHASH, update,
 						      &count,
 						      pass);
@@ -343,29 +331,45 @@ static int check_eattr_entries(struct fsck_inode *ip, osi_buf_t *bh,
 			stack;
 			return -1;
 		}
-		if(error == 0) {
-			if(pass->check_eattr_extentry && ea_hdr->ea_num_ptrs) {
-				ea_data_ptr = ((uint64_t *)((char *)ea_hdr +
-							    sizeof(struct gfs_ea_header) +
-							    ((ea_hdr->ea_name_len + 7) & ~7)));
+		if(error == 0 && pass->check_eattr_extentry &&
+		   ea_hdr->ea_num_ptrs) {
+			uint32_t tot_ealen = 0;
+			struct fsck_sb *sdp = ip->i_sbd;
 
-				/* It is possible when a EA is shrunk
-				** to have ea_num_ptrs be greater than
-				** the number required for ** data.
-				** In this case, the EA ** code leaves
-				** the blocks ** there for **
-				** reuse...........  */
-				for(i = 0; i < ea_hdr->ea_num_ptrs; i++){
-					if(pass->check_eattr_extentry(ip,
-								      ea_data_ptr,
-								      bh, ea_hdr,
-								      ea_hdr_prev,
-								      pass->private)) {
-						stack;
+			ea_data_ptr = ((uint64_t *)((char *)ea_hdr +
+						    sizeof(struct gfs_ea_header) +
+						    ((ea_hdr->ea_name_len + 7) & ~7)));
+
+			/* It is possible when a EA is shrunk
+			** to have ea_num_ptrs be greater than
+			** the number required for ** data.
+			** In this case, the EA ** code leaves
+			** the blocks ** there for **
+			** reuse...........  */
+			for(i = 0; i < ea_hdr->ea_num_ptrs; i++){
+				if(pass->check_eattr_extentry(ip,
+							      ea_data_ptr,
+							      bh, ea_hdr,
+							      ea_hdr_prev,
+							      pass->private)) {
+					if (query(ip->i_sbd,
+						  "Repair the bad EA? "
+						  "(y/n) ")) {
+						ea_hdr->ea_num_ptrs = i;
+						ea_hdr->ea_data_len =
+							cpu_to_be32(tot_ealen);
+						*ea_data_ptr = 0;
+						/* Endianness doesn't matter
+						   in this case because it's
+						   a single byte. */
+						write_buf(sdp, bh, 0);
 						return -1;
 					}
-					ea_data_ptr++;
+					log_err("The bad EA was not fixed.\n");
 				}
+				tot_ealen += sdp->sb.sb_bsize -
+					sizeof(struct gfs_meta_header);
+				ea_data_ptr++;
 			}
 		}
 		offset += gfs32_to_cpu(ea_hdr->ea_rec_len);
@@ -407,12 +411,13 @@ static int check_leaf_eattr(struct fsck_inode *ip, uint64_t block,
 			relse_buf(ip->i_sbd, bh);
 			return 1;
 		}
+		if (bh) {
+			error = check_eattr_entries(ip, bh, pass);
+			relse_buf(ip->i_sbd, bh);
+		}
+		return error;
 	}
 
-	check_eattr_entries(ip, bh, pass);
-
-	relse_buf(ip->i_sbd, bh);
-
 	return 0;
 }
 
@@ -437,9 +442,13 @@ static int check_indirect_eattr(struct fsck_inode *ip, uint64_t indirect,
 
 	log_debug("Checking EA indirect block #%"PRIu64".\n", indirect);
 
-	if (!pass->check_eattr_indir ||
-	    !pass->check_eattr_indir(ip, indirect, ip->i_di.di_num.no_addr,
-				     &indirect_buf, pass->private)) {
+	if (!pass->check_eattr_indir)
+		return 0;
+	error = pass->check_eattr_indir(ip, indirect, ip->i_di.di_num.no_addr,
+					&indirect_buf, pass->private);
+	if (!error) {
+		int leaf_pointers = 0, leaf_pointer_errors = 0;
+
 		ea_leaf_ptr = (uint64 *)(BH_DATA(indirect_buf)
 					 + sizeof(struct gfs_indirect));
 		end = ea_leaf_ptr
@@ -448,14 +457,29 @@ static int check_indirect_eattr(struct fsck_inode *ip, uint64_t indirect,
 
 		while(*ea_leaf_ptr && (ea_leaf_ptr < end)){
 			block = gfs64_to_cpu(*ea_leaf_ptr);
-			/* FIXME: should I call check_leaf_eattr if we
-			 * find a dup? */
+			leaf_pointers++;
 			error = check_leaf_eattr(ip, block, indirect, pass);
+			if (error)
+				leaf_pointer_errors++;
 			ea_leaf_ptr++;
 		}
+		if (pass->finish_eattr_indir) {
+			int indir_ok = 1;
+
+			if (leaf_pointer_errors == leaf_pointers)
+				indir_ok = 0;
+			pass->finish_eattr_indir(ip, indir_ok, pass->private);
+			if (!indir_ok) {
+				fs_set_bitmap(sdp, indirect, GFS_BLKST_FREE);
+				block_clear(sdp->bl, indirect, indir_blk);
+				block_set(sdp->bl, indirect, block_free);
+				error = 1;
+			}
+		}
 	}
 
-	relse_buf(sdp, indirect_buf);
+	if (indirect_buf)
+		relse_buf(sdp, indirect_buf);
 	return error;
 }
 
@@ -673,7 +697,7 @@ end:
 
 
 /* Checks stuffed inode directories */
-int check_linear_dir(struct fsck_inode *ip, osi_buf_t *bh, int *update,
+static int check_linear_dir(struct fsck_inode *ip, osi_buf_t *bh, int *update,
 		     struct metawalk_fxns *pass)
 {
 	int error = 0;
