@@ -1,16 +1,3 @@
-/******************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004-2005 Red Hat, Inc.  All rights reserved.
-**
-**  This copyrighted material is made available to anyone wishing to use,
-**  modify, copy, or redistribute it subject to the terms and conditions
-**  of the GNU General Public License v.2.
-**
-*******************************************************************************
-******************************************************************************/
-
 #include "fsck_incore.h"
 #include "fsck.h"
 #include "osi_list.h"
@@ -138,6 +125,8 @@ static int find_dentry(struct fsck_inode *ip, struct gfs_dirent *de,
 	osi_list_t *tmp1, *tmp2;
 	struct blocks *b;
 	struct inode_with_dups *id;
+	struct gfs_leaf leaf;
+
 	osi_list_foreach(tmp1, &ip->i_sbd->dup_list) {
 		b = osi_list_entry(tmp1, struct blocks, list);
 		osi_list_foreach(tmp2, &b->ref_inode_list) {
@@ -162,6 +151,10 @@ static int find_dentry(struct fsck_inode *ip, struct gfs_dirent *de,
 			}
 		}
 	}
+	/* Return the number of leaf entries so metawalk doesn't flag this
+	   leaf as having none. */
+	gfs_leaf_in(&leaf, bh->b_data);
+	*count = leaf.lf_entries;
 	return 0;
 }
 
@@ -324,7 +317,7 @@ static int clear_eattr_extentry(struct fsck_inode *ip, uint64_t *ea_data_ptr,
 }
 
 /* Finds all references to duplicate blocks in the metadata */
-int find_block_ref(struct fsck_sb *sbp, uint64_t inode, struct blocks *b)
+static int find_block_ref(struct fsck_sb *sbp, uint64_t inode, struct blocks *b)
 {
 	struct fsck_inode *ip;
 	struct fxn_info myfi = {b->block_no, 0, 1};
@@ -378,7 +371,7 @@ int find_block_ref(struct fsck_sb *sbp, uint64_t inode, struct blocks *b)
 }
 
 /* Finds all blocks marked in the duplicate block bitmap */
-int find_dup_blocks(struct fsck_sb *sbp)
+static int find_dup_blocks(struct fsck_sb *sbp)
 {
 	uint64_t block_no = 0;
 	struct blocks *b;
@@ -403,7 +396,7 @@ int find_dup_blocks(struct fsck_sb *sbp)
 
 
 
-int handle_dup_blk(struct fsck_sb *sbp, struct blocks *b)
+static int handle_dup_blk(struct fsck_sb *sbp, struct blocks *b)
 {
 	osi_list_t *tmp;
 	struct inode_with_dups *id;
@@ -469,9 +462,10 @@ int pass1b(struct fsck_sb *sbp)
 	struct block_query q;
 	osi_list_t *tmp;
 	struct metawalk_fxns find_dirents = {0};
-	find_dirents.check_dentry = &find_dentry;
 	int rc = 0;
 
+	find_dirents.check_dentry = &find_dentry;
+
 	osi_list_init(&sbp->dup_list);
 	/* Shove all blocks marked as duplicated into a list */
 	log_info("Looking for duplicate blocks...\n");
