@@ -14,7 +14,6 @@
 #include <term.h>
 #include <sys/ioctl.h>
 #include <limits.h>
-#include <linux_endian.h>
 #include <sys/time.h>
 #include <linux/gfs2_ondisk.h>
 
@@ -41,323 +40,6 @@ uint64_t gfs1_journal_size = 0; /* in blocks */
 int journals_found = 0;
 
 extern void read_superblock(void);
-uint64_t masterblock(const char *fn);
-
-static __inline__ int fs_is_jdata(struct gfs2_inode *ip)
-{
-        return ip->i_di.di_flags & GFS2_DIF_JDATA;
-}
-
-static struct metapath *find_metapath(struct gfs2_inode *ip, uint64_t block)
-{
-	struct gfs2_sbd *sdp = ip->i_sbd;
-	struct metapath *mp;
-	uint64_t b = block;
-	unsigned int i;
-
-	zalloc(mp, sizeof(struct metapath));
-
-	for (i = ip->i_di.di_height; i--;)
-		mp->mp_list[i] = do_div(b, sdp->sd_inptrs);
-
-	return mp;
-}
-
-static __inline__ uint64_t *
-metapointer(struct gfs2_buffer_head *bh, unsigned int height,
-			struct metapath *mp)
-{
-	unsigned int head_size = (height > 0) ?
-		sizeof(struct gfs2_meta_header) : sizeof(struct gfs2_dinode);
-
-	return ((uint64_t *)(bh->b_data + head_size)) + mp->mp_list[height];
-}
-
-static void lookup_block(struct gfs2_inode *ip,
-	     struct gfs2_buffer_head *bh, unsigned int height, struct metapath *mp,
-	     int create, int *new, uint64_t *block)
-{
-	uint64_t *ptr = metapointer(bh, height, mp);
-
-	if (*ptr) {
-		*block = be64_to_cpu(*ptr);
-		return;
-	}
-
-	*block = 0;
-
-	if (!create)
-		return;
-
-	if (height == ip->i_di.di_height - 1&&
-	    !(S_ISDIR(ip->i_di.di_mode)))
-		*block = data_alloc(ip);
-	else
-		*block = meta_alloc(ip);
-
-	*ptr = cpu_to_be64(*block);
-	ip->i_di.di_blocks++;
-
-	*new = 1;
-}
-
-void gfs1_block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
-		    uint64_t *dblock, uint32_t *extlen, int prealloc)
-{
-	struct gfs2_sbd *sdp = ip->i_sbd;
-	struct gfs2_buffer_head *bh;
-	struct metapath *mp;
-	int create = *new;
-	unsigned int bsize;
-	unsigned int height;
-	unsigned int end_of_metadata;
-	unsigned int x;
-	enum update_flags f;
-
-	f = not_updated;
-	*new = 0;
-	*dblock = 0;
-	if (extlen)
-		*extlen = 0;
-
-	if (!ip->i_di.di_height) { /* stuffed */
-		if (!lblock) {
-			*dblock = ip->i_di.di_num.no_addr;
-			if (extlen)
-				*extlen = 1;
-		}
-		return;
-	}
-
-	bsize = (fs_is_jdata(ip)) ? sdp->sd_jbsize : sdp->bsize;
-
-	height = calc_tree_height(ip, (lblock + 1) * bsize);
-	if (ip->i_di.di_height < height) {
-		if (!create)
-			return;
-
-		build_height(ip, height);
-	}
-
-	mp = find_metapath(ip, lblock);
-	end_of_metadata = ip->i_di.di_height - 1;
-
-	bh = bhold(ip->i_bh);
-
-	for (x = 0; x < end_of_metadata; x++) {
-		lookup_block(ip, bh, x, mp, create, new, dblock);
-		brelse(bh, not_updated);
-		if (!*dblock)
-			goto out;
-
-		if (*new) {
-			struct gfs2_meta_header mh;
-
-			bh = bget(sdp, *dblock);
-			mh.mh_magic = GFS2_MAGIC;
-			mh.mh_type = GFS2_METATYPE_IN;
-			mh.mh_format = GFS2_FORMAT_IN;
-			gfs2_meta_header_out(&mh, bh->b_data);
-			f = updated;
-		} else
-			bh = bread(sdp, *dblock);
-	}
-
-	if (!prealloc)
-		lookup_block(ip, bh, end_of_metadata, mp, create, new, dblock);
-
-	if (extlen && *dblock) {
-		*extlen = 1;
-
-		if (!*new) {
-			uint64_t tmp_dblock;
-			int tmp_new;
-			unsigned int nptrs;
-
-			nptrs = (end_of_metadata) ? sdp->sd_inptrs : sdp->sd_diptrs;
-
-			while (++mp->mp_list[end_of_metadata] < nptrs) {
-				lookup_block(ip, bh, end_of_metadata, mp, FALSE, &tmp_new,
-							 &tmp_dblock);
-
-				if (*dblock + *extlen != tmp_dblock)
-					break;
-
-				(*extlen)++;
-			}
-		}
-	}
-
-	brelse(bh, f);
-
- out:
-	free(mp);
-}
-
-int gfs1_readi(struct gfs2_inode *ip, void *buf,
-	       uint64_t offset, unsigned int size)
-{
-	struct gfs2_sbd *sdp = ip->i_sbd;
-	struct gfs2_buffer_head *bh;
-	uint64_t lblock, dblock;
-	uint32_t extlen = 0;
-	unsigned int amount;
-	int not_new = 0;
-	int journaled = fs_is_jdata(ip);
-	int copied = 0;
-
-	if (offset >= ip->i_di.di_size)
-		return 0;
-
-	if ((offset + size) > ip->i_di.di_size)
-		size = ip->i_di.di_size - offset;
-
-	if (!size)
-		return 0;
-
-	if (journaled) {
-		lblock = offset / sdp->sd_jbsize;
-		offset %= sdp->sd_jbsize;
-	} else {
-		lblock = offset >> sdp->sd_sb.sb_bsize_shift;
-		offset &= sdp->sd_sb.sb_bsize - 1;
-	}
-
-	if (!ip->i_di.di_height) /* stuffed */
-		offset += sizeof(struct gfs2_dinode);
-	else if (journaled)
-		offset += sizeof(struct gfs2_meta_header);
-
-	while (copied < size) {
-		amount = size - copied;
-		if (amount > sdp->bsize - offset)
-			amount = sdp->bsize - offset;
-
-		if (!extlen)
-			gfs1_block_map(ip, lblock, &not_new, &dblock,
-				       &extlen, FALSE);
-
-		if (dblock) {
-			bh = bread(sdp, dblock);
-			dblock++;
-			extlen--;
-		} else
-			bh = NULL;
-
-
-		if (bh) {
-			memcpy(buf+copied, bh->b_data + offset, amount);
-			brelse(bh, not_updated);
-		} else
-			memset(buf+copied, 0, amount);
-		copied += amount;
-		lblock++;
-
-		offset = (journaled) ? sizeof(struct gfs2_meta_header) : 0;
-	}
-
-	return copied;
-}
-
-/**
- * gfs1_rindex_read - read in the rg index file
- *                  Stolen from libgfs2/super.c, but modified to handle gfs1.
- * @sdp: the incore superblock pointer
- * fd: optional file handle for rindex file (if meta_fs file system is mounted)
- *     (if fd is <= zero, it will read from raw device)
- * @count1: return count of the rgs.
- *
- * Returns: 0 on success, -1 on failure
- */
-int gfs1_rindex_read(struct gfs2_sbd *sdp, int fd, int *count1)
-{
-	unsigned int rg;
-	int error;
-	struct gfs2_rindex buf;
-	struct rgrp_list *rgd, *prev_rgd;
-	uint64_t prev_length = 0;
-
-	*count1 = 0;
-	prev_rgd = NULL;
-	for (rg = 0; ; rg++) {
-		if (fd > 0)
-			error = read(fd, &buf, sizeof(struct gfs2_rindex));
-		else
-			error = gfs1_readi(sdp->md.riinode, (char *)&buf,
-					   (rg * sizeof(struct gfs2_rindex)),
-					   sizeof(struct gfs2_rindex));
-		if (!error)
-			break;
-		if (error != sizeof(struct gfs2_rindex))
-			return -1;
-
-		rgd = (struct rgrp_list *)malloc(sizeof(struct rgrp_list));
-		// FIXME: handle failed malloc
-		memset(rgd, 0, sizeof(struct rgrp_list));
-		osi_list_add_prev(&rgd->list, &sdp->rglist);
-
-		gfs2_rindex_in(&rgd->ri, (char *)&buf);
-
-		rgd->start = rgd->ri.ri_addr;
-		if (prev_rgd) {
-			prev_length = rgd->start - prev_rgd->start;
-			prev_rgd->length = prev_length;
-		}
-
-		if(gfs2_compute_bitstructs(sdp, rgd))
-			return -1;
-
-		(*count1)++;
-		prev_rgd = rgd;
-	}
-	if (prev_rgd)
-		prev_rgd->length = prev_length;
-	return 0;
-}
-
-/**
- * gfs1_ri_update - attach rgrps to the super block
- *                  Stolen from libgfs2/super.c, but modified to handle gfs1.
- * @sdp:
- *
- * Given the rgrp index inode, link in all rgrps into the super block
- * and be sure that they can be read.
- *
- * Returns: 0 on success, -1 on failure.
- */
-int gfs1_ri_update(struct gfs2_sbd *sdp, int fd, int *rgcount)
-{
-	struct rgrp_list *rgd;
-	osi_list_t *tmp;
-	int count1 = 0, count2 = 0;
-	uint64_t errblock = 0;
-
-	if (gfs1_rindex_read(sdp, fd, &count1))
-	    goto fail;
-	for (tmp = sdp->rglist.next; tmp != &sdp->rglist; tmp = tmp->next) {
-		enum update_flags f;
-
-		f = not_updated;
-		rgd = osi_list_entry(tmp, struct rgrp_list, list);
-		errblock = gfs2_rgrp_read(sdp, rgd);
-		if (errblock)
-			return errblock;
-		else
-			gfs2_rgrp_relse(rgd, f);
-		count2++;
-	}
-
-	*rgcount = count1;
-	if (count1 != count2)
-		goto fail;
-
-	return 0;
-
- fail:
-	gfs2_rgrp_free(&sdp->rglist, not_updated);
-	return -1;
-}
-
 
 /*
  * get_gfs_struct_info - get block type and structure length
@@ -368,14 +50,14 @@ int gfs1_ri_update(struct gfs2_sbd *sdp, int fd, int *rgcount)
  * returns: 0 if successful
  *          -1 if this isn't gfs metadata.
  */
-int get_gfs_struct_info(char *buf, int *block_type, int *struct_len)
+static int get_gfs_struct_info(char *gbuf, int *block_type, int *gstruct_len)
 {
 	struct gfs2_meta_header mh;
 
 	*block_type = 0;
-	*struct_len = sbd.bsize;
+	*gstruct_len = sbd.bsize;
 
-	gfs2_meta_header_in(&mh, buf);
+	gfs2_meta_header_in(&mh, gbuf);
 	if (mh.mh_magic != GFS2_MAGIC)
 		return -1;
 
@@ -383,40 +65,52 @@ int get_gfs_struct_info(char *buf, int *block_type, int *struct_len)
 
 	switch (mh.mh_type) {
 	case GFS2_METATYPE_SB:   /* 1 (superblock) */
-		*struct_len = sizeof(struct gfs_sb);
+		*gstruct_len = sizeof(struct gfs_sb);
 		break;
 	case GFS2_METATYPE_RG:   /* 2 (rsrc grp hdr) */
-		*struct_len = sbd.bsize; /*sizeof(struct gfs_rgrp);*/
+		*gstruct_len = sbd.bsize; /*sizeof(struct gfs_rgrp);*/
 		break;
 	case GFS2_METATYPE_RB:   /* 3 (rsrc grp bitblk) */
-		*struct_len = sbd.bsize;
+		*gstruct_len = sbd.bsize;
 		break;
 	case GFS2_METATYPE_DI:   /* 4 (disk inode) */
-		*struct_len = sbd.bsize; /*sizeof(struct gfs_dinode);*/
+		*gstruct_len = sbd.bsize; /*sizeof(struct gfs_dinode);*/
 		break;
 	case GFS2_METATYPE_IN:   /* 5 (indir inode blklst) */
-		*struct_len = sbd.bsize; /*sizeof(struct gfs_indirect);*/
+		*gstruct_len = sbd.bsize; /*sizeof(struct gfs_indirect);*/
 		break;
 	case GFS2_METATYPE_LF:   /* 6 (leaf dinode blklst) */
-		*struct_len = sbd.bsize; /*sizeof(struct gfs_leaf);*/
+		*gstruct_len = sbd.bsize; /*sizeof(struct gfs_leaf);*/
 		break;
 	case GFS2_METATYPE_JD:   /* 7 (journal data) */
-		*struct_len = sizeof(struct gfs2_meta_header);
+		/* GFS1 keeps indirect pointers in GFS2_METATYPE_JD blocks
+		   so we need to save the whole block.  For GFS2, we don't
+		   want to, or we might capture user data, which is bad.  */
+		if (gfs1)
+			*gstruct_len = sbd.bsize;
+		else
+			*gstruct_len = sizeof(struct gfs2_meta_header);
 		break;
 	case GFS2_METATYPE_LH:   /* 8 (log header) */
-		*struct_len = sizeof(struct gfs2_log_header);
+		if (gfs1)
+			*gstruct_len = 512; /* gfs copies the log header
+					       twice and compares the copy,
+					       so we need to save all 512
+					       bytes of it. */
+		else
+			*gstruct_len = sizeof(struct gfs2_log_header);
 		break;
 	case GFS2_METATYPE_LD:   /* 9 (log descriptor) */
-		*struct_len = sbd.bsize;
+		*gstruct_len = sbd.bsize;
 		break;
 	case GFS2_METATYPE_EA:   /* 10 (extended attr hdr) */
-		*struct_len = sbd.bsize;
+		*gstruct_len = sbd.bsize;
 		break;
 	case GFS2_METATYPE_ED:   /* 11 (extended attr data) */
-		*struct_len = sbd.bsize;
+		*gstruct_len = sbd.bsize;
 		break;
 	default:
-		*struct_len = sbd.bsize;
+		*gstruct_len = sbd.bsize;
 		break;
 	}
 	return 0;
@@ -428,12 +122,12 @@ int get_gfs_struct_info(char *buf, int *block_type, int *struct_len)
 /* checking every block kills performance.  We only report    */
 /* every second because we don't need 100 extra messages in   */
 /* logs made from verbose mode.                               */
-void warm_fuzzy_stuff(uint64_t block, int force, int save)
+static void warm_fuzzy_stuff(uint64_t wfsblock, int force, int save)
 {
         static struct timeval tv;
         static uint32_t seconds = 0;
         
-	last_reported_block = block;
+	last_reported_block = wfsblock;
 	gettimeofday(&tv, NULL);
 	if (!seconds)
 		seconds = tv.tv_sec;
@@ -446,7 +140,7 @@ void warm_fuzzy_stuff(uint64_t block, int force, int save)
 			if (save) {
 				percent = (block * 100) / last_fs_block;
 				printf("%" PRIu64 " metadata blocks (%"
-				       PRIu64 "%%) processed, ", block,
+				       PRIu64 "%%) processed, ", wfsblock,
 				       percent);
 			}
 			if (total_out < 1024 * 1024)
@@ -466,7 +160,7 @@ void warm_fuzzy_stuff(uint64_t block, int force, int save)
 	}
 }
 
-int block_is_a_journal(void)
+static int block_is_a_journal(void)
 {
 	int j;
 
@@ -476,31 +170,43 @@ int block_is_a_journal(void)
 	return FALSE;
 }
 
-int block_is_systemfile(void)
+static int block_is_systemfile(void)
 {
-	return block_is_jindex() ||
-		block_is_inum_file() ||
-		block_is_statfs_file() ||
-		block_is_quota_file() ||
-		block_is_rindex() ||
-		block_is_a_journal();
+	return block_is_jindex() || block_is_inum_file() ||
+		block_is_statfs_file() || block_is_quota_file() ||
+		block_is_rindex() || block_is_a_journal() ||
+		block_is_per_node() || block_is_in_per_node();
 }
 
-int save_block(int fd, int out_fd, uint64_t blk)
+static int save_block(int fd, int out_fd, uint64_t blk)
 {
 	int blktype, blklen, outsz;
 	uint16_t trailing0;
 	char *p;
 
 	if (blk > last_fs_block) {
-		fprintf(stderr, "\nWarning: bad block pointer ignored in "
-			"block (block %llu (%llx))",
+		fprintf(stderr, "\nWarning: bad block pointer '0x%llx' "
+			"ignored in block (block %llu (%llx))",
+			(unsigned long long)blk,
 			(unsigned long long)block, (unsigned long long)block);
 		return 0;
 	}
 	memset(savedata, 0, sizeof(struct saved_metablock));
-	do_lseek(fd, blk * sbd.bsize);
-	do_read(fd, savedata->buf, sbd.bsize); /* read in the block */
+	if (lseek(fd, blk * sbd.bsize, SEEK_SET) != blk * sbd.bsize) {
+		fprintf(stderr, "bad seek: %s from %s:%d: "
+			"block %lld (0x%llx)\n", strerror(errno),
+			__FUNCTION__, __LINE__,
+			(unsigned long long)blk, (unsigned long long)blk);
+		exit(-1);
+	}
+	/* read in the block */
+	if (read(fd, savedata->buf, sbd.bsize) != sbd.bsize) {
+		fprintf(stderr, "bad read: %s from %s:%d: "
+			"block %lld (0x%llx)\n", strerror(errno),
+			__FUNCTION__, __LINE__,
+			(unsigned long long)blk, (unsigned long long)blk);
+		exit(-1);
+	}
 
 	/* If this isn't metadata and isn't a system file, we don't want it.
 	   Note that we're checking "block" here rather than blk.  That's
@@ -517,25 +223,77 @@ int save_block(int fd, int out_fd, uint64_t blk)
 		p--;
 	}
 	savedata->blk = cpu_to_be64(blk);
-	do_write(out_fd, &savedata->blk, sizeof(savedata->blk));
+	if (write(out_fd, &savedata->blk, sizeof(savedata->blk)) !=
+	    sizeof(savedata->blk)) {
+		fprintf(stderr, "write error: %s from %s:%d: "
+			"block %lld (0x%llx)\n", strerror(errno),
+			__FUNCTION__, __LINE__,
+			(unsigned long long)savedata->blk,
+			(unsigned long long)savedata->blk);
+		exit(-1);
+	}
 	outsz = blklen - trailing0;
 	savedata->siglen = cpu_to_be16(outsz);
-	do_write(out_fd, &savedata->siglen, sizeof(savedata->siglen));
-	do_write(out_fd, savedata->buf, outsz);
+	if (write(out_fd, &savedata->siglen, sizeof(savedata->siglen)) !=
+	    sizeof(savedata->siglen)) {
+		fprintf(stderr, "write error: %s from %s:%d: "
+			"block %lld (0x%llx)\n", strerror(errno),
+			__FUNCTION__, __LINE__,
+			(unsigned long long)savedata->blk,
+			(unsigned long long)savedata->blk);
+		exit(-1);
+	}
+	if (write(out_fd, savedata->buf, outsz) != outsz) {
+		fprintf(stderr, "write error: %s from %s:%d: "
+			"block %lld (0x%llx)\n", strerror(errno),
+			__FUNCTION__, __LINE__,
+			(unsigned long long)savedata->blk,
+			(unsigned long long)savedata->blk);
+		exit(-1);
+	}
 	total_out += sizeof(savedata->blk) + sizeof(savedata->siglen) + outsz;
 	blks_saved++;
 	return blktype;
 }
 
+/*
+ * save_ea_block - save off an extended attribute block
+ */
+static void save_ea_block(int out_fd, struct gfs2_buffer_head *metabh)
+{
+	int i, e, ea_len = sbd.bsize;
+	struct gfs2_ea_header ea;
+
+	for (e = sizeof(struct gfs2_meta_header); e < sbd.bsize; e += ea_len) {
+		uint64_t blk, *b;
+		int charoff;
+
+		gfs2_ea_header_in(&ea, metabh->b_data + e);
+		for (i = 0; i < ea.ea_num_ptrs; i++) {
+			charoff = e + ea.ea_name_len +
+				sizeof(struct gfs2_ea_header) +
+				sizeof(uint64_t) - 1;
+			charoff /= sizeof(uint64_t);
+			b = (uint64_t *)(metabh->b_data);
+			b += charoff + i;
+			blk = be64_to_cpu(*b);
+			save_block(sbd.device_fd, out_fd, blk);
+		}
+		if (!ea.ea_rec_len)
+			break;
+		ea_len = ea.ea_rec_len;
+	}
+}
+
 /*
  * save_indirect_blocks - save all indirect blocks for the given buffer
  */
-void save_indirect_blocks(int out_fd, osi_list_t *cur_list,
+static void save_indirect_blocks(int out_fd, osi_list_t *cur_list,
 			  struct gfs2_buffer_head *mybh, int height, int hgt)
 {
 	uint64_t old_block = 0, indir_block;
 	uint64_t *ptr;
-	int head_size;
+	int head_size, blktype;
 	struct gfs2_buffer_head *nbh;
 
 	head_size = (hgt > 1 ?
@@ -550,9 +308,14 @@ void save_indirect_blocks(int out_fd, osi_list_t *cur_list,
 		if (indir_block == old_block)
 			continue;
 		old_block = indir_block;
-		save_block(sbd.device_fd, out_fd, indir_block);
+		blktype = save_block(sbd.device_fd, out_fd, indir_block);
+		if (blktype == GFS2_METATYPE_EA) {
+			nbh = bread(&sbd.buf_list, indir_block);
+			save_ea_block(out_fd, nbh);
+			brelse(nbh, not_updated);
+		}
 		if (height != hgt) { /* If not at max height */
-			nbh = bread(&sbd, indir_block);
+			nbh = bread(&sbd.buf_list, indir_block);
 			osi_list_add_prev(&nbh->b_altlist,
 					  cur_list);
 			brelse(nbh, not_updated);
@@ -575,7 +338,7 @@ void save_indirect_blocks(int out_fd, osi_list_t *cur_list,
  * For file system journals, the "data" is a mixture of metadata and
  * journaled data.  We want all the metadata and none of the user data.
  */
-void save_inode_data(int out_fd)
+static void save_inode_data(int out_fd)
 {
 	uint32_t height;
 	struct gfs2_inode *inode;
@@ -586,8 +349,11 @@ void save_inode_data(int out_fd)
 
 	for (i = 0; i < GFS2_MAX_META_HEIGHT; i++)
 		osi_list_init(&metalist[i]);
-	metabh = bread(&sbd, block);
-	inode = inode_get(&sbd, metabh);
+	metabh = bread(&sbd.buf_list, block);
+	if (gfs1)
+		inode = inode_get(&sbd, metabh);
+	else
+		inode = gfs_inode_get(&sbd, metabh);
 	height = inode->i_di.di_height;
 	/* If this is a user inode, we don't follow to the file height.
 	   We stop one level less.  That way we save off the indirect
@@ -624,36 +390,27 @@ void save_inode_data(int out_fd)
 		}
 	}
 	if (inode->i_di.di_eattr) { /* if this inode has extended attributes */
-		struct gfs2_ea_header ea;
 		struct gfs2_meta_header mh;
-		int e;
 
-		metabh = bread(&sbd, inode->i_di.di_eattr);
+		metabh = bread(&sbd.buf_list, inode->i_di.di_eattr);
 		save_block(sbd.device_fd, out_fd, inode->i_di.di_eattr);
 		gfs2_meta_header_in(&mh, metabh->b_data);
-		if (mh.mh_magic == GFS2_MAGIC) {
-			for (e = sizeof(struct gfs2_meta_header);
-			     e < sbd.bsize; e += ea.ea_rec_len) {
-				uint64_t blk, *b;
-				int charoff;
-
-				gfs2_ea_header_in(&ea, metabh->b_data + e);
-				for (i = 0; i < ea.ea_num_ptrs; i++) {
-					charoff = e + ea.ea_name_len +
-						sizeof(struct gfs2_ea_header) +
-						sizeof(uint64_t) - 1;
-					charoff /= sizeof(uint64_t);
-					b = (uint64_t *)(metabh->b_data);
-					b += charoff + i;
-					blk = be64_to_cpu(*b);
-					save_block(sbd.device_fd, out_fd, blk);
-				}
-				if (!ea.ea_rec_len)
-					break;
-			}
-		} else {
-			fprintf(stderr, "\nWarning: corrupt extended attribute"
-				" at block %llu (0x%llx) detected.\n",
+		if (mh.mh_magic == GFS2_MAGIC &&
+		    mh.mh_type == GFS2_METATYPE_EA)
+			save_ea_block(out_fd, metabh);
+		else if (mh.mh_magic == GFS2_MAGIC &&
+			 mh.mh_type == GFS2_METATYPE_IN)
+			save_indirect_blocks(out_fd, cur_list, metabh, 2, 2);
+		else {
+			if (mh.mh_magic == GFS2_MAGIC) /* if it's metadata */
+				save_block(sbd.device_fd, out_fd,
+					   inode->i_di.di_eattr);
+			fprintf(stderr,
+				"\nWarning: corrupt extended "
+				"attribute at block %llu (0x%llx) "
+				"detected in inode %lld (0x%llx).\n",
+				(unsigned long long)inode->i_di.di_eattr,
+				(unsigned long long)inode->i_di.di_eattr,
 				(unsigned long long)block,
 				(unsigned long long)block);
 		}
@@ -662,7 +419,7 @@ void save_inode_data(int out_fd)
 	inode_put(inode, not_updated);
 }
 
-void get_journal_inode_blocks(void)
+static void get_journal_inode_blocks(void)
 {
 	int journal;
 	struct gfs2_buffer_head *bh;
@@ -688,12 +445,12 @@ void get_journal_inode_blocks(void)
 			struct gfs_jindex ji;
 			char jbuf[sizeof(struct gfs_jindex)];
 
-			bh = bread(&sbd, sbd1->sb_jindex_di.no_addr);
-			j_inode = inode_get(&sbd, bh);
-			brelse(bh, not_updated);
+			bh = bread(&sbd.buf_list, sbd1->sb_jindex_di.no_addr);
+			j_inode = gfs_inode_get(&sbd, bh);
 			amt = gfs2_readi(j_inode, (void *)&jbuf,
 					 journal * sizeof(struct gfs_jindex),
 					 sizeof(struct gfs_jindex));
+			brelse(bh, not_updated);
 			if (!amt)
 				break;
 			gfs_jindex_in(&ji, jbuf);
@@ -703,7 +460,7 @@ void get_journal_inode_blocks(void)
 			if (journal > indirect->ii[0].dirents - 3)
 				break;
 			jblock = indirect->ii[0].dirent[journal + 2].block;
-			bh = bread(&sbd, jblock);
+			bh = bread(&sbd.buf_list, jblock);
 			j_inode = inode_get(&sbd, bh);
 			gfs2_dinode_in(&jdi, bh->b_data);
 			inode_put(j_inode, not_updated);
@@ -712,6 +469,41 @@ void get_journal_inode_blocks(void)
 	}
 }
 
+static int next_rg_freemeta(struct rgrp_list *rgd, uint64_t *nrfblock, int first)
+{
+	struct gfs2_bitmap *bits = NULL;
+	uint32_t length = rgd->ri.ri_length;
+	uint32_t blk = (first)? 0: (uint32_t)((*nrfblock+1)-rgd->ri.ri_data0);
+	int i;
+
+	if(!first && (*nrfblock < rgd->ri.ri_data0)) {
+		log_err("next_rg_freemeta:  Start block is outside rgrp "
+			"bounds.\n");
+		exit(1);
+	}
+	for(i=0; i < length; i++){
+		bits = &rgd->bits[i];
+		if(blk < bits->bi_len*GFS2_NBBY)
+			break;
+		blk -= bits->bi_len*GFS2_NBBY;
+	}
+	for(; i < length; i++){
+		bits = &rgd->bits[i];
+		blk = gfs2_bitfit((unsigned char *)rgd->bh[i]->b_data +
+				  bits->bi_offset, bits->bi_len, blk,
+				  GFS2_BLKST_UNLINKED);
+		if(blk != BFITNOENT){
+			*nrfblock = blk + (bits->bi_start * GFS2_NBBY) +
+				rgd->ri.ri_data0;
+			break;
+		}
+		blk=0;
+	}
+	if(i == length)
+		return -1;
+	return 0;
+}
+
 void savemeta(char *out_fn, int saveoption)
 {
 	int out_fd;
@@ -742,24 +534,49 @@ void savemeta(char *out_fn, int saveoption)
 	if (!savedata)
 		die("Can't allocate memory for the operation.\n");
 
-	do_lseek(sbd.device_fd, 0);
+	lseek(sbd.device_fd, 0, SEEK_SET);
 	blks_saved = total_out = last_reported_block = 0;
-	sbd.bsize = BUFSIZE;
+	if (!gfs1)
+		sbd.bsize = BUFSIZE;
 	if (!slow) {
-		int i;
-
-		device_geometry(&sbd);
-		fix_device_geometry(&sbd);
+		if (device_geometry(&sbd)) {
+			fprintf(stderr, "Geometery error\n");
+			exit(-1);
+		}
+		if (fix_device_geometry(&sbd)) {
+			fprintf(stderr, "Device is too small (%"PRIu64" bytes)\n",
+				sbd.device.length << GFS2_BASIC_BLOCK_SHIFT);
+			exit(-1);
+		}
 		osi_list_init(&sbd.rglist);
-		osi_list_init(&sbd.buf_list);
-		for(i = 0; i < BUF_HASH_SIZE; i++)
-			osi_list_init(&sbd.buf_hash[i]);
-		sbd.sd_sb.sb_bsize = GFS2_DEFAULT_BSIZE;
-		compute_constants(&sbd);
-		if(!gfs1 && read_sb(&sbd) < 0)
-			slow = TRUE;
-		else
-			sbd.bsize = sbd.bsize = sbd.sd_sb.sb_bsize;
+		init_buf_list(&sbd, &sbd.buf_list, 128 << 20);
+		init_buf_list(&sbd, &sbd.nvbuf_list, 0xffffffff);
+		if (!gfs1)
+			sbd.sd_sb.sb_bsize = GFS2_DEFAULT_BSIZE;
+		if (compute_constants(&sbd)) {
+			fprintf(stderr, "Bad constants (1)\n");
+			exit(-1);
+		}
+		if(gfs1) {
+			sbd.bsize = sbd.sd_sb.sb_bsize;
+			sbd.sd_inptrs = (sbd.bsize -
+					 sizeof(struct gfs_indirect)) /
+				sizeof(uint64_t);
+			sbd.sd_diptrs = (sbd.bsize -
+					  sizeof(struct gfs_dinode)) /
+				sizeof(uint64_t);
+		} else {
+			if (read_sb(&sbd) < 0)
+				slow = TRUE;
+			else {
+				sbd.sd_inptrs = (sbd.bsize -
+					 sizeof(struct gfs2_meta_header)) /
+					sizeof(uint64_t);
+				sbd.sd_diptrs = (sbd.bsize -
+					  sizeof(struct gfs2_dinode)) /
+					sizeof(uint64_t);
+			}
+		}
 	}
 	last_fs_block = lseek(sbd.device_fd, 0, SEEK_END) / sbd.bsize;
 	printf("There are %" PRIu64 " blocks of %u bytes.\n",
@@ -779,7 +596,7 @@ void savemeta(char *out_fn, int saveoption)
 					    &sbd.md.riinode);
 			jindex_block = masterblock("jindex");
 		}
-		bh = bread(&sbd, jindex_block);
+		bh = bread(&sbd.buf_list, jindex_block);
 		gfs2_dinode_in(&di, bh->b_data);
 		if (!gfs1)
 			do_dinode_extended(&di, bh->b_data);
@@ -833,8 +650,10 @@ void savemeta(char *out_fn, int saveoption)
 			slow = gfs2_rgrp_read(&sbd, rgd);
 			if (slow)
 				continue;
-			log_debug("RG at %"PRIu64" is %u long\n",
-				  rgd->ri.ri_addr, rgd->ri.ri_length);
+			log_debug("RG at %lld (0x%llx) is %u long\n",
+				  (unsigned long long)rgd->ri.ri_addr,
+				  (unsigned long long)rgd->ri.ri_addr,
+				  rgd->ri.ri_length);
 			for (i = 0; i < rgd->ri.ri_length; i++) {
 				if(gfs2_block_set(&sbd, blocklist,
 						  rgd->ri.ri_addr + i,
@@ -850,9 +669,9 @@ void savemeta(char *out_fn, int saveoption)
 			}
 			/* Save off the other metadata: inodes, etc. */
 			if (saveoption != 2) {
-				while (!gfs2_next_rg_meta(rgd, &block, first)) {
-					int blktype;
+				int blktype;
 
+				while (!gfs2_next_rg_meta(rgd, &block, first)) {
 					warm_fuzzy_stuff(block, FALSE, TRUE);
 					blktype = save_block(sbd.device_fd,
 							     out_fd, block);
@@ -860,6 +679,14 @@ void savemeta(char *out_fn, int saveoption)
 						save_inode_data(out_fd);
 					first = 0;
 				}
+				/* Save off the free/unlinked meta blocks too.
+				   If we don't, we may run into metadata
+				   allocation issues. */
+				while (!next_rg_freemeta(rgd, &block, first)) {
+					blktype = save_block(sbd.device_fd,
+							     out_fd, block);
+					first = 0;
+				}
 			}
 			gfs2_rgrp_relse(rgd, not_updated);
 		}
@@ -883,15 +710,40 @@ void savemeta(char *out_fn, int saveoption)
 	exit(0);
 }
 
-int restore_data(int fd, int in_fd, int printblocksonly)
+static int restore_data(int fd, int in_fd, int printblocksonly)
 {
 	size_t rs;
 	uint64_t buf64, writes = 0;
 	uint16_t buf16;
-	int first = 1;
+	int first = 1, pos;
+	char rdbuf[256];
+	char gfs_superblock_id[8] = {0x01, 0x16, 0x19, 0x70,
+				     0x00, 0x00, 0x00, 0x01};
 
 	if (!printblocksonly)
-		do_lseek(fd, 0);
+		lseek(fd, 0, SEEK_SET);
+	lseek(in_fd, 0, SEEK_SET);
+	rs = read(in_fd, rdbuf, sizeof(rdbuf));
+	if (rs != sizeof(rdbuf)) {
+		fprintf(stderr, "Error: File is too small.\n");
+		return -1;
+	}
+	for (pos = 0; pos < sizeof(rdbuf) - sizeof(uint64_t) - sizeof(uint16_t);
+	     pos++) {
+		if (!memcmp(&rdbuf[pos + sizeof(uint64_t) + sizeof(uint16_t)],
+			    gfs_superblock_id, sizeof(gfs_superblock_id))) {
+			break;
+		}
+	}
+	if (pos == sizeof(rdbuf) - sizeof(uint64_t) - sizeof(uint16_t))
+		pos = 0;
+	if (lseek(in_fd, pos, SEEK_SET) != pos) {
+		fprintf(stderr, "bad seek: %s from %s:%d: "
+			"offset %lld (0x%llx)\n", strerror(errno),
+			__FUNCTION__, __LINE__, (unsigned long long)pos,
+			(unsigned long long)pos);
+		exit(-1);
+	}
 	blks_saved = total_out = 0;
 	last_fs_block = 0;
 	while (TRUE) {
@@ -917,9 +769,18 @@ int restore_data(int fd, int in_fd, int printblocksonly)
 		}
 		rs = read(in_fd, &buf16, sizeof(uint16_t));
 		savedata->siglen = be16_to_cpu(buf16);
-		if (savedata->siglen > 0 &&
-		    savedata->siglen <= sizeof(savedata->buf)) {
-			do_read(in_fd, savedata->buf, savedata->siglen);
+		if (savedata->siglen <= sizeof(savedata->buf)) {
+			if (savedata->siglen &&
+			    read(in_fd, savedata->buf, savedata->siglen) !=
+			    savedata->siglen) {
+				fprintf(stderr, "read error: %s from %s:%d: "
+					"block %lld (0x%llx)\n",
+					strerror(errno), __FUNCTION__,
+					__LINE__,
+					(unsigned long long)savedata->blk,
+					(unsigned long long)savedata->blk);
+				exit(-1);
+			}
 			if (first) {
 				gfs2_sb_in(&sbd.sd_sb, savedata->buf);
 				sbd1 = (struct gfs_sb *)&sbd.sd_sb;
@@ -955,17 +816,37 @@ int restore_data(int fd, int in_fd, int printblocksonly)
 			} else {
 				warm_fuzzy_stuff(savedata->blk, FALSE, FALSE);
 				if (savedata->blk >= last_fs_block) {
-					printf("Out of space on the destination "
+					printf("\nOut of space on the destination "
 					       "device; quitting.\n");
 					break;
 				}
-				do_lseek(fd, savedata->blk * sbd.bsize);
-				do_write(fd, savedata->buf, sbd.bsize);
+				if (lseek(fd, savedata->blk * sbd.bsize,
+					  SEEK_SET) !=
+				    savedata->blk * sbd.bsize) {
+					fprintf(stderr, "bad seek: %s from %s:"
+						"%d: block %lld (0x%llx)\n",
+						strerror(errno), __FUNCTION__,
+						__LINE__, (unsigned long long)
+						savedata->blk,
+						(unsigned long long)
+						savedata->blk);
+					exit(-1);
+				}
+				if (write(fd, savedata->buf, sbd.bsize) !=
+				    sbd.bsize) {
+					fprintf(stderr, "write error: %s from "
+						"%s:%d: block %lld (0x%llx)\n",
+						strerror(errno),
+						__FUNCTION__, __LINE__,
+						(unsigned long long)savedata->blk,
+						(unsigned long long)savedata->blk);
+					exit(-1);
+				}
 				writes++;
 			}
 			blks_saved++;
 		} else {
-			fprintf(stderr, "Bad record length: %d for #%"
+			fprintf(stderr, "\nBad record length: %d for block #%"
 				PRIu64".\n", savedata->siglen, savedata->blk);
 			return -1;
 		}
@@ -975,7 +856,7 @@ int restore_data(int fd, int in_fd, int printblocksonly)
 	return 0;
 }
 
-void complain(const char *complaint)
+static void complain(const char *complaint)
 {
 	fprintf(stderr, "%s\n", complaint);
 	die("Format is: \ngfs2_edit restoremeta <file to restore> "
