@@ -1,18 +1,17 @@
 #include <errno.h>
 #include <inttypes.h>
-#include <linux_endian.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#include <libintl.h>
+#define _(String) gettext(String)
 
 #include "fsck.h"
 #include "fs_recovery.h"
 #include "libgfs2.h"
 #include "util.h"
 
-#define RANDOM(values) ((values) * (random() / (RAND_MAX + 1.0)))
-
 unsigned int sd_found_jblocks = 0, sd_replayed_jblocks = 0;
 unsigned int sd_found_metablocks = 0, sd_replayed_metablocks = 0;
 unsigned int sd_found_revokes = 0;
@@ -58,7 +57,7 @@ int gfs2_revoke_check(struct gfs2_sbd *sdp, uint64_t blkno, unsigned int where)
 {
 	osi_list_t *tmp;
 	struct gfs2_revoke_replay *rr;
-	int wrap, a, b, revoke;
+	int wrap, a, b;
 	int found = 0;
 
 	osi_list_foreach(tmp, &sd_revoke_list) {
@@ -75,8 +74,7 @@ int gfs2_revoke_check(struct gfs2_sbd *sdp, uint64_t blkno, unsigned int where)
 	wrap = (rr->rr_where < sd_replay_tail);
 	a = (sd_replay_tail < where);
 	b = (where < rr->rr_where);
-	revoke = (wrap) ? (a || b) : (a && b);
-	return revoke;
+	return (wrap) ? (a || b) : (a && b);
 }
 
 void gfs2_revoke_clean(struct gfs2_sbd *sdp)
@@ -121,7 +119,7 @@ static int buf_lo_scan_elements(struct gfs2_inode *ip, unsigned int start,
 		if (error)
 			return error;
 
-		bh_ip = bget(sdp, blkno);
+		bh_ip = bget(&sdp->buf_list, blkno);
 		memcpy(bh_ip->b_data, bh_log->b_data, sdp->bsize);
 
 		check_magic = ((struct gfs2_meta_header *)
@@ -220,7 +218,7 @@ static int databuf_lo_scan_elements(struct gfs2_inode *ip, unsigned int start,
 		if (error)
 			return error;
 
-		bh_ip = bget(sdp, blkno);
+		bh_ip = bget(&sdp->buf_list, blkno);
 		memcpy(bh_ip->b_data, bh_log->b_data, sdp->bsize);
 
 		/* Unescape */
@@ -249,7 +247,7 @@ static int databuf_lo_scan_elements(struct gfs2_inode *ip, unsigned int start,
  * Returns: errno
  */
 
-int foreach_descriptor(struct gfs2_inode *ip, unsigned int start,
+static int foreach_descriptor(struct gfs2_inode *ip, unsigned int start,
 		       unsigned int end, int pass)
 {
 	struct gfs2_buffer_head *bh;
@@ -320,6 +318,59 @@ int foreach_descriptor(struct gfs2_inode *ip, unsigned int start,
 	return 0;
 }
 
+/**
+ * fix_journal_seq_no - Fix log header sequencing problems
+ * @ip: the journal incore inode
+ */
+static int fix_journal_seq_no(struct gfs2_inode *ip)
+{
+	int error = 0, wrapped = 0;
+	uint32_t jd_blocks = ip->i_di.di_size / ip->i_sbd->sd_sb.sb_bsize;
+	uint32_t blk;
+	struct gfs2_log_header lh;
+	uint64_t highest_seq = 0, lowest_seq = 0, prev_seq = 0;
+	int new = 0;
+	uint64_t dblock;
+	uint32_t extlen;
+	struct gfs2_buffer_head *bh;
+
+	for (blk = 0; blk < jd_blocks; blk++) {
+		error = get_log_header(ip, blk, &lh);
+		if (error == 1) /* if not a log header */
+			continue; /* just journal data--ignore it */
+		if (!lowest_seq || lh.lh_sequence < lowest_seq)
+			lowest_seq = lh.lh_sequence;
+		if (!highest_seq || lh.lh_sequence > highest_seq)
+			highest_seq = lh.lh_sequence;
+		if (lh.lh_sequence > prev_seq) {
+			prev_seq = lh.lh_sequence;
+			continue;
+		}
+		/* The sequence number is not higher than the previous one,
+		   so it's either wrap-around or a sequencing problem. */
+		if (!wrapped && lh.lh_sequence == lowest_seq) {
+			wrapped = 1;
+			prev_seq = lh.lh_sequence;
+			continue;
+		}
+		log_err( _("Journal block %u (0x%x): sequence no. 0x%llx "
+			   "out of order.\n"), blk, blk, lh.lh_sequence);
+		log_info( _("Low: 0x%llx, High: 0x%llx, Prev: 0x%llx\n"),
+			  (unsigned long long)lowest_seq,
+			  (unsigned long long)highest_seq,
+			  (unsigned long long)prev_seq);
+		highest_seq++;
+		lh.lh_sequence = highest_seq;
+		prev_seq = lh.lh_sequence;
+		log_warn( _("Renumbering it as 0x%llx\n"), lh.lh_sequence);
+		block_map(ip, blk, &new, &dblock, &extlen, FALSE, not_updated);
+		bh = bread(&ip->i_sbd->buf_list, dblock);
+		gfs2_log_header_out(&lh, bh->b_data);
+		brelse(bh, updated);
+	}
+	return 0;
+}
+
 /**
  * gfs2_recover_journal - recovery a given journal
  * @ip: the journal incore inode
@@ -330,27 +381,48 @@ int foreach_descriptor(struct gfs2_inode *ip, unsigned int start,
  * Returns: errno
  */
 
-int gfs2_recover_journal(struct gfs2_inode *ip, int j)
+static int gfs2_recover_journal(struct gfs2_inode *ip, int j)
 {
 	struct gfs2_sbd *sdp = ip->i_sbd;
 	struct gfs2_log_header head;
 	unsigned int pass;
 	int error;
 
-	log_info("jid=%u: Looking at journal...\n", j);
+	log_info( _("jid=%u: Looking at journal...\n"), j);
 
 	osi_list_init(&sd_revoke_list);
 	error = gfs2_find_jhead(ip, &head);
-	if (error)
-		goto out;
-
+	if (error) {
+		if (!query(&opts, _("\nJournal #%d (\"journal%d\") is "
+				    "corrupt.  Okay to repair it? (y/n)"),
+			   j+1, j)) {
+			log_err( _("jid=%u: The journal was not repaired.\n"),
+				 j);
+			goto out;
+		}
+		log_info( _("jid=%u: Repairing journal...\n"), j);
+		error = fix_journal_seq_no(ip);
+		if (error) {
+			log_err( _("jid=%u: Unable to repair the bad "
+				   "journal.\n"), j);
+			goto out;
+		}
+		error = gfs2_find_jhead(ip, &head);
+		if (error) {
+			log_err( _("jid=%u: Unable to fix the bad journal.\n"),
+				 j);
+			goto out;
+		}
+		log_err( _("jid=%u: The journal was successfully fixed.\n"),
+			 j);
+	}
 	if (head.lh_flags & GFS2_LOG_HEAD_UNMOUNT) {
-		log_info("jid=%u: Journal is clean.\n", j);
+		log_info( _("jid=%u: Journal is clean.\n"), j);
 		return 0;
 	}
-	if (query(&opts, "\nJournal #%d (\"journal%d\") is dirty.  Okay to replay it? (y/n)",
+	if (query(&opts, _("\nJournal #%d (\"journal%d\") is dirty.  Okay to replay it? (y/n)"),
 		    j+1, j)) {
-		log_info("jid=%u: Replaying journal...\n", j);
+		log_info( _("jid=%u: Replaying journal...\n"), j);
 
 		sd_found_jblocks = sd_replayed_jblocks = 0;
 		sd_found_metablocks = sd_replayed_metablocks = 0;
@@ -362,28 +434,28 @@ int gfs2_recover_journal(struct gfs2_inode *ip, int j)
 			if (error)
 				goto out;
 		}
-		log_info("jid=%u: Found %u revoke tags\n", j,
+		log_info( _("jid=%u: Found %u revoke tags\n"), j,
 			 sd_found_revokes);
 		gfs2_revoke_clean(sdp);
 		error = clean_journal(ip, &head);
 		if (error)
 			goto out;
-		log_err("jid=%u: Replayed %u of %u journaled data blocks\n",
+		log_err( _("jid=%u: Replayed %u of %u journaled data blocks\n"),
 			j, sd_replayed_jblocks, sd_found_jblocks);
-		log_err("jid=%u: Replayed %u of %u metadata blocks\n",
+		log_err( _("jid=%u: Replayed %u of %u metadata blocks\n"),
 			j, sd_replayed_metablocks, sd_found_metablocks);
 	} else {
-		if (query(&opts, "Do you want to clear the dirty journal instead? (y/n)")) {
+		if (query(&opts, _("Do you want to clear the dirty journal instead? (y/n)"))) {
 			write_journal(sdp, sdp->md.journal[j], j,
 				      sdp->md.journal[j]->i_di.di_size /
 				      sdp->sd_sb.sb_bsize);
 			
 		} else
-			log_err("jid=%u: Dirty journal not replayed or cleared.\n", j);
+			log_err( _("jid=%u: Dirty journal not replayed or cleared.\n"), j);
 	}
 
 out:
-	log_info("jid=%u: %s\n", j, (error) ? "Failed" : "Done");
+	log_info( _("jid=%u: %s\n"), j, (error) ? "Failed" : "Done");
 	return error;
 }
 
@@ -400,7 +472,7 @@ out:
 int replay_journals(struct gfs2_sbd *sdp){
 	int i;
 
-	log_notice("Recovering journals (this may take a while)");
+	log_notice( _("Recovering journals (this may take a while)"));
 
 	/* Get master dinode */
 	sdp->master_dir = gfs2_load_inode(sdp,
@@ -409,7 +481,7 @@ int replay_journals(struct gfs2_sbd *sdp){
 
 	/* read in the journal index data */
 	if (ji_update(sdp)){
-		log_err("Unable to read in jindex inode.\n");
+		log_err( _("Unable to read in jindex inode.\n"));
 		return -1;
 	}
 
@@ -420,10 +492,11 @@ int replay_journals(struct gfs2_sbd *sdp){
 		inode_put(sdp->md.journal[i],
 			  (opts.no ? not_updated : updated));
 	}
-	log_notice("\nJournal recovery complete.\n");
+	log_notice( _("\nJournal recovery complete.\n"));
 	inode_put(sdp->master_dir, not_updated);
 	inode_put(sdp->md.jiinode, not_updated);
 	/* Sync the buffers to disk so we get a fresh start. */
-	bsync(sdp);
+	bsync(&sdp->buf_list);
+	bsync(&sdp->nvbuf_list);
 	return 0;
 }
