@@ -1,11 +1,12 @@
 #include <inttypes.h>
-#include <linux_endian.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <libintl.h>
+#define _(String) gettext(String)
 
 #include "libgfs2.h"
 #include "fsck.h"
@@ -62,7 +63,11 @@ struct gfs2_inode *fsck_inode_get(struct gfs2_sbd *sdp,
 {
 	struct gfs2_inode *ip, *sysip;
 
-	zalloc(ip, sizeof(struct gfs2_inode));
+	ip = calloc(1, sizeof(struct gfs2_inode));
+	if (ip == NULL) {
+		fprintf(stderr, _("Out of memory in %s\n"), __FUNCTION__);
+		exit(-1);
+	}
 	gfs2_dinode_in(&ip->i_di, bh->b_data);
 	ip->i_bh = bh;
 	ip->i_sbd = sdp;
@@ -91,7 +96,7 @@ void fsck_inode_put(struct gfs2_inode *ip, enum update_flags update)
 	}
 }
 
-int dirent_repair(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
+static int dirent_repair(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 		  struct gfs2_dirent *de, struct gfs2_dirent *dent,
 		  int type, int first)
 {
@@ -130,8 +135,8 @@ int dirent_repair(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 	return 0;
 }
 
-int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
-		  int index, int type, enum update_flags *update,
+static int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
+		  int eindex, int type, enum update_flags *update,
 		  uint16_t *count, struct metawalk_fxns *pass)
 {
 	struct gfs2_leaf *leaf = NULL;
@@ -150,11 +155,11 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 	else if (type == DIR_EXHASH) {
 		dent = (struct gfs2_dirent *)(bh->b_data + sizeof(struct gfs2_leaf));
 		leaf = (struct gfs2_leaf *)bh->b_data;
-		log_debug("Checking leaf %" PRIu64 " (0x%" PRIx64 ")\n",
+		log_debug( _("Checking leaf %" PRIu64 " (0x%" PRIx64 ")\n"),
 				  bh->b_blocknr, bh->b_blocknr);
 	}
 	else {
-		log_err("Invalid directory type %d specified\n", type);
+		log_err( _("Invalid directory type %d specified\n"), type);
 		return -1;
 	}
 
@@ -163,47 +168,57 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 		return 0;
 
 	while(1) {
+		if (skip_this_pass || fsck_abort)
+			return FSCK_OK;
 		memset(&de, 0, sizeof(struct gfs2_dirent));
 		gfs2_dirent_in(&de, (char *)dent);
 		filename = (char *)dent + sizeof(struct gfs2_dirent);
 
 		if (de.de_rec_len < sizeof(struct gfs2_dirent) +
 		    de.de_name_len || !de.de_name_len) {
-			log_err("Directory block %" PRIu64 "(0x%"
-				PRIx64 "), entry %d of directory %"
-				PRIu64 "(0x%" PRIx64 ") is corrupt.\n",
-				bh->b_blocknr, bh->b_blocknr, (*count) + 1,
-				ip->i_di.di_num.no_addr,
-				ip->i_di.di_num.no_addr);
-			if (query(&opts, "Attempt to repair it? (y/n) ")) {
+			log_err( _("Directory block %llu (0x%llx"
+				"), entry %d of directory %llu"
+				"(0x%llx) is corrupt.\n"),
+				(unsigned long long)bh->b_blocknr,
+				(unsigned long long)bh->b_blocknr,
+				(*count) + 1,
+				(unsigned long long)ip->i_di.di_num.no_addr,
+				(unsigned long long)ip->i_di.di_num.no_addr);
+			errors_found++;
+			if (query(&opts, _("Attempt to repair it? (y/n) "))) {
 				if (dirent_repair(ip, bh, &de, dent, type,
 						  first))
 					break;
-				else
+				else {
+					errors_corrected++;
 					*update = updated;
+				}
 			}
 			else {
-				log_err("Corrupt directory entry ignored, "
-					"stopped after checking %d entries.\n",
+				log_err( _("Corrupt directory entry ignored, "
+					"stopped after checking %d entries.\n"),
 					*count);
 				break;
 			}
 		}
 		if (!de.de_inum.no_formal_ino){
 			if(first){
-				log_debug("First dirent is a sentinel (place holder).\n");
+				log_debug( _("First dirent is a sentinel (place holder).\n"));
 				first = 0;
 			} else {
 				/* FIXME: Do something about this */
-				log_err("Directory entry with inode number of zero in leaf %"
-						PRIu64 "(0x%" PRIx64 ") of directory %" PRIu64
-						" (0x%" PRIx64 ")!\n", bh->b_blocknr, bh->b_blocknr,
-						ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
+				log_err( _("Directory entry with inode number of "
+					"zero in leaf %llu (0x%llx) of "
+					"directory %llu (0x%llx)!\n"),
+					(unsigned long long)bh->b_blocknr,
+					(unsigned long long)bh->b_blocknr,
+					(unsigned long long)ip->i_di.di_num.no_addr,
+					(unsigned long long)ip->i_di.di_num.no_addr);
 				return 1;
 			}
 		} else {
 			if (!de.de_inum.no_addr && first) { /* reverse sentinel */
-				log_debug("First dirent is a Sentinel (place holder).\n");
+				log_debug( _("First dirent is a Sentinel (place holder).\n"));
 				/* Swap the two to silently make it a proper sentinel */
 				de.de_inum.no_addr = de.de_inum.no_formal_ino;
 				de.de_inum.no_formal_ino = 0;
@@ -228,7 +243,7 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 		}
 
 		if ((char *)dent + de.de_rec_len >= bh_end){
-			log_debug("Last entry processed.\n");
+			log_debug( _("Last entry processed.\n"));
 			break;
 		}
 
@@ -247,49 +262,80 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 /* so that they replace the bad ones.  We have to hack up the old    */
 /* leaf a bit, but it's better than deleting the whole directory,    */
 /* which is what used to happen before.                              */
-void warn_and_patch(struct gfs2_inode *ip, uint64_t *leaf_no, 
-		    uint64_t *bad_leaf, uint64_t old_leaf, int index,
-		    const char *msg)
+static void warn_and_patch(struct gfs2_inode *ip, uint64_t *leaf_no, 
+			   uint64_t *bad_leaf, uint64_t old_leaf,
+			   uint64_t first_ok_leaf, int pindex, const char *msg)
 {
 	if (*bad_leaf != *leaf_no) {
-		log_err("Directory Inode %" PRIu64 "(0x%"
-			PRIx64 ") points to leaf %" PRIu64 "(0x%"
-			PRIx64 ") %s.\n", ip->i_di.di_num.no_addr,
-			ip->i_di.di_num.no_addr, *leaf_no, *leaf_no, msg);
+		log_err( _("Directory Inode %llu (0x%llx) points to leaf %llu"
+			" (0x%llx) %s.\n"),
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)*leaf_no,
+			(unsigned long long)*leaf_no, msg);
 	}
+	errors_found++;
 	if (*leaf_no == *bad_leaf ||
-	    query(&opts, "Attempt to patch around it? (y/n) ")) {
-		gfs2_put_leaf_nr(ip, index, old_leaf);
+	    query(&opts, _("Attempt to patch around it? (y/n) "))) {
+		errors_corrected++;
+		if (gfs2_check_range(ip->i_sbd, old_leaf) == 0)
+			gfs2_put_leaf_nr(ip, pindex, old_leaf);
+		else
+			gfs2_put_leaf_nr(ip, pindex, first_ok_leaf);
+		log_err( _("Directory Inode %llu (0x%llx) repaired.\n"),
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr);
 	}
 	else
-		log_err("Bad leaf left in place.\n");
+		log_err( _("Bad leaf left in place.\n"));
 	*bad_leaf = *leaf_no;
 	*leaf_no = old_leaf;
 }
 
 /* Checks exhash directory entries */
-int check_leaf(struct gfs2_inode *ip, enum update_flags *update,
+static int check_leaf(struct gfs2_inode *ip, enum update_flags *update,
 	       struct metawalk_fxns *pass)
 {
 	int error;
 	struct gfs2_leaf leaf, oldleaf;
 	uint64_t leaf_no, old_leaf, bad_leaf = -1;
+	uint64_t first_leaf_ptr = -1, first_ok_leaf = -1;
 	struct gfs2_buffer_head *lbh;
-	int index;
+	int lindex;
 	struct gfs2_sbd *sbp = ip->i_sbd;
 	uint16_t count;
 	int ref_count = 0, exp_count = 0;
 
-	old_leaf = 0;
+	/* Find the first valid leaf pointer in range and use it as our "old"
+	   leaf. That way, bad blocks at the beginning will be overwritten
+	   with the first valid leaf. */
+	first_ok_leaf = -1;
+	for(lindex = 0; lindex < (1 << ip->i_di.di_depth); lindex++) {
+		gfs2_get_leaf_nr(ip, lindex, &first_ok_leaf);
+		if (first_leaf_ptr == -1)
+			first_leaf_ptr = first_ok_leaf;
+		if(gfs2_check_range(ip->i_sbd, first_ok_leaf) == 0) {
+			lbh = bread(&sbp->buf_list, first_ok_leaf);
+			/* Make sure it's really a valid leaf block. */
+			if (gfs2_check_meta(lbh, GFS2_METATYPE_LF) == 0) {
+				brelse(lbh, not_updated);
+				break;
+			}
+			brelse(lbh, not_updated);
+		}
+	}
+	old_leaf = -1;
 	memset(&oldleaf, 0, sizeof(oldleaf));
-	for(index = 0; index < (1 << ip->i_di.di_depth); index++) {
-		gfs2_get_leaf_nr(ip, index, &leaf_no);
+	for(lindex = 0; lindex < (1 << ip->i_di.di_depth); lindex++) {
+		if (fsck_abort)
+			break;
+		gfs2_get_leaf_nr(ip, lindex, &leaf_no);
 
 		/* GFS has multiple indirect pointers to the same leaf
 		 * until those extra pointers are needed, so skip the
 		 * dups */
 		if (leaf_no == bad_leaf) {
-			gfs2_put_leaf_nr(ip, index, old_leaf); /* fill w/old
+			gfs2_put_leaf_nr(ip, lindex, old_leaf); /* fill w/old
 								  leaf info */
 			ref_count++;
 			continue;
@@ -297,50 +343,56 @@ int check_leaf(struct gfs2_inode *ip, enum update_flags *update,
 		else if(old_leaf == leaf_no) {
 			ref_count++;
 			continue;
-		} else {
-			if(ref_count != exp_count){
-				log_err("Dir #%" PRIu64 " (0x%"
-					PRIx64 ") has an incorrect "
-					"number of pointers to leaf #%"
-					PRIu64 " (0x%" PRIx64
-					")\n\tFound: %u,  Expected: %u\n",
-					ip->i_di.di_num.no_addr,
-					ip->i_di.di_num.no_addr,
-					old_leaf, old_leaf, ref_count,
-					exp_count);
-				if (query(&opts, "Attempt to fix it? (y/n) "))
-				{
-					int factor = 0, divisor = ref_count;
-
-					lbh = bread(&sbp->buf_list, old_leaf);
-					while (divisor > 1) {
-						factor++;
-						divisor /= 2;
-					}
-					oldleaf.lf_depth = ip->i_di.di_depth -
-						factor;
-					gfs2_leaf_out(&oldleaf, lbh->b_data);
-					brelse(lbh, updated);
-				}
-				else
-					return 1;
-			}
-			ref_count = 1;
 		}
+		if(gfs2_check_range(ip->i_sbd, old_leaf) == 0 &&
+		   ref_count != exp_count){
+			log_err( _("Dir #%llu (0x%llx) has an incorrect "
+				   "number of pointers to leaf #%llu "
+				   " (0x%llx)\n\tFound: %u,  Expected: "
+				   "%u\n"), (unsigned long long)
+				 ip->i_di.di_num.no_addr,
+				 (unsigned long long)
+				 ip->i_di.di_num.no_addr,
+				 (unsigned long long)old_leaf,
+				 (unsigned long long)old_leaf,
+				 ref_count, exp_count);
+			errors_found++;
+			if (query(&opts, _("Attempt to fix it? (y/n) "))) {
+				int factor = 0, divisor = ref_count;
+
+				errors_corrected++;
+				lbh = bread(&sbp->buf_list, old_leaf);
+				while (divisor > 1) {
+					factor++;
+					divisor /= 2;
+				}
+				gfs2_leaf_in(&oldleaf, lbh->b_data);
+				oldleaf.lf_depth = ip->i_di.di_depth - factor;
+				gfs2_leaf_out(&oldleaf, lbh->b_data);
+				brelse(lbh, updated);
+			}
+			else
+				return 1;
+		}
+		ref_count = 1;
 
 		count = 0;
 		do {
+			if (fsck_abort)
+				break;
 			/* Make sure the block number is in range. */
 			if(gfs2_check_range(ip->i_sbd, leaf_no)){
-				log_err("Leaf block #%" PRIu64 " (0x%"
-					PRIx64 ") is out of range for "
-					"directory #%" PRIu64 " (0x%"
-					PRIx64 ").\n", leaf_no, leaf_no,
+				log_err( _("Leaf block #%llu (0x%llx) is out "
+					"of range for directory #%llu (0x%llx"
+					").\n"), (unsigned long long)leaf_no,
+					(unsigned long long)leaf_no,
+					(unsigned long long)
 					ip->i_di.di_num.no_addr,
+					(unsigned long long)
 					ip->i_di.di_num.no_addr);
 				warn_and_patch(ip, &leaf_no, &bad_leaf,
-					       old_leaf, index,
-					       "that is out of range");
+					       old_leaf, first_ok_leaf, lindex,
+					       _("that is out of range"));
 				memcpy(&leaf, &oldleaf, sizeof(oldleaf));
 				break;
 			}
@@ -351,10 +403,10 @@ int check_leaf(struct gfs2_inode *ip, enum update_flags *update,
 			/* Make sure it's really a valid leaf block. */
 			if (gfs2_check_meta(lbh, GFS2_METATYPE_LF)) {
 				warn_and_patch(ip, &leaf_no, &bad_leaf,
-					       old_leaf, index,
-					       "that is not really a leaf");
+					       old_leaf, first_ok_leaf, lindex,
+					       _("that is not really a leaf"));
 				memcpy(&leaf, &oldleaf, sizeof(oldleaf));
-				brelse(lbh, (opts.no ? not_updated : updated));
+				brelse(lbh, not_updated);
 				break;
 			}
 			gfs2_leaf_in(&leaf, lbh->b_data);
@@ -371,29 +423,34 @@ int check_leaf(struct gfs2_inode *ip, enum update_flags *update,
 			 * values and replace them with the correct value. */
 
 			if (leaf.lf_dirent_format == (GFS2_FORMAT_DE << 16)) {
-				log_debug("incorrect lf_dirent_format at leaf #%" PRIu64 "\n", leaf_no);
+				log_debug( _("incorrect lf_dirent_format at leaf #%" PRIu64 "\n"), leaf_no);
 				leaf.lf_dirent_format = GFS2_FORMAT_DE;
 				gfs2_leaf_out(&leaf, lbh->b_data);
-				log_debug("Fixing lf_dirent_format.\n");
+				log_debug( _("Fixing lf_dirent_format.\n"));
 				*update = (opts.no ? not_updated : updated);
 			}
 
 			/* Make sure it's really a leaf. */
 			if (leaf.lf_header.mh_type != GFS2_METATYPE_LF) {
-				log_err("Inode %" PRIu64 " (0x%"
-					PRIx64 ") points to bad leaf "
-					PRIu64 " (0x%" PRIx64 ").\n",
-					ip->i_di.di_num.no_addr, leaf_no);
+				log_err( _("Inode %llu (0x%llx"
+					") points to bad leaf %llu"
+					" (0x%llx).\n"),
+					(unsigned long long)
+					ip->i_di.di_num.no_addr,
+					(unsigned long long)
+					ip->i_di.di_num.no_addr,
+					(unsigned long long)leaf_no,
+					(unsigned long long)leaf_no);
 				brelse(lbh, *update);
 				break;
 			}
 			exp_count = (1 << (ip->i_di.di_depth - leaf.lf_depth));
-			log_debug("expected count %u - di_depth %u, leaf depth %u\n",
+			log_debug( _("expected count %u - di_depth %u, leaf depth %u\n"),
 					  exp_count, ip->i_di.di_depth, leaf.lf_depth);
 
 			if(pass->check_dentry &&
 			   S_ISDIR(ip->i_di.di_mode)) {
-				error = check_entries(ip, lbh, index,
+				error = check_entries(ip, lbh, lindex,
 						      DIR_EXHASH, update,
 						      &count, pass);
 
@@ -417,18 +474,27 @@ int check_leaf(struct gfs2_inode *ip, enum update_flags *update,
 					lbh = bread(&sbp->buf_list, leaf_no);
 					gfs2_leaf_in(&leaf, lbh->b_data);
 
-					log_err("Leaf %"PRIu64" (0x%" PRIx64
-							") entry count in directory %" PRIu64
-							" doesn't match number of entries found - is %u, found %u\n",
-							leaf_no, leaf_no, ip->i_di.di_num.no_addr,
-							leaf.lf_entries, count);
-					if(query(&opts, "Update leaf entry count? (y/n) ")) {
+					log_err( _("Leaf %llu (0x%llx) entry "
+						"count in directory %llu"
+						" (0x%llx) doesn't match "
+						"number of entries found "
+						"- is %u, found %u\n"),
+						(unsigned long long)leaf_no,
+						(unsigned long long)leaf_no,
+						(unsigned long long)
+						ip->i_di.di_num.no_addr,
+						(unsigned long long)
+						ip->i_di.di_num.no_addr,
+						leaf.lf_entries, count);
+					errors_found++;
+					if(query(&opts, _("Update leaf entry count? (y/n) "))) {
+						errors_corrected++;
 						leaf.lf_entries = count;
 						gfs2_leaf_out(&leaf, lbh->b_data);
-						log_warn("Leaf entry count updated\n");
+						log_warn( _("Leaf entry count updated\n"));
 						f = updated;
 					} else
-						log_err("Leaf entry count left in inconsistant state\n");
+						log_err( _("Leaf entry count left in inconsistant state\n"));
 					brelse(lbh, f);
 				}
 				/* FIXME: Need to get entry count and
@@ -439,7 +505,7 @@ int check_leaf(struct gfs2_inode *ip, enum update_flags *update,
 				if(!leaf.lf_next)
 					break;
 				leaf_no = leaf.lf_next;
-				log_debug("Leaf chain detected.\n");
+				log_debug( _("Leaf chain detected.\n"));
 			}
 		} while(1);
 		old_leaf = leaf_no;
@@ -497,8 +563,12 @@ static int check_eattr_entries(struct gfs2_inode *ip,
 							      ea_hdr_prev,
 							      update_it,
 							      pass->private)) {
-					if (query(&opts, "Repair the bad EA? "
-						  "(y/n) ")) {
+					errors_found++;
+					error = 1;
+					if (query(&opts, _("Repair the bad "
+							 "Extended Attribute? "
+							   "(y/n) "))) {
+						errors_corrected++;
 						ea_hdr->ea_num_ptrs = i;
 						ea_hdr->ea_data_len =
 							cpu_to_be32(tot_ealen);
@@ -507,9 +577,14 @@ static int check_eattr_entries(struct gfs2_inode *ip,
 						/* Endianness doesn't matter
 						   in this case because it's
 						   a single byte. */
-						return -1;
-					}
-					log_err("The bad EA was not fixed.\n");
+						gfs2_block_set(sdp, bl,
+							     ip->i_di.di_eattr,
+							     gfs2_meta_eattr);
+						log_err( _("The EA was "
+							   "fixed.\n"));
+					} else
+						log_err( _("The bad EA was "
+							   "not fixed.\n"));
 				}
 				tot_ealen += sdp->sd_sb.sb_bsize -
 					sizeof(struct gfs2_meta_header);
@@ -527,7 +602,7 @@ static int check_eattr_entries(struct gfs2_inode *ip,
 			 be32_to_cpu(ea_hdr->ea_rec_len));
 	}
 
-	return 0;
+	return error;
 }
 
 /**
@@ -535,7 +610,7 @@ static int check_eattr_entries(struct gfs2_inode *ip,
  * @ip: the inode the eattr comes from
  * @block: block number of the leaf
  *
- * Returns: 0 on success, -1 if removal is needed
+ * Returns: 0 on success, 1 if removal is needed, -1 on error
  */
 static int check_leaf_eattr(struct gfs2_inode *ip, uint64_t block,
 			    uint64_t parent, enum update_flags *want_updated,
@@ -544,7 +619,7 @@ static int check_leaf_eattr(struct gfs2_inode *ip, uint64_t block,
 	struct gfs2_buffer_head *bh = NULL;
 	int error = 0;
 
-	log_debug("Checking EA leaf block #%"PRIu64" (0x%" PRIx64 ").\n",
+	log_debug( _("Checking EA leaf block #%"PRIu64" (0x%" PRIx64 ").\n"),
 			  block, block);
 
 	if(pass->check_eattr_leaf) {
@@ -557,9 +632,10 @@ static int check_leaf_eattr(struct gfs2_inode *ip, uint64_t block,
 		if(error > 0) {
 			return 1;
 		}
-		check_eattr_entries(ip, bh, pass, want_updated);
+		error = check_eattr_entries(ip, bh, pass, want_updated);
 		if (bh)
 			brelse(bh, *want_updated);
+		return error;
 	}
 
 	return 0;
@@ -582,7 +658,7 @@ static int check_indirect_eattr(struct gfs2_inode *ip, uint64_t indirect,
 	struct gfs2_sbd *sdp = ip->i_sbd;
 
 	*want_updated = not_updated;
-	log_debug("Checking EA indirect block #%"PRIu64" (0x%" PRIx64 ").\n",
+	log_debug( _("Checking EA indirect block #%"PRIu64" (0x%" PRIx64 ").\n"),
 			  indirect, indirect);
 
 	if (!pass->check_eattr_indir)
@@ -618,8 +694,6 @@ static int check_indirect_eattr(struct gfs2_inode *ip, uint64_t indirect,
 				if (*want_updated)
 					gfs2_set_bitmap(sdp, indirect,
 							GFS2_BLKST_FREE);
-				gfs2_block_clear(sdp, bl, indirect,
-						 gfs2_indir_blk);
 				gfs2_block_set(sdp, bl, indirect,
 					       gfs2_block_free);
 				error = 1;
@@ -647,8 +721,9 @@ int check_inode_eattr(struct gfs2_inode *ip, enum update_flags *want_updated,
 		return 0;
 	}
 
-	log_debug("Extended attributes exist for inode #%" PRIu64 " (0x%" PRIx64
-		  ").\n", ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
+	log_debug( _("Extended attributes exist for inode #%llu (0x%llx).\n"),
+		  (unsigned long long)ip->i_di.di_num.no_addr,
+		  (unsigned long long)ip->i_di.di_num.no_addr);
 
 	if(ip->i_di.di_flags & GFS2_DIF_EA_INDIRECT){
 		if((error = check_indirect_eattr(ip, ip->i_di.di_eattr,
@@ -696,9 +771,17 @@ static int build_and_check_metalist(struct gfs2_inode *ip,
 			bh = osi_list_entry(tmp, struct gfs2_buffer_head,
 					    b_altlist);
 
-			head_size = (i > 1 ?
-				     sizeof(struct gfs2_meta_header) :
-				     sizeof(struct gfs2_dinode));
+			if (i > 1) {
+				/* if this isn't really a block list skip it */
+				if (gfs2_check_meta(bh, GFS2_METATYPE_IN))
+					continue;
+				head_size = sizeof(struct gfs2_meta_header);
+			} else {
+				/* if this isn't really a dinode, skip it */
+				if (gfs2_check_meta(bh, GFS2_METATYPE_DI))
+					continue;
+				head_size = sizeof(struct gfs2_dinode);
+			}
 
 			for (ptr = (uint64_t *)(bh->b_data + head_size);
 			     (char *)ptr < (bh->b_data + ip->i_sbd->bsize);
@@ -718,8 +801,8 @@ static int build_and_check_metalist(struct gfs2_inode *ip,
 					goto fail;
 				}
 				if(err > 0) {
-					log_debug("Skipping block %" PRIu64
-						  " (0x%" PRIx64 ")\n",
+					log_debug( _("Skipping block %" PRIu64
+						  " (0x%" PRIx64 ")\n"),
 						  block, block);
 					continue;
 				}
@@ -779,8 +862,9 @@ int check_metatree(struct gfs2_inode *ip, struct metawalk_fxns *pass)
 	/* We don't need to record directory blocks - they will be
 	 * recorded later...i think... */
         if (S_ISDIR(ip->i_di.di_mode))
-		log_debug("Directory with height > 0 at %"PRIu64"\n",
-			  ip->i_di.di_num.no_addr);
+		log_debug( _("Directory with height > 0 at %llu (0x%llx)\n"),
+			  (unsigned long long)ip->i_di.di_num.no_addr,
+			  (unsigned long long)ip->i_di.di_num.no_addr);
 
 	/* check data blocks */
 	list = &metalist[height - 1];
@@ -788,8 +872,17 @@ int check_metatree(struct gfs2_inode *ip, struct metawalk_fxns *pass)
 	for (tmp = list->next; tmp != list; tmp = tmp->next) {
 		bh = osi_list_entry(tmp, struct gfs2_buffer_head, b_altlist);
 
-		head_size = (height != 1 ? sizeof(struct gfs2_meta_header) :
-			     sizeof(struct gfs2_dinode));
+		if (height > 1) {
+			/* if this isn't really a block list skip it */
+			if (gfs2_check_meta(bh, GFS2_METATYPE_IN))
+				continue;
+			head_size = sizeof(struct gfs2_meta_header);
+		} else {
+			/* if this isn't really a dinode, skip it */
+			if (gfs2_check_meta(bh, GFS2_METATYPE_DI))
+				continue;
+			head_size = sizeof(struct gfs2_dinode);
+		}
 		ptr = (uint64_t *)(bh->b_data + head_size);
 
 		for ( ; (char *)ptr < (bh->b_data + ip->i_sbd->bsize); ptr++) {
@@ -835,7 +928,7 @@ end:
 }
 
 /* Checks stuffed inode directories */
-int check_linear_dir(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
+static int check_linear_dir(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 		     enum update_flags *update, struct metawalk_fxns *pass)
 {
 	int error = 0;
@@ -914,10 +1007,10 @@ int remove_dentry_from_dir(struct gfs2_sbd *sbp, uint64_t dir,
 	struct gfs2_block_query q;
 	int error;
 
-	log_debug("Removing dentry %" PRIu64 " (0x%" PRIx64 ") from directory %"
-			  PRIu64" (0x%" PRIx64 ")\n", dentryblock, dentryblock, dir, dir);
+	log_debug( _("Removing dentry %" PRIu64 " (0x%" PRIx64 ") from directory %"
+			  PRIu64" (0x%" PRIx64 ")\n"), dentryblock, dentryblock, dir, dir);
 	if(gfs2_check_range(sbp, dir)) {
-		log_err("Parent directory out of range\n");
+		log_err( _("Parent directory out of range\n"));
 		return 1;
 	}
 	remove_dentry_fxns.private = &dentryblock;
@@ -928,7 +1021,7 @@ int remove_dentry_from_dir(struct gfs2_sbd *sbp, uint64_t dir,
 		return -1;
 	}
 	if(q.block_type != gfs2_inode_dir) {
-		log_info("Parent block is not a directory...ignoring\n");
+		log_info( _("Parent block is not a directory...ignoring\n"));
 		return 1;
 	}
 	/* Need to run check_dir with a private var of dentryblock,
