@@ -16,12 +16,13 @@
 #include <time.h>
 #include <sys/ioctl.h>
 #include <inttypes.h>
+#include <libintl.h>
+#define _(String) gettext(String)
 
 #include "libgfs2.h"
 #include "fsck.h"
 #include "util.h"
 #include "link.h"
-#include "linux_endian.h"
 #include "metawalk.h"
 
 struct block_count {
@@ -73,7 +74,7 @@ static int leaf(struct gfs2_inode *ip, uint64_t block,
 {
 	struct block_count *bc = (struct block_count *) private;
 
-	log_debug("\tLeaf block at %15" PRIu64 " (0x%" PRIx64 ")\n",
+	log_debug( _("\tLeaf block at %15" PRIu64 " (0x%" PRIx64 ")\n"),
 			  block, block);
 	gfs2_block_set(ip->i_sbd, bl, block, gfs2_leaf_blk);
 	bc->indir_count++;
@@ -93,7 +94,7 @@ static int check_metalist(struct gfs2_inode *ip, uint64_t block,
 	if (gfs2_check_range(ip->i_sbd, block)){ /* blk outside of FS */
 		gfs2_block_set(ip->i_sbd, bl, ip->i_di.di_num.no_addr,
 			       gfs2_bad_block);
-		log_debug("Bad indirect block pointer (out of range).\n");
+		log_debug( _("Bad indirect block pointer (out of range).\n"));
 
 		return 1;
 	}
@@ -102,16 +103,16 @@ static int check_metalist(struct gfs2_inode *ip, uint64_t block,
 		return -1;
 	}
 	if(q.block_type != gfs2_block_free) {
-		log_debug("Found duplicate block in indirect block -"
-				  " was marked %d\n", q.block_type);
-		gfs2_block_mark(ip->i_sbd, bl, block, gfs2_dup_block);
+		log_err( _("Found duplicate block referenced as metadata in "
+			   "indirect block - was marked %d\n"), q.block_type);
+		gfs2_block_set(ip->i_sbd, bl, block, gfs2_dup_block);
 		found_dup = 1;
 	}
-	nbh = bread(ip->i_sbd, block);
+	nbh = bread(&ip->i_sbd->buf_list, block);
 
 	if (gfs2_check_meta(nbh, GFS2_METATYPE_IN)){
-		log_debug("Bad indirect block pointer "
-				  "(points to something that is not an indirect block).\n");
+		log_debug( _("Bad indirect block pointer "
+				  "(points to something that is not an indirect block).\n"));
 		if(!found_dup) {
 			gfs2_block_set(ip->i_sbd, bl, block, gfs2_meta_inval);
 			brelse(nbh, not_updated);
@@ -120,9 +121,11 @@ static int check_metalist(struct gfs2_inode *ip, uint64_t block,
 	}else  /* blk check ok */
 		*bh = nbh;
 
-	log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to indirect block.\n",
-			  block, block);
-	gfs2_block_set(ip->i_sbd, bl, block, gfs2_indir_blk);
+	if (!found_dup) {
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to indirect "
+			     "block.\n"), block, block);
+		gfs2_block_set(ip->i_sbd, bl, block, gfs2_indir_blk);
+	}
 	bc->indir_count++;
 
 	return 0;
@@ -132,9 +135,10 @@ static int check_data(struct gfs2_inode *ip, uint64_t block, void *private)
 {
 	struct gfs2_block_query q = {0};
 	struct block_count *bc = (struct block_count *) private;
+	int error = 0, btype;
 
 	if (gfs2_check_range(ip->i_sbd, block)) {
-		log_err( "Bad data block pointer (out of range)\n");
+		log_err( _("Bad data block pointer (out of range)\n"));
 		/* Mark the owner of this block with the bad_block
 		 * designator so we know to check it for out of range blocks later */
 		gfs2_block_set(ip->i_sbd, bl, ip->i_di.di_num.no_addr,
@@ -146,17 +150,64 @@ static int check_data(struct gfs2_inode *ip, uint64_t block, void *private)
 		return -1;
 	}
 	if(q.block_type != gfs2_block_free) {
-		log_debug("Found duplicate block at %" PRIu64 " (0x%"PRIx64 ")\n",
-				  block, block);
+		log_err( _("Found duplicate block referenced as data at %"
+			   PRIu64 " (0x%"PRIx64 ")\n"), block, block);
+		if (q.block_type != gfs2_meta_inval) {
+			gfs2_block_mark(ip->i_sbd, bl, block, gfs2_dup_block);
+			return 1;
+		}
+		/* An explanation is in order here.  At this point we found
+		   a duplicate block, a block that was already referenced
+		   somewhere else.  We'll resolve those duplicates in pass1b.
+		   However, if the block is marked "invalid" that's a special
+		   case.  It's likely that the block was discovered to be
+		   invalid metadata--i.e. doesn't have a metadata header.
+		   However, it still may be a valid data block, since they
+		   won't have metadata headers.  In that case, the block is
+		   marked as duplicate, but also as a data block. */
+		error = 1;
+		gfs2_block_unmark(ip->i_sbd, bl, block, gfs2_meta_inval);
 		gfs2_block_mark(ip->i_sbd, bl, block, gfs2_dup_block);
-		bc->data_count++;
-		return 1;
 	}
-	log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to data block\n", block,
-			  block);
-	gfs2_block_set(ip->i_sbd, bl, block, gfs2_block_used);
+	log_debug( _("Marking block %llu (0x%llx) as data block\n"),
+		   (unsigned long long)block, (unsigned long long)block);
+	gfs2_block_mark(ip->i_sbd, bl, block, gfs2_block_used);
+
+	/* This is also confusing, so I'll clarify.  There are two bitmaps:
+	   (1) The gfs2_bmap that fsck uses to keep track of what block
+	   type has been discovered, and (2) The rgrp bitmap.  Function
+	   gfs2_block_set is used to set the former and gfs2_set_bitmap
+	   is used to set the latter.  In this function we need to set both
+	   because we found a "data" block that could be "meta" in the rgrp
+	   bitmap.  If we don't we could run into the data block again as
+	   metadata when we're traversing the metadata with gfs2_next_rg_meta
+	   in func pass1().  If that happens, it will look at the block,
+	   say "hey this isn't metadata" and mark it incorrectly as an
+	   invalid metadata block and free it.  Ordinarily, fsck will wait
+	   until pass5 to sync (2) so that it agrees with (1).  However, in
+	   this case, it's better to do it upfront.  The duplicate solving
+	   code in pass1b.c is better at resolving metadata referencing a
+	   data block than it is at resolving a data block referencing a
+	   metadata block. */
+	btype = gfs2_get_bitmap(ip->i_sbd, block, NULL);
+	if (btype != GFS2_BLKST_USED) {
+		const char *allocdesc[] = {"free space", "data", "unlinked",
+					   "metadata", "reserved"};
+
+		errors_found++;
+		log_err( _("Block %llu (0x%llx) seems to be data, but is "
+			   "marked as %s.\n"), (unsigned long long)block,
+			   (unsigned long long)block, allocdesc[btype]);
+		if(query(&opts, _("Okay to mark it as 'data'? (y/n)"))) {
+			errors_corrected++;
+			gfs2_set_bitmap(ip->i_sbd, block, GFS2_BLKST_USED);
+			log_err( _("The block was reassigned as data.\n"));
+		} else {
+			log_err( _("The invalid block was ignored.\n"));
+		}
+	}
 	bc->data_count++;
-	return 0;
+	return error;
 }
 
 /* clear_eas - clear the extended attributes for an inode
@@ -175,18 +226,19 @@ static int clear_eas(struct gfs2_inode *ip, struct block_count *bc,
 	struct gfs2_sbd *sdp = ip->i_sbd;
 
 	*want_updated = not_updated;
-	log_err("Inode #%" PRIu64 " (0x%" PRIx64 "): %s",
-		ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr, emsg);
+	log_err( _("Inode #%llu (0x%llx): %s"),
+		(unsigned long long)ip->i_di.di_num.no_addr,
+		(unsigned long long)ip->i_di.di_num.no_addr, emsg);
 	if (block)
-		log_err(" at block #%" PRIu64 " (0x%" PRIx64 ")",
+		log_err( _(" at block #%" PRIu64 " (0x%" PRIx64 ")"),
 			block, block);
 	log_err(".\n");
-	if (query(&opts, "Clear the bad EA? (y/n) ")) {
+	errors_found++;
+	if ((errors_corrected += query(&opts, _("Clear the bad Extended "
+						"Attribute? (y/n) ")))) {
 		if (block == 0)
 			block = ip->i_di.di_eattr;
-		gfs2_block_clear(sdp, bl, block, gfs2_eattr_block);
 		if (!duplicate) {
-			gfs2_block_clear(sdp, bl, block, gfs2_indir_blk);
 			gfs2_block_set(sdp, bl, block, gfs2_block_free);
 			gfs2_set_bitmap(sdp, block, GFS2_BLKST_FREE);
 		}
@@ -197,9 +249,10 @@ static int clear_eas(struct gfs2_inode *ip, struct block_count *bc,
 		ip->i_di.di_blocks = 1 + bc->indir_count + bc->data_count;
 		gfs2_dinode_out(&ip->i_di, ip->i_bh->b_data);
 		*want_updated = updated;
+		log_err( _("The bad Extended Attribute was removed.\n"));
 		return 1;
 	} else {
-		log_err("The bad EA was not fixed.\n");
+		log_err( _("The bad Extended Attribute was not fixed.\n"));
 		bc->ea_count++;
 		return 0;
 	}
@@ -233,31 +286,38 @@ static int check_eattr_indir(struct gfs2_inode *ip, uint64_t indirect,
 	   check if it really is an EA.  If it is, let duplicate
 	   handling sort it out.  If it isn't, clear it but don't
 	   count it as a duplicate. */
-	*bh = bread(sdp, indirect);
+	*bh = bread(&sdp->buf_list, indirect);
 	if(gfs2_check_meta(*bh, GFS2_METATYPE_IN)) {
 		if(q.block_type != gfs2_block_free) { /* Duplicate? */
 			if (!clear_eas(ip, bc, indirect, 1, want_updated,
-				       "Bad indirect EA duplicate found"))
+				       _("Bad indirect Extended Attribute "
+					 "duplicate found"))) {
 				gfs2_block_set(sdp, bl, indirect,
 					       gfs2_dup_block);
+				bc->ea_count++;
+			}
 			return 1;
 		}
 		clear_eas(ip, bc, indirect, 0, want_updated,
-			  "EA indirect block has incorrect type");
+			  _("Extended Attribute indirect block has incorrect "
+			    "type"));
 		return 1;
 	}
 	if(q.block_type != gfs2_block_free) { /* Duplicate? */
-		log_err("Inode #%" PRIu64 " (0x%" PRIx64
-			"): Duplicate EA indirect block found at #%" PRIu64
-			" (0x%" PRIx64 ").\n",
-			ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr,
-			indirect, indirect);
+		log_err( _("Inode #%llu (0x%llx): Duplicate Extended "
+			   "Attribute indirect block found at #%llu "
+			   "(0x%llx).\n"),
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)indirect,
+			 (unsigned long long)indirect);
 		gfs2_block_set(sdp, bl, indirect, gfs2_dup_block);
 		bc->ea_count++;
 		ret = 1;
 	} else {
-		log_debug("Setting #%" PRIu64 " (0x%" PRIx64
-			  ") to indirect EA block\n", indirect, indirect);
+		log_debug( _("Setting #%" PRIu64 " (0x%" PRIx64
+			  ") to indirect Extended Attribute block\n"),
+			   indirect, indirect);
 		gfs2_block_set(sdp, bl, indirect, gfs2_indir_blk);
 		bc->ea_count++;
 	}
@@ -268,9 +328,9 @@ static int finish_eattr_indir(struct gfs2_inode *ip, int indir_ok,
 			      enum update_flags *want_updated, void *private)
 {
 	if (indir_ok) {
-		log_debug("Marking inode #%" PRIu64 " (0x%"
-			  PRIx64 ") with eattr block\n",
-			  ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
+		log_debug( _("Marking inode #%llu (0x%llx) with eattr block\n"),
+			  (unsigned long long)ip->i_di.di_num.no_addr,
+			  (unsigned long long)ip->i_di.di_num.no_addr);
 		/* Mark the inode as having an eattr in the block map
 		   so pass1c can check it. */
 		gfs2_block_mark(ip->i_sbd, bl, ip->i_di.di_num.no_addr,
@@ -278,7 +338,7 @@ static int finish_eattr_indir(struct gfs2_inode *ip, int indir_ok,
 		return 0;
 	}
 	clear_eas(ip, (struct block_count *)private, 0, 0, want_updated,
-		  "has unrecoverable indirect EA errors");
+		  _("has unrecoverable indirect Extended Attribute errors"));
 	return 0;
 }
 
@@ -308,11 +368,15 @@ static int check_extended_leaf_eattr(struct gfs2_inode *ip, uint64_t *data_ptr,
 	int ret = 0;
 
 	if(gfs2_check_range(sdp, el_blk)){
-		log_err("Inode #%" PRIu64 " (0x%" PRIx64 "): EA extended "
-			"leaf block #%" PRIu64 " (0x%" PRIx64
-			") is out of range.\n",
-			ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr,
-			el_blk, el_blk);
+		log_err( _("Inode #%llu (0x%llx): Extended Attribute block "
+			   "%llu (0x%llx) has an extemded leaf block #%llu "
+			   "(0x%llx) that is out of range.\n"),
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_eattr,
+			 (unsigned long long)ip->i_di.di_eattr,
+			 (unsigned long long)el_blk,
+			 (unsigned long long)el_blk);
 		gfs2_block_set(sdp, bl, ip->i_di.di_eattr, gfs2_bad_block);
 		return 1;
 	}
@@ -321,7 +385,7 @@ static int check_extended_leaf_eattr(struct gfs2_inode *ip, uint64_t *data_ptr,
 		stack;
 		return -1;
 	}
-	el_buf = bread(sdp, el_blk);
+	el_buf = bread(&sdp->buf_list, el_blk);
 
 	/* Special duplicate processing:  If we have an EA block,
 	   check if it really is an EA.  If it is, let duplicate
@@ -330,22 +394,24 @@ static int check_extended_leaf_eattr(struct gfs2_inode *ip, uint64_t *data_ptr,
 	if(gfs2_check_meta(el_buf, GFS2_METATYPE_ED)) {
 		if(q.block_type != gfs2_block_free) /* Duplicate? */
 			clear_eas(ip, bc, el_blk, 1, want_updated,
-				  "has bad extended EA duplicate");
+				  _("has bad extended Extended Attribute "
+				    "duplicate"));
 		else
 			clear_eas(ip, bc, el_blk, 0, want_updated,
-				  "EA extended leaf block has incorrect type");
+				  _("Extended Attribute extended leaf block "
+				    "has incorrect type"));
 		ret = 1;
 	} else { /* If this looks like an EA */
 		if(q.block_type != gfs2_block_free) { /* Duplicate? */
-			log_debug("Duplicate block found at #%" PRIu64
-				  " (0x%" PRIx64 ").\n",
+			log_debug( _("Duplicate block found at #%" PRIu64
+				  " (0x%" PRIx64 ").\n"),
 				  el_blk, el_blk);
 			gfs2_block_set(sdp, bl, el_blk, gfs2_dup_block);
 			bc->ea_count++;
 			ret = 1;
 		} else {
-			log_debug("Setting block #%" PRIu64
-				  " (0x%" PRIx64 ") to eattr block\n",
+			log_debug( _("Setting block #%" PRIu64
+				  " (0x%" PRIx64 ") to eattr block\n"),
 				  el_blk, el_blk);
 			gfs2_block_set(sdp, bl, el_blk, gfs2_meta_eattr);
 			bc->ea_count++;
@@ -366,18 +432,22 @@ static int check_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 	struct gfs2_block_query q = {0};
 	struct block_count *bc = (struct block_count *) private;
 
+	/* Figure out the parent's block type */
+	gfs2_block_check(sdp, bl, parent, &q);
 	/* This inode contains an eattr - it may be invalid, but the
 	 * eattr attributes points to a non-zero block */
-	if (parent != ip->i_di.di_num.no_addr) { /* if parent isn't the inode */
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to eattr block\n",
+	if (parent != ip->i_di.di_num.no_addr && /* if parent isn't the inode*/
+	    q.block_type != gfs2_indir_blk) { /* parent isn't indirect block */
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to eattr block\n"),
 				  parent, parent);
 		gfs2_block_set(sdp, bl, parent, gfs2_eattr_block);
 	}
 	if(gfs2_check_range(sdp, block)){
-		log_warn("Inode #%" PRIu64 " (0x%" PRIx64 "): EA leaf block "
-			 "#%" PRIu64 " (0x%" PRIx64 ") is out of range.\n",
-			 ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr,
-			 block, block);
+		log_warn( _("Inode #%llu (0x%llx): Extended Attribute leaf "
+			    "block #%llu (0x%llx) is out of range.\n"),
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)block, (unsigned long long)block);
 		gfs2_block_set(sdp, bl, ip->i_di.di_eattr, gfs2_bad_block);
 		ret = 1;
 	}
@@ -390,29 +460,31 @@ static int check_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 		   check if it really is an EA.  If it is, let duplicate
 		   handling sort it out.  If it isn't, clear it but don't
 		   count it as a duplicate. */
-		leaf_bh = bread(sdp, block);
+		leaf_bh = bread(&sdp->buf_list, block);
 		if(gfs2_check_meta(leaf_bh, GFS2_METATYPE_EA)) {
 			if(q.block_type != gfs2_block_free) { /* Duplicate? */
 				clear_eas(ip, bc, block, 1, want_updated,
-					  "Bad EA duplicate found");
+					  _("Bad Extended Attribute duplicate "
+					    "found"));
 			} else {
 				clear_eas(ip, bc, block, 0, want_updated,
-					  "EA leaf block has incorrect type");
+					  _("Extended Attribute leaf block "
+					    "has incorrect type"));
 			}
 			ret = 1;
 			brelse(leaf_bh, not_updated);
 		} else { /* If this looks like an EA */
 			if(q.block_type != gfs2_block_free) { /* Duplicate? */
-				log_debug("Duplicate block found at #%" PRIu64
-					  " (0x%" PRIx64 ").\n",
+				log_debug( _("Duplicate block found at #%" PRIu64
+					  " (0x%" PRIx64 ").\n"),
 					  block, block);
 				gfs2_block_set(sdp, bl, block, gfs2_dup_block);
 				bc->ea_count++;
 				ret = 1;
 				brelse(leaf_bh, not_updated);
 			} else {
-				log_debug("Setting block #%" PRIu64
-					  " (0x%" PRIx64 ") to eattr block\n",
+				log_debug( _("Setting block #%" PRIu64
+					  " (0x%" PRIx64 ") to eattr block\n"),
 					  block, block);
 				gfs2_block_set(sdp, bl, block,
 					       gfs2_meta_eattr);
@@ -460,14 +532,14 @@ static int check_eattr_entries(struct gfs2_inode *ip,
 		if(max_ptrs > ea_hdr->ea_num_ptrs) {
 			return 1;
 		} else {
-			log_debug("  Pointers Required: %d\n  Pointers Reported: %d\n",
+			log_debug( _("  Pointers Required: %d\n  Pointers Reported: %d\n"),
 				  max_ptrs, ea_hdr->ea_num_ptrs);
 		}
 	}
 	return 0;
 }
 
-int clear_metalist(struct gfs2_inode *ip, uint64_t block,
+static int clear_metalist(struct gfs2_inode *ip, uint64_t block,
 		   struct gfs2_buffer_head **bh, void *private)
 {
 	struct gfs2_block_query q = {0};
@@ -485,7 +557,7 @@ int clear_metalist(struct gfs2_inode *ip, uint64_t block,
 	return 0;
 }
 
-int clear_data(struct gfs2_inode *ip, uint64_t block, void *private)
+static int clear_data(struct gfs2_inode *ip, uint64_t block, void *private)
 {
 	struct gfs2_block_query q = {0};
 
@@ -501,7 +573,7 @@ int clear_data(struct gfs2_inode *ip, uint64_t block, void *private)
 
 }
 
-int clear_leaf(struct gfs2_inode *ip, uint64_t block,
+static int clear_leaf(struct gfs2_inode *ip, uint64_t block,
 	       struct gfs2_buffer_head *bh, void *private)
 {
 	struct gfs2_block_query q = {0};
@@ -512,7 +584,7 @@ int clear_leaf(struct gfs2_inode *ip, uint64_t block,
 		return -1;
 	}
 	if(!q.dup_block) {
-		log_crit("Setting leaf #%" PRIu64 " (0x%" PRIx64 ") invalid\n",
+		log_crit( _("Setting leaf #%" PRIu64 " (0x%" PRIx64 ") invalid\n"),
 				 block, block);
 		if(gfs2_block_set(ip->i_sbd, bl, block, gfs2_block_free)) {
 			stack;
@@ -535,17 +607,17 @@ int add_to_dir_list(struct gfs2_sbd *sbp, uint64_t block)
 	 * matter too much */
 	find_di(sbp, block, &di);
 	if(di) {
-		log_err("Attempting to add directory block #%" PRIu64
-				" (0x%" PRIx64 ") which is already in list\n", block, block);
+		log_err( _("Attempting to add directory block #%" PRIu64
+				" (0x%" PRIx64 ") which is already in list\n"), block, block);
 		return -1;
 	}
 
 	if(!(newdi = (struct dir_info *) malloc(sizeof(struct dir_info)))) {
-		log_crit("Unable to allocate dir_info structure\n");
+		log_crit( _("Unable to allocate dir_info structure\n"));
 		return -1;
 	}
 	if(!memset(newdi, 0, sizeof(*newdi))) {
-		log_crit("Error while zeroing dir_info structure\n");
+		log_crit( _("Error while zeroing dir_info structure\n"));
 		return -1;
 	}
 
@@ -554,7 +626,7 @@ int add_to_dir_list(struct gfs2_sbd *sbp, uint64_t block)
 	return 0;
 }
 
-int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
+static int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 			  uint64_t block)
 {
 	struct gfs2_block_query q = {0};
@@ -571,18 +643,21 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 
 	ip = fsck_inode_get(sdp, bh);
 	if (ip->i_di.di_num.no_addr != block) {
-		log_err("Inode #%" PRIu64 " (0x%" PRIx64
-				"): Bad inode address found: %"	PRIu64 " (0x%" PRIx64 ")\n",
-				block, block, ip->i_di.di_num.no_addr,
-				ip->i_di.di_num.no_addr);
-		if(query(&opts, "Fix address in inode at block #%"
-				 PRIu64 " (0x%" PRIx64 ")? (y/n) ", block, block)) {
+		log_err( _("Inode #%llu (0x%llx): Bad inode address found: %llu "
+			"(0x%llx)\n"), (unsigned long long)block,
+			(unsigned long long)block,
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)ip->i_di.di_num.no_addr);
+		errors_found++;
+		if((errors_corrected +=
+		    query(&opts, _("Fix address in inode at block #%"
+			  PRIu64 " (0x%" PRIx64 ")? (y/n) "), block, block))) {
 			ip->i_di.di_num.no_addr = ip->i_di.di_num.no_formal_ino = block;
 			gfs2_dinode_out(&ip->i_di, ip->i_bh->b_data);
 			f = updated;
 		} else
-			log_err("Address in inode at block #%" PRIu64
-				 " (0x%" PRIx64 ") not fixed\n", block, block);
+			log_err( _("Address in inode at block #%" PRIu64
+				 " (0x%" PRIx64 ") not fixed\n"), block, block);
 	}
 
 	if(gfs2_block_check(sdp, bl, block, &q)) {
@@ -591,8 +666,8 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		return -1;
 	}
 	if(q.block_type != gfs2_block_free) {
-		log_debug("Found duplicate block at #%" PRIu64 " (0x%" PRIx64 ")\n",
-				  block, block);
+		log_err( _("Found duplicate block referenced as an inode at "
+			   "#%" PRIu64 " (0x%" PRIx64 ")\n"), block, block);
 		if(gfs2_block_mark(sdp, bl, block, gfs2_dup_block)) {
 			stack;
 			fsck_inode_put(ip, f);
@@ -605,7 +680,7 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 	switch(ip->i_di.di_mode & S_IFMT) {
 
 	case S_IFDIR:
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to directory inode.\n",
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to directory inode.\n"),
 				  block, block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_inode_dir)) {
 			stack;
@@ -619,7 +694,7 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		}
 		break;
 	case S_IFREG:
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to file inode.\n",
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to file inode.\n"),
 				  block, block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_inode_file)) {
 			stack;
@@ -628,7 +703,7 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		}
 		break;
 	case S_IFLNK:
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to symlink inode.\n",
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to symlink inode.\n"),
 				  block, block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_inode_lnk)) {
 			stack;
@@ -637,7 +712,7 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		}
 		break;
 	case S_IFBLK:
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to block dev inode.\n",
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to block dev inode.\n"),
 				  block, block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_inode_blk)) {
 			stack;
@@ -646,7 +721,7 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		}
 		break;
 	case S_IFCHR:
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to char dev inode.\n",
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to char dev inode.\n"),
 				  block, block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_inode_chr)) {
 			stack;
@@ -655,7 +730,7 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		}
 		break;
 	case S_IFIFO:
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to fifo inode.\n",
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to fifo inode.\n"),
 				  block, block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_inode_fifo)) {
 			stack;
@@ -664,7 +739,7 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		}
 		break;
 	case S_IFSOCK:
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to socket inode.\n",
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to socket inode.\n"),
 				  block, block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_inode_sock)) {
 			stack;
@@ -673,7 +748,7 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		}
 		break;
 	default:
-		log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to invalid.\n",
+		log_debug( _("Setting %" PRIu64 " (0x%" PRIx64 ") to invalid.\n"),
 				  block, block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_meta_inval)) {
 			stack;
@@ -690,35 +765,17 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 		return -1;
 	}
 
-	/* FIXME: fix height and depth here - wasn't implemented in
-	 * old fsck either, so no biggy... */
-	if (ip->i_di.di_height < compute_height(sdp, ip->i_di.di_size)){
-		log_warn("Dinode #%" PRIu64 " (0x%" PRIx64 ") has bad height  "
-				 "Found %u, Expected >= %u\n", ip->i_di.di_num.no_addr, 
-				 ip->i_di.di_num.no_addr, ip->i_di.di_height,
-				 compute_height(sdp, ip->i_di.di_size));
-			/* once implemented, remove continue statement */
-		log_warn("Marking inode invalid\n");
-		if(gfs2_block_set(sdp, bl, block, gfs2_meta_inval)) {
-			stack;
-			fsck_inode_put(ip, f);
-			return -1;
-		}
-		gfs2_set_bitmap(sdp, block, GFS2_BLKST_FREE);
-		fsck_inode_put(ip, f);
-		return 0;
-	}
-
 	if (S_ISDIR(ip->i_di.di_mode) &&
 	    (ip->i_di.di_flags & GFS2_DIF_EXHASH)) {
 		if (((1 << ip->i_di.di_depth) * sizeof(uint64_t)) != ip->i_di.di_size){
-			log_warn("Directory dinode #%" PRIu64 " (0x%" PRIx64
-					 ") has bad depth.  Found %u, Expected %u\n",
-					 ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr,
-					 ip->i_di.di_depth,
-					 (1 >> (ip->i_di.di_size/sizeof(uint64_t))));
+			log_warn( _("Directory dinode #%llu (0x%llx"
+				 ") has bad depth.  Found %u, Expected %u\n"),
+				 (unsigned long long)ip->i_di.di_num.no_addr,
+				 (unsigned long long)ip->i_di.di_num.no_addr,
+				 ip->i_di.di_depth,
+				 (1 >> (ip->i_di.di_size/sizeof(uint64_t))));
 			/* once implemented, remove continue statement */
-			log_warn("Marking inode invalid\n");
+			log_warn( _("Marking inode invalid\n"));
 			if(gfs2_block_set(sdp, bl, block, gfs2_meta_inval)) {
 				stack;
 				fsck_inode_put(ip, f);
@@ -733,13 +790,14 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 	pass1_fxns.private = &bc;
 
 	error = check_metatree(ip, &pass1_fxns);
-	if(error < 0) {
+	if (fsck_abort || error < 0) {
 		fsck_inode_put(ip, f);
 		return 0;
 	}
 	if(error > 0) {
-		log_warn("Marking inode #%" PRIu64 " (0x%" PRIx64 ") invalid\n",
-			 ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
+		log_warn( _("Marking inode #%llu (0x%llx) invalid\n"),
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr);
 		/* FIXME: Must set all leaves invalid as well */
 		check_metatree(ip, &invalidate_metatree);
 		gfs2_block_set(sdp, bl, ip->i_di.di_num.no_addr,
@@ -753,41 +811,55 @@ int handle_di(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 
 	if (ip->i_di.di_blocks != 
 		(1 + bc.indir_count + bc.data_count + bc.ea_count)) {
-		log_err("Inode #%" PRIu64 " (0x%" PRIx64 "): Ondisk block count (%"
-				PRIu64 ") does not match what fsck found (%" PRIu64 ")\n",
-				ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr,
-				ip->i_di.di_blocks,
-				1 + bc.indir_count + bc.data_count + bc.ea_count);
-		if(query(&opts, "Fix ondisk block count? (y/n) ")) {
+		log_err( _("Inode #%llu (0x%llx): Ondisk block count (%llu"
+			") does not match what fsck found (%llu)\n"),
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)ip->i_di.di_blocks,
+			(unsigned long long)1 + bc.indir_count +
+			bc.data_count + bc.ea_count);
+		errors_found++;
+		if ((errors_corrected +=
+		     query(&opts, _("Fix ondisk block count? (y/n) ")))) {
 			ip->i_di.di_blocks = 1 + bc.indir_count + bc.data_count +
 				bc.ea_count;
 			gfs2_dinode_out(&ip->i_di, ip->i_bh->b_data);
 			f = updated;
 		} else
-			log_err("Bad block count for #%" PRIu64 " (0x%" PRIx64
-					") not fixed\n", ip->i_di.di_num.no_addr,
-					ip->i_di.di_num.no_addr);
+			log_err( _("Bad block count for #%llu (0x%llx"
+				") not fixed\n"),
+				(unsigned long long)ip->i_di.di_num.no_addr,
+				(unsigned long long)ip->i_di.di_num.no_addr);
 	}
 
 	fsck_inode_put(ip, f);
 	return 0;
 }
 
-int scan_meta(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
+static int scan_meta(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh,
 			  uint64_t block)
 {
 	if (gfs2_check_meta(bh, 0)) {
-		log_debug("Found invalid metadata at #%" PRIu64 " (0x%" PRIx64 ")\n",
-				  block, block);
+		errors_found++;
+
+		log_info( _("Found invalid metadata at #%llu (0x%llx)\n"),
+			  (unsigned long long)block,
+			  (unsigned long long)block);
 		if(gfs2_block_set(sdp, bl, block, gfs2_meta_inval)) {
 			stack;
 			return -1;
 		}
-		gfs2_set_bitmap(sdp, block, GFS2_BLKST_FREE);
+		if(query(&opts, _("Okay to free the invalid block? (y/n)"))) {
+			errors_corrected++;
+			gfs2_set_bitmap(sdp, block, GFS2_BLKST_FREE);
+			log_err( _("The invalid block was freed.\n"));
+		} else {
+			log_err( _("The invalid block was ignored.\n"));
+		}
 		return 0;
 	}
 
-	log_debug("Checking metadata block #%" PRIu64 " (0x%" PRIx64 ")\n", block,
+	log_debug( _("Checking metadata block #%" PRIu64 " (0x%" PRIx64 ")\n"), block,
 			  block);
 
 	if (!gfs2_check_meta(bh, GFS2_METATYPE_DI)) {
@@ -851,20 +923,22 @@ int pass1(struct gfs2_sbd *sbp)
 
 	for (tmp = sbp->rglist.next; tmp != &sbp->rglist;
 	     tmp = tmp->next, rg_count++){
-		log_info("Checking metadata in Resource Group #%" PRIu64 "\n",
+		log_info( _("Checking metadata in Resource Group #%" PRIu64 "\n"),
 				 rg_count);
 		rgd = osi_list_entry(tmp, struct rgrp_list, list);
 		if(gfs2_rgrp_read(sbp, rgd)){
 			stack;
-			return -1;
+			return FSCK_ERROR;
 		}
-		log_debug("RG at %" PRIu64 " (0x%" PRIx64 ") is %u long\n",
-				  rgd->ri.ri_addr, rgd->ri.ri_addr, rgd->ri.ri_length);
+		log_debug( _("RG at %llu (0x%llx) is %u long\n"),
+			  (unsigned long long)rgd->ri.ri_addr,
+			  (unsigned long long)rgd->ri.ri_addr,
+			  rgd->ri.ri_length);
 		for (i = 0; i < rgd->ri.ri_length; i++) {
 			if(gfs2_block_set(sbp, bl, rgd->ri.ri_addr + i,
 					  gfs2_meta_other)){
 				stack;
-				return -1;
+				return FSCK_ERROR;
 			}
 		}
 
@@ -879,25 +953,25 @@ int pass1(struct gfs2_sbd *sbp)
 			warm_fuzzy_stuff(block);
 			if (fsck_abort) { /* if asked to abort */
 				gfs2_rgrp_relse(rgd, not_updated);
-				return 0;
+				return FSCK_OK;
 			}
 			if (skip_this_pass) {
-				printf("Skipping pass 1 is not a good idea.\n");
+				printf( _("Skipping pass 1 is not a good idea.\n"));
 				skip_this_pass = FALSE;
 				fflush(stdout);
 			}
-			bh = bread(sbp, block);
+			bh = bread(&sbp->buf_list, block);
 
 			if (scan_meta(sbp, bh, block)) {
 				stack;
 				brelse(bh, not_updated);
 				gfs2_rgrp_relse(rgd, not_updated);
-				return -1;
+				return FSCK_ERROR;
 			}
 			brelse(bh, not_updated);
 			first = 0;
 		}
 		gfs2_rgrp_relse(rgd, not_updated);
 	}
-	return 0;
+	return FSCK_OK;
 }
