@@ -1,8 +1,9 @@
 #include <inttypes.h>
-#include <linux_endian.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <libintl.h>
+#define _(String) gettext(String)
 
 #include "libgfs2.h"
 #include "fsck.h"
@@ -20,12 +21,6 @@ struct inode_with_dups {
 	char *name;
 };
 
-struct blocks {
-	osi_list_t list;
-	uint64_t block_no;
-	osi_list_t ref_inode_list;
-};
-
 struct fxn_info {
 	uint64_t block;
 	int found;
@@ -33,7 +28,7 @@ struct fxn_info {
 };
 
 struct dup_handler {
-	struct blocks *b;
+	struct dup_blocks *b;
 	struct inode_with_dups *id;
 	int ref_inode_count;
 	int ref_count;
@@ -72,7 +67,7 @@ static int check_eattr_indir(struct gfs2_inode *ip, uint64_t block,
 
 	*want_updated = not_updated;
 	inc_if_found(block, 0, private);
-	indir_bh = bread(sbp, block);
+	indir_bh = bread(&sbp->buf_list, block);
 	*bh = indir_bh;
 
 	return 0;
@@ -87,17 +82,16 @@ static int check_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 
 	*want_updated = not_updated;
 	inc_if_found(block, 0, private);
-	leaf_bh = bread(sbp, block);
+	leaf_bh = bread(&sbp->buf_list, block);
 
 	*bh = leaf_bh;
 	return 0;
 }
 
 static int check_eattr_entry(struct gfs2_inode *ip,
-							 struct gfs2_buffer_head *leaf_bh,
-							 struct gfs2_ea_header *ea_hdr,
-							 struct gfs2_ea_header *ea_hdr_prev,
-							 void *private)
+			     struct gfs2_buffer_head *leaf_bh,
+			     struct gfs2_ea_header *ea_hdr,
+			     struct gfs2_ea_header *ea_hdr_prev, void *private)
 {
 	return 0;
 }
@@ -122,12 +116,12 @@ static int find_dentry(struct gfs2_inode *ip, struct gfs2_dirent *de,
 		       enum update_flags *update, uint16_t *count, void *priv)
 {
 	osi_list_t *tmp1, *tmp2;
-	struct blocks *b;
+	struct dup_blocks *b;
 	struct inode_with_dups *id;
 	struct gfs2_leaf leaf;
 
 	osi_list_foreach(tmp1, &ip->i_sbd->dup_blocks.list) {
-		b = osi_list_entry(tmp1, struct blocks, list);
+		b = osi_list_entry(tmp1, struct dup_blocks, list);
 		osi_list_foreach(tmp2, &b->ref_inode_list) {
 			id = osi_list_entry(tmp2, struct inode_with_dups,
 					    list);
@@ -139,11 +133,16 @@ static int find_dentry(struct gfs2_inode *ip, struct gfs2_dirent *de,
 			if(id->block_no == de->de_inum.no_addr) {
 				id->name = strdup(filename);
 				id->parent = ip->i_di.di_num.no_addr;
-				log_debug("Duplicate block %" PRIu64 " (0x%" PRIx64
-						  ") is in file or directory %" PRIu64
-						  " (0x%" PRIx64 ") named %s\n", id->block_no,
-						  id->block_no, ip->i_di.di_num.no_addr,
-						  ip->i_di.di_num.no_addr, filename);
+				log_debug( _("Duplicate block %llu (0x%llx"
+					  ") is in file or directory %llu"
+					  " (0x%llx) named %s\n"),
+					  (unsigned long long)id->block_no,
+					  (unsigned long long)id->block_no,
+					  (unsigned long long)
+					  ip->i_di.di_num.no_addr,
+					  (unsigned long long)
+					  ip->i_di.di_num.no_addr,
+					  filename);
 				/* If there are duplicates of
 				 * duplicates, I guess we'll miss them
 				 * here */
@@ -166,13 +165,15 @@ static int clear_dup_metalist(struct gfs2_inode *ip, uint64_t block,
 	if(dh->ref_count == 1)
 		return 1;
 	if(block == dh->b->block_no) {
-		log_err("Found dup in inode \"%s\" (block #%"PRIu64
-				") with block #%"PRIu64"\n",
-				dh->id->name ? dh->id->name : "unknown name",
-				ip->i_di.di_num.no_addr, block);
-		log_err("Inode %s is in directory %"PRIu64" (0x%" PRIx64 ")\n",
-				dh->id->name ? dh->id->name : "",
-				dh->id->parent, dh->id->parent);
+		log_err( _("Found duplicate reference in inode \"%s\" at "
+			   "block #%llu (0x%llx) with block #%llu (0x%llx)\n"),
+			 dh->id->name ? dh->id->name : _("unknown name"),
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)block, (unsigned long long)block);
+		log_err( _("Inode %s is in directory %"PRIu64" (0x%" PRIx64 ")\n"),
+			 dh->id->name ? dh->id->name : "", dh->id->parent,
+			 dh->id->parent);
 		inode_hash_remove(inode_hash, ip->i_di.di_num.no_addr);
 		/* Setting the block to invalid means the inode is
 		 * cleared in pass2 */
@@ -184,29 +185,9 @@ static int clear_dup_metalist(struct gfs2_inode *ip, uint64_t block,
 
 static int clear_dup_data(struct gfs2_inode *ip, uint64_t block, void *private)
 {
-	struct dup_handler *dh = (struct dup_handler *) private;
-
-	if(dh->ref_count == 1) {
-		return 1;
-	}
-	if(block == dh->b->block_no) {
-		log_err("Found dup in inode \"%s\" for block #%" PRIu64
-				" (0x%" PRIx64 ") at block #%" PRIu64 " (0x%" PRIx64 ")\n",
-				dh->id->name ? dh->id->name : "unknown name",
-				ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr, block,
-				block);
-		log_err("Inode %s is in directory %"PRIu64" (0x%" PRIx64 ")\n",
-				dh->id->name ? dh->id->name : "", dh->id->parent,
-				dh->id->parent);
-		inode_hash_remove(inode_hash, ip->i_di.di_num.no_addr);
-		/* Setting the block to invalid means the inode is
-		 * cleared in pass2 */
-		gfs2_block_set(ip->i_sbd, bl, ip->i_di.di_num.no_addr,
-			       gfs2_meta_inval);
-	}
-
-	return 0;
+	return clear_dup_metalist(ip, block, NULL, private);
 }
+
 static int clear_dup_eattr_indir(struct gfs2_inode *ip, uint64_t block,
 				 uint64_t parent, struct gfs2_buffer_head **bh,
 				 enum update_flags *want_updated,
@@ -220,12 +201,14 @@ static int clear_dup_eattr_indir(struct gfs2_inode *ip, uint64_t block,
 	if(dh->ref_count == 1)
 		return 1;
 	if(block == dh->b->block_no) {
-		log_err("Found dup in inode \"%s\" with address #%" PRIu64
-				" (0x%" PRIx64 ") with block #%" PRIu64 " (0x%" PRIx64 ")\n",
-				dh->id->name ? dh->id->name : "unknown name",
-				ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr, block,
-				block);
-		log_err("Inode %s is in directory %" PRIu64 " (0x%" PRIx64 ")\n",
+		log_err( _("Found dup in inode \"%s\" with address #%llu"
+			" (0x%llx) with block #%llu (0x%llx)\n"),
+			dh->id->name ? dh->id->name : _("unknown name"),
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)block,
+			(unsigned long long)block);
+		log_err( _("Inode %s is in directory %" PRIu64 " (0x%" PRIx64 ")\n"),
 				dh->id->name ? dh->id->name : "",
 				dh->id->parent, dh->id->parent);
 		gfs2_block_set(ip->i_sbd, bl, ip->i_di.di_eattr,
@@ -245,12 +228,14 @@ static int clear_dup_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 	if(dh->ref_count == 1)
 		return 1;
 	if(block == dh->b->block_no) {
-		log_err("Found dup in inode \"%s\" with address #%" PRIu64
-				" (0x%" PRIx64 ") with block #%" PRIu64 " (0x%" PRIx64 ")\n",
-				dh->id->name ? dh->id->name : "unknown name",
-				ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr, block,
-				block);
-		log_err("Inode %s is in directory %" PRIu64 " (0x%" PRIx64 ")\n",
+		log_err( _("Found dup in inode \"%s\" with address #%llu"
+			" (0x%llx) with block #%llu (0x%llx)\n"),
+			dh->id->name ? dh->id->name : _("unknown name"),
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)block,
+			(unsigned long long)block);
+		log_err( _("Inode %s is in directory %" PRIu64 " (0x%" PRIx64 ")\n"),
 				dh->id->name ? dh->id->name : "",
 				dh->id->parent, dh->id->parent);
 		/* mark the main eattr block invalid */
@@ -296,7 +281,7 @@ static int clear_eattr_entry (struct gfs2_inode *ip,
 		if(max_ptrs > ea_hdr->ea_num_ptrs)
 			return 1;
 		else {
-			log_debug("  Pointers Required: %d\n  Pointers Reported: %d\n",
+			log_debug( _("  Pointers Required: %d\n  Pointers Reported: %d\n"),
 					  max_ptrs, ea_hdr->ea_num_ptrs);
 		}
 	}
@@ -317,12 +302,13 @@ static int clear_eattr_extentry(struct gfs2_inode *ip, uint64_t *ea_data_ptr,
 	if(dh->ref_count == 1)
 		return 1;
 	if(block == dh->b->block_no) {
-		log_err("Found dup in inode \"%s\" with address #%" PRIu64
-				" (0x%" PRIx64 ") with block #%" PRIu64 " (0x%" PRIx64 ")\n",
-				dh->id->name ? dh->id->name : "unknown name",
-				ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr, block,
-				block);
-		log_err("Inode %s is in directory %" PRIu64 " (0x%" PRIx64 ")\n",
+		log_err( _("Found dup in inode \"%s\" with address #%llu"
+			" (0x%llx) with block #%llu (0x%llx)\n"),
+			dh->id->name ? dh->id->name : _("unknown name"),
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)ip->i_di.di_num.no_addr,
+			(unsigned long long)block, (unsigned long long)block);
+		log_err( _("Inode %s is in directory %" PRIu64 " (0x%" PRIx64 ")\n"),
 				dh->id->name ? dh->id->name : "",
 				dh->id->parent, dh->id->parent);
 		/* mark the main eattr block invalid */
@@ -335,7 +321,7 @@ static int clear_eattr_extentry(struct gfs2_inode *ip, uint64_t *ea_data_ptr,
 }
 
 /* Finds all references to duplicate blocks in the metadata */
-int find_block_ref(struct gfs2_sbd *sbp, uint64_t inode, struct blocks *b)
+static int find_block_ref(struct gfs2_sbd *sbp, uint64_t inode, struct dup_blocks *b)
 {
 	struct gfs2_inode *ip;
 	struct fxn_info myfi = {b->block_no, 0, 1};
@@ -354,15 +340,15 @@ int find_block_ref(struct gfs2_sbd *sbp, uint64_t inode, struct blocks *b)
 	enum update_flags update;
 
 	ip = fsck_load_inode(sbp, inode); /* bread, inode_get */
-	log_info("Checking inode %" PRIu64 " (0x%" PRIx64
-			 ")'s metatree for references to block %" PRIu64 " (0x%" PRIx64
-			 ")\n", inode, inode, b->block_no, b->block_no);
+	log_debug( _("Checking inode %" PRIu64 " (0x%" PRIx64 ")'s "
+		     "metatree for references to block %" PRIu64 " (0x%" PRIx64
+		     ")\n"), inode, inode, b->block_no, b->block_no);
 	if(check_metatree(ip, &find_refs)) {
 		stack;
 		fsck_inode_put(ip, not_updated); /* out, brelse, free */
 		return -1;
 	}
-	log_info("Done checking metatree\n");
+	log_debug( _("Done checking metatree\n"));
 	/* Check for ea references in the inode */
 	if(check_inode_eattr(ip, &update, &find_refs) < 0){
 		stack;
@@ -371,15 +357,15 @@ int find_block_ref(struct gfs2_sbd *sbp, uint64_t inode, struct blocks *b)
 	}
 	if (myfi.found) {
 		if(!(id = malloc(sizeof(*id)))) {
-			log_crit("Unable to allocate inode_with_dups structure\n");
+			log_crit( _("Unable to allocate inode_with_dups structure\n"));
 			return -1;
 		}
 		if(!(memset(id, 0, sizeof(*id)))) {
-			log_crit("Unable to zero inode_with_dups structure\n");
+			log_crit( _("Unable to zero inode_with_dups structure\n"));
 			return -1;
 		}
-		log_debug("Found %d entries with block %" PRIu64
-				  " (0x%" PRIx64 ") in inode #%" PRIu64 " (0x%" PRIx64 ")\n",
+		log_debug( _("Found %d entries with block %" PRIu64
+				  " (0x%" PRIx64 ") in inode #%" PRIu64 " (0x%" PRIx64 ")\n"),
 				  myfi.found, b->block_no, b->block_no, inode, inode);
 		id->dup_count = myfi.found;
 		id->block_no = inode;
@@ -390,7 +376,7 @@ int find_block_ref(struct gfs2_sbd *sbp, uint64_t inode, struct blocks *b)
 	return 0;
 }
 
-int handle_dup_blk(struct gfs2_sbd *sbp, struct blocks *b)
+static int handle_dup_blk(struct gfs2_sbd *sbp, struct dup_blocks *b)
 {
 	osi_list_t *tmp;
 	struct inode_with_dups *id;
@@ -414,18 +400,72 @@ int handle_dup_blk(struct gfs2_sbd *sbp, struct blocks *b)
 		dh.ref_inode_count++;
 		dh.ref_count += id->dup_count;
 	}
-	log_notice("Block %" PRIu64 " (0x%" PRIx64 ") has %d inodes referencing it"
-			   " for a total of %d duplicate references\n",
-			   b->block_no, b->block_no, dh.ref_inode_count,
-			   dh.ref_inode_count, dh.ref_count);
+	/* A single reference to the block implies a possible situation where
+	   a data pointer points to a metadata block.  In other words, the
+	   duplicate reference in the file system is (1) Metadata block X and
+	   (2) A dinode reference such as a data pointer pointing to block X.
+	   We can't really check for that in pass1 because user data might
+	   just _look_ like metadata by coincidence, and at the time we're
+	   checking, we might not have processed the referenced block.
+	   Here in pass1b we're sure. */
+	if (dh.ref_count == 1) {
+		struct gfs2_buffer_head *bh;
+		uint32_t cmagic;
+
+		bh = bread(&sbp->buf_list, b->block_no);
+		cmagic = ((struct gfs2_meta_header *)(bh->b_data))->mh_magic;
+		brelse(bh, not_updated);
+		if (be32_to_cpu(cmagic) == GFS2_MAGIC) {
+			tmp = b->ref_inode_list.next;
+			id = osi_list_entry(tmp, struct inode_with_dups, list);
+			log_warn( _("Inode %s (%lld/0x%llx) has a reference to"
+				    " data block %llu (0x%llx), "
+				    "but the block is really metadata.\n"),
+				  id->name, (unsigned long long)id->block_no,
+				  (unsigned long long)id->block_no,
+				  (unsigned long long)b->block_no,
+				  (unsigned long long)b->block_no);
+			errors_found++;
+			if ((errors_corrected +=
+			     query(&opts, _("Clear the inode? (y/n) ")))) {
+				log_warn( _("Clearing...\n"));
+				ip = fsck_load_inode(sbp, id->block_no);
+				inode_hash_remove(inode_hash,
+						  ip->i_di.di_num.no_addr);
+				/* Setting the block to invalid means the inode
+				   is cleared in pass2 */
+				gfs2_block_set(ip->i_sbd, bl,
+					       ip->i_di.di_num.no_addr,
+					       gfs2_meta_inval);
+				fsck_inode_put(ip, updated);
+			} else {
+				log_warn( _("The bad inode was not cleared."));
+			}
+			return 0;
+		}
+	}
+
+	log_notice( _("Block %llu (0x%llx) has %d inodes referencing it"
+		   " for a total of %d duplicate references\n"),
+		   (unsigned long long)b->block_no,
+		   (unsigned long long)b->block_no,
+		   dh.ref_inode_count, dh.ref_count);
 
 	osi_list_foreach(tmp, &b->ref_inode_list) {
 		id = osi_list_entry(tmp, struct inode_with_dups, list);
-		log_warn("Inode %s has %d reference(s) to block %"PRIu64
-				 " (0x%" PRIx64 ")\n", id->name, id->dup_count, b->block_no,
-				 b->block_no);
-		/* FIXME: User input */
-		log_warn("Clearing...\n");
+		log_warn( _("Inode %s (%lld/0x%llx) has %d reference(s) to "
+			    "block %llu (0x%llx)\n"), id->name,
+			  (unsigned long long)id->block_no,
+			  (unsigned long long)id->block_no,
+			  id->dup_count, (unsigned long long)b->block_no,
+			  (unsigned long long)b->block_no);
+		errors_found++;
+		if (!(errors_corrected +=
+		      query(&opts, _("Clear the inode? (y/n) ")))) {
+			log_warn( _("The bad inode was not cleared...\n"));
+			continue;
+		}
+		log_warn( _("Clearing...\n"));
 		ip = fsck_load_inode(sbp, id->block_no);
 		dh.b = b;
 		dh.id = id;
@@ -436,7 +476,9 @@ int handle_dup_blk(struct gfs2_sbd *sbp, struct blocks *b)
 		if(!id->ea_only)
 			check_metatree(ip, &clear_dup_fxns);
 
-		fsck_inode_put(ip, not_updated); /* out, brelse, free */
+		gfs2_block_set(ip->i_sbd, bl, ip->i_di.di_num.no_addr,
+			       gfs2_meta_inval);
+		fsck_inode_put(ip, updated); /* out, brelse, free */
 		dh.ref_inode_count--;
 		if(dh.ref_inode_count == 1)
 			break;
@@ -454,36 +496,36 @@ int handle_dup_blk(struct gfs2_sbd *sbp, struct blocks *b)
  * use in pass2 */
 int pass1b(struct gfs2_sbd *sbp)
 {
-	struct blocks *b;
+	struct dup_blocks *b;
 	uint64_t i;
 	struct gfs2_block_query q;
-	osi_list_t *tmp = NULL;
+	osi_list_t *tmp = NULL, *x;
 	struct metawalk_fxns find_dirents = {0};
+	int rc = FSCK_OK;
 	find_dirents.check_dentry = &find_dentry;
-	int rc = 0;
 
-	log_info("Looking for duplicate blocks...\n");
+	log_info( _("Looking for duplicate blocks...\n"));
 
 	/* If there were no dups in the bitmap, we don't need to do anymore */
 	if(osi_list_empty(&sbp->dup_blocks.list)) {
-		log_info("No duplicate blocks found\n");
-		return 0;
+		log_info( _("No duplicate blocks found\n"));
+		return FSCK_OK;
 	}
 
 	/* Rescan the fs looking for pointers to blocks that are in
 	 * the duplicate block map */
-	log_info("Scanning filesystem for inodes containing duplicate blocks...\n");
-	log_debug("Filesystem has %"PRIu64" (0x%" PRIx64 ") blocks total\n",
+	log_info( _("Scanning filesystem for inodes containing duplicate blocks...\n"));
+	log_debug( _("Filesystem has %"PRIu64" (0x%" PRIx64 ") blocks total\n"),
 			  last_fs_block, last_fs_block);
 	for(i = 0; i < last_fs_block; i += 1) {
 		warm_fuzzy_stuff(i);
 		if (skip_this_pass || fsck_abort) /* if asked to skip the rest */
 			goto out;
-		log_debug("Scanning block %" PRIu64 " (0x%" PRIx64 ") for inodes\n",
+		log_debug( _("Scanning block %" PRIu64 " (0x%" PRIx64 ") for inodes\n"),
 				  i, i);
 		if(gfs2_block_check(sbp, bl, i, &q)) {
 			stack;
-			rc = -1;
+			rc = FSCK_ERROR;
 			goto out;
 		}
 		if((q.block_type == gfs2_inode_dir) ||
@@ -493,11 +535,12 @@ int pass1b(struct gfs2_sbd *sbp)
 		   (q.block_type == gfs2_inode_chr) ||
 		   (q.block_type == gfs2_inode_fifo) ||
 		   (q.block_type == gfs2_inode_sock)) {
-			osi_list_foreach(tmp, &sbp->dup_blocks.list) {
-				b = osi_list_entry(tmp, struct blocks, list);
+			osi_list_foreach_safe(tmp, &sbp->dup_blocks.list, x) {
+				b = osi_list_entry(tmp, struct dup_blocks,
+						   list);
 				if(find_block_ref(sbp, i, b)) {
 					stack;
-					rc = -1;
+					rc = FSCK_ERROR;
 					goto out;
 				}
 			}
@@ -509,11 +552,11 @@ int pass1b(struct gfs2_sbd *sbp)
 
 	/* Fix dups here - it's going to slow things down a lot to fix
 	 * it later */
-	log_info("Handling duplicate blocks\n");
+	log_info( _("Handling duplicate blocks\n"));
 out:
 	while (!osi_list_empty(&sbp->dup_blocks.list)) {
-		b = osi_list_entry(&sbp->dup_blocks.list.next, struct blocks,
-				   list);
+		b = osi_list_entry(sbp->dup_blocks.list.next,
+				   struct dup_blocks, list);
 		if (!skip_this_pass && !rc) /* no error & not asked to skip the rest */
 			handle_dup_blk(sbp, b);
 		osi_list_del(&b->list);
