@@ -4,6 +4,8 @@
 #include <string.h>
 #include <inttypes.h>
 #include <sys/stat.h>
+#include <libintl.h>
+#define _(String) gettext(String)
 
 #include "libgfs2.h"
 #include "fsck.h"
@@ -16,7 +18,7 @@
 
 /* Set children's parent inode in dir_info structure - ext2 does not set
  * dotdot inode here, but instead in pass3 - should we? */
-int set_parent_dir(struct gfs2_sbd *sbp, uint64_t childblock,
+static int set_parent_dir(struct gfs2_sbd *sbp, uint64_t childblock,
 				   uint64_t parentblock)
 {
 	struct dir_info *di;
@@ -24,9 +26,9 @@ int set_parent_dir(struct gfs2_sbd *sbp, uint64_t childblock,
 	if(!find_di(sbp, childblock, &di)) {
 		if(di->dinode == childblock) {
 			if (di->treewalk_parent) {
-				log_err("Another directory at block %" PRIu64
+				log_err( _("Another directory at block %" PRIu64
 						" (0x%" PRIx64 ") already contains"
-						" this child - checking %" PRIu64 " (0x%" PRIx64 ")\n",
+						" this child - checking %" PRIu64 " (0x%" PRIx64 ")\n"),
 						di->treewalk_parent, di->treewalk_parent,
 						parentblock, parentblock);
 				return 1;
@@ -34,8 +36,8 @@ int set_parent_dir(struct gfs2_sbd *sbp, uint64_t childblock,
 			di->treewalk_parent = parentblock;
 		}
 	} else {
-		log_err("Unable to find block %"PRIu64" (0x%" PRIx64
-				") in dir_info list\n",	childblock,	childblock);
+		log_err( _("Unable to find block %"PRIu64" (0x%" PRIx64
+				") in dir_info list\n"),	childblock,	childblock);
 		return -1;
 	}
 
@@ -43,7 +45,7 @@ int set_parent_dir(struct gfs2_sbd *sbp, uint64_t childblock,
 }
 
 /* Set's the child's '..' directory inode number in dir_info structure */
-int set_dotdot_dir(struct gfs2_sbd *sbp, uint64_t childblock,
+static int set_dotdot_dir(struct gfs2_sbd *sbp, uint64_t childblock,
 				   uint64_t parentblock)
 {
 	struct dir_info *di;
@@ -55,17 +57,17 @@ int set_dotdot_dir(struct gfs2_sbd *sbp, uint64_t childblock,
 			if(di->dotdot_parent && sbp->md.rooti->i_di.di_num.no_addr
 			   != di->dinode) {
 				/* This should never happen */
-				log_crit("Dotdot parent already set for"
+				log_crit( _("Dotdot parent already set for"
 						 " block %"PRIu64" (0x%" PRIx64 ") -> %" PRIu64
-						 " (0x%" PRIx64 ")\n", childblock, childblock,
+						 " (0x%" PRIx64 ")\n"), childblock, childblock,
 						 di->dotdot_parent, di->dotdot_parent);
 				return -1;
 			}
 			di->dotdot_parent = parentblock;
 		}
 	} else {
-		log_err("Unable to find block %"PRIu64" (0x%" PRIx64
-				") in dir_info list\n", childblock, childblock);
+		log_err( _("Unable to find block %"PRIu64" (0x%" PRIx64
+				") in dir_info list\n"), childblock, childblock);
 		return -1;
 	}
 
@@ -90,7 +92,7 @@ static int check_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 	return 0;
 }
 
-const char *de_type_string(uint8_t de_type)
+static const char *de_type_string(uint8_t de_type)
 {
 	const char *de_types[15] = {"unknown", "fifo", "chrdev", "invalid",
 								"directory", "invalid", "blkdev", "invalid",
@@ -133,16 +135,48 @@ static int check_file_type(uint8_t de_type, uint8_t block_type)
 			return 1;
 		break;
 	default:
-		log_err("Invalid block type\n");
+		log_err( _("Invalid block type\n"));
 		return -1;
 		break;
 	}
 	return 0;
 }
 
+static int delete_blocks(struct gfs2_inode *ip, uint64_t block,
+			 struct gfs2_buffer_head **bh, void *private)
+{
+	struct gfs2_block_query q = {0};
+
+	if (gfs2_check_range(ip->i_sbd, block) == 0) {
+		log_info(_("Deleting block %lld (0x%llx) as part of inode "
+			   "%lld (0x%llx)\n"), (unsigned long long)block,
+			 (unsigned long long)block,
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr);
+		if (gfs2_block_check(ip->i_sbd, bl, block, &q))
+			return 0;
+		if (!q.dup_block) {
+			gfs2_block_set(ip->i_sbd, bl, block, gfs2_block_free);
+			gfs2_free_block(ip->i_sbd, block);
+		}
+	}
+	return 0;
+}
+
+static int delete_data(struct gfs2_inode *ip, uint64_t block, void *private)
+{
+	return delete_blocks(ip, block, NULL, private);
+}
+
+struct metawalk_fxns pass2_fxns_delete = {
+	.private = NULL,
+	.check_metalist = delete_blocks,
+	.check_data = delete_data,
+};
+
 /* FIXME: should maybe refactor this a bit - but need to deal with
  * FIXMEs internally first */
-int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
+static int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 		 struct gfs2_dirent *prev_de,
 		 struct gfs2_buffer_head *bh, char *filename,
 		 enum update_flags *update, uint16_t *count, void *priv)
@@ -158,7 +192,6 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 	struct gfs2_dirent dentry, *de;
 	uint32_t calculated_hash;
 
-	*update = not_updated;
 	memset(&dentry, 0, sizeof(struct gfs2_dirent));
 	gfs2_dirent_in(&dentry, (char *)dent);
 	de = &dentry;
@@ -172,9 +205,9 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 
 	/* Start of checks */
 	if (de->de_rec_len < GFS2_DIRENT_SIZE(de->de_name_len)){
-		log_err("Dir entry with bad record or name length\n"
+		log_err( _("Dir entry with bad record or name length\n"
 			"\tRecord length = %u\n"
-			"\tName length = %u\n",
+			"\tName length = %u\n"),
 			de->de_rec_len,
 			de->de_name_len);
 		gfs2_block_set(sbp, bl, ip->i_di.di_num.no_addr,
@@ -186,21 +219,23 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 	
 	calculated_hash = gfs2_disk_hash(filename, de->de_name_len);
 	if (de->de_hash != calculated_hash){
-	        log_err("Dir entry with bad hash or name length\n"
+	        log_err( _("Dir entry with bad hash or name length\n"
 					"\tHash found         = %u (0x%x)\n"
-					"\tFilename           = %s\n", de->de_hash, de->de_hash,
+					"\tFilename           = %s\n"), de->de_hash, de->de_hash,
 					filename);
-			log_err("\tName length found  = %u\n"
-					"\tHash expected      = %u (0x%x)\n",
+			log_err( _("\tName length found  = %u\n"
+					"\tHash expected      = %u (0x%x)\n"),
 					de->de_name_len, calculated_hash, calculated_hash);
-			if(query(&opts, "Fix directory hash for %s? (y/n) ",
+			errors_found++;
+			if(query(&opts, _("Fix directory hash for %s? (y/n) "),
 					 filename)) {
+				errors_corrected++;
 				de->de_hash = calculated_hash;
 				gfs2_dirent_out(de, (char *)dent);
-				log_err("Directory entry hash for %s fixed.\n", filename);
+				log_err( _("Directory entry hash for %s fixed.\n"), filename);
 			}
 			else {
-				log_err("Directory entry hash for %s not fixed.\n", filename);
+				log_err( _("Directory entry hash for %s not fixed.\n"), filename);
 				return 1;
 			}
 	}
@@ -213,16 +248,18 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 		strncpy(tmp_name, filename, MAX_FILENAME - 1);
 
 	if(gfs2_check_range(ip->i_sbd, entryblock)) {
-		log_err("Block # referenced by directory entry %s is out of range\n",
+		log_err( _("Block # referenced by directory entry %s is out of range\n"),
 				tmp_name);
+		errors_found++;
 		if(query(&opts, 
-				 "Clear directory entry tp out of range block? (y/n) ")) {
-			log_err("Clearing %s\n", tmp_name);
+				 _("Clear directory entry tp out of range block? (y/n) "))) {
+			errors_corrected++;
+			log_err( _("Clearing %s\n"), tmp_name);
 			dirent2_del(ip, bh, prev_de, dent);
 			*update = updated;
 			return 1;
 		} else {
-			log_err("Directory entry to out of range block remains\n");
+			log_err( _("Directory entry to out of range block remains\n"));
 			(*count)++;
 			ds->entry_count++;
 			return 0;
@@ -236,20 +273,17 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 	if(q.bad_block) {
 		/* This entry's inode has bad blocks in it */
 
-		/* FIXME: user interface */
-		/* FIXME: do i want to kill the inode here? */
 		/* Handle bad blocks */
-		log_err("Found a bad directory entry: %s\n", filename);
-
-		if(query(&opts, "Clear entry to inode containing bad blocks? (y/n)")) {
+		log_err( _("Found a bad directory entry: %s\n"), filename);
 
+		errors_found++;
+		if(query(&opts, _("Clear inode containing bad blocks? (y/n)"))) {
+			errors_corrected++;
 			entry_ip = fsck_load_inode(sbp, de->de_inum.no_addr);
 			check_inode_eattr(entry_ip, update, &clear_eattrs);
 			fsck_inode_put(entry_ip, not_updated);
 
-			/* FIXME: make sure all blocks referenced by
-			 * this inode are cleared in the bitmap */
-
+			check_metatree(entry_ip, &pass2_fxns_delete);
 			dirent2_del(ip, bh, prev_de, dent);
 
 			gfs2_block_set(sbp, bl, de->de_inum.no_addr,
@@ -257,7 +291,7 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 			*update = updated;
 			return 1;
 		} else {
-			log_warn("Entry to inode containing bad blocks remains\n");
+			log_warn( _("Entry to inode containing bad blocks remains\n"));
 			(*count)++;
 			ds->entry_count++;
 			return 0;
@@ -268,23 +302,48 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 	   q.block_type != gfs2_inode_lnk && q.block_type != gfs2_inode_blk &&
 	   q.block_type != gfs2_inode_chr && q.block_type != gfs2_inode_fifo &&
 	   q.block_type != gfs2_inode_sock) {
-		log_err("Directory entry '%s' at block %" PRIu64 " (0x%" PRIx64
-			") in dir inode %" PRIu64 " (0x%" PRIx64
-			") has an invalid block type: %d.\n", tmp_name,
-			de->de_inum.no_addr, de->de_inum.no_addr,
-			ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr,
-			q.block_type);
+		log_err( _("Directory entry '%s' at block %llu (0x%llx"
+			   ") in dir inode %llu (0x%llx"
+			   ") block type %d: %s.\n"), tmp_name,
+			 (unsigned long long)de->de_inum.no_addr,
+			 (unsigned long long)de->de_inum.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 (unsigned long long)ip->i_di.di_num.no_addr,
+			 q.block_type, q.block_type == gfs2_meta_inval ?
+			 _("previously marked invalid") :
+			 _("is not an inode"));
 
-		if(query(&opts, "Clear directory entry to non-inode block? (y/n) ")) {
-			/* FIXME: make sure all blocks referenced by
-			 * this inode are cleared in the bitmap */
+		errors_found++;
+		if(query(&opts, _("Clear directory entry to non-inode block? "
+				  "(y/n) "))) {
+			struct gfs2_buffer_head *bhi;
 
+			errors_corrected++;
 			dirent2_del(ip, bh, prev_de, dent);
 			*update = updated;
-			log_warn("Directory entry '%s' cleared\n", tmp_name);
+			log_warn( _("Directory entry '%s' cleared\n"), tmp_name);
+			/* If it was previously marked invalid (i.e. known
+			   to be bad, not just a free block, etc.) then
+			   delete any metadata it holds.  If not, return. */
+			if (q.block_type != gfs2_meta_inval)
+				return 1;
+
+			/* Now try to clear the dinode, if it is an dinode */
+			bhi = bread(&sbp->buf_list, de->de_inum.no_addr);
+			error = gfs2_check_meta(bhi, GFS2_METATYPE_DI);
+			brelse(bhi, updated);
+			if (error)
+				return 1; /* not a dinode: nothing to delete */
+
+			entry_ip = fsck_load_inode(sbp, de->de_inum.no_addr);
+			check_metatree(entry_ip, &pass2_fxns_delete);
+			fsck_inode_put(entry_ip, updated);
+			gfs2_block_set(sbp, bl, de->de_inum.no_addr,
+				       gfs2_block_free);
+
 			return 1;
 		} else {
-			log_err("Directory entry to non-inode block remains\n");
+			log_err( _("Directory entry to non-inode block remains\n"));
 			(*count)++;
 			ds->entry_count++;
 			return 0;
@@ -297,21 +356,25 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 		return -1;
 	}
 	if(error > 0) {
-		log_warn("Type '%s' in dir entry (%s, %" PRIu64 "/0x%" PRIx64 ") "
-				 "conflicts with type '%s' in dinode. (Dir entry is stale.)\n",
-				 de_type_string(de->de_type), tmp_name, 
-				 de->de_inum.no_addr, de->de_inum.no_addr,
-				 block_type_string(&q));
-		if(query(&opts, "Clear stale directory entry? (y/n) ")) {
+		log_warn( _("Type '%s' in dir entry (%s, %llu/0x%llx) conflicts"
+			 " with type '%s' in dinode. (Dir entry is stale.)\n"),
+			 de_type_string(de->de_type), tmp_name,
+			 (unsigned long long)de->de_inum.no_addr,
+			 (unsigned long long)de->de_inum.no_addr,
+			 block_type_string(&q));
+		errors_found++;
+		if(query(&opts, _("Clear stale directory entry? (y/n) "))) {
+			errors_corrected++;
 			entry_ip = fsck_load_inode(sbp, de->de_inum.no_addr);
 			check_inode_eattr(entry_ip, update, &clear_eattrs);
 			fsck_inode_put(entry_ip, not_updated);
 
 			dirent2_del(ip, bh, prev_de, dent);
 			*update = updated;
+			log_err( _("Stale directory entry deleted\n"));
 			return 1;
 		} else {
-			log_err("Stale directory entry remains\n");
+			log_err( _("Stale directory entry remains\n"));
 			(*count)++;
 			ds->entry_count++;
 			return 0;
@@ -319,14 +382,17 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 	}
 
 	if(!strcmp(".", tmp_name)) {
-		log_debug("Found . dentry\n");
+		log_debug( _("Found . dentry\n"));
 
 		if(ds->dotdir) {
-			log_err("Already found '.' entry in directory %" PRIu64 " (0x%"
-					PRIx64 ")\n",
-					ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
-			if(query(&opts, "Clear duplicate '.' entry? (y/n) ")) {
+			log_err( _("Already found '.' entry in directory %llu"
+				" (0x%llx)\n"),
+				(unsigned long long)ip->i_di.di_num.no_addr,
+				(unsigned long long)ip->i_di.di_num.no_addr);
+			errors_found++;
+			if(query(&opts, _("Clear duplicate '.' entry? (y/n) "))) {
 
+				errors_corrected++;
 				entry_ip = fsck_load_inode(sbp, de->de_inum.no_addr);
 				check_inode_eattr(entry_ip, update,
 						  &clear_eattrs);
@@ -336,7 +402,7 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 				*update = updated;
 				return 1;
 			} else {
-				log_err("Duplicate '.' entry remains\n");
+				log_err( _("Duplicate '.' entry remains\n"));
 				/* FIXME: Should we continue on here
 				 * and check the rest of the '.'
 				 * entry? */
@@ -352,13 +418,19 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 
 		/* check that '.' refers to this inode */
 		if(de->de_inum.no_addr != ip->i_di.di_num.no_addr) {
-			log_err("'.' entry's value incorrect in directory %" PRIu64
-					" (0x%" PRIx64 ").  Points to %"PRIu64
-					" (0x%" PRIx64 ") when it should point to %" PRIu64
-					" (0x%" PRIx64 ").\n",
-					de->de_inum.no_addr, de->de_inum.no_addr,
-					ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
-			if(query(&opts, "Remove '.' reference? (y/n) ")) {
+			log_err( _("'.' entry's value incorrect in directory %llu"
+				" (0x%llx).  Points to %llu"
+				" (0x%llx) when it should point to %llu"
+				" (0x%llx).\n"),
+				(unsigned long long)de->de_inum.no_addr,
+				(unsigned long long)de->de_inum.no_addr,
+				(unsigned long long)de->de_inum.no_addr,
+				(unsigned long long)de->de_inum.no_addr,
+				(unsigned long long)ip->i_di.di_num.no_addr,
+				(unsigned long long)ip->i_di.di_num.no_addr);
+			errors_found++;
+			if(query(&opts, _("Remove '.' reference? (y/n) "))) {
+				errors_corrected++;
 				entry_ip = fsck_load_inode(sbp, de->de_inum.no_addr);
 				check_inode_eattr(entry_ip, update,
 						  &clear_eattrs);
@@ -369,7 +441,7 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 				return 1;
 
 			} else {
-				log_err("Invalid '.' reference remains\n");
+				log_err( _("Invalid '.' reference remains\n"));
 				/* Not setting ds->dotdir here since
 				 * this '.' entry is invalid */
 				increment_link(sbp, de->de_inum.no_addr);
@@ -387,13 +459,16 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 		return 0;
 	}
 	if(!strcmp("..", tmp_name)) {
-		log_debug("Found .. dentry\n");
+		log_debug( _("Found .. dentry\n"));
 		if(ds->dotdotdir) {
-			log_err("Already found '..' entry in directory %" PRIu64 " (0x%"
-					PRIx64 ")\n",
-					ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
-			if(query(&opts, "Clear duplicate '..' entry? (y/n) ")) {
+			log_err( _("Already found '..' entry in directory %llu"
+				"(0x%llx)\n"),
+				(unsigned long long)ip->i_di.di_num.no_addr,
+				(unsigned long long)ip->i_di.di_num.no_addr);
+			errors_found++;
+			if(query(&opts, _("Clear duplicate '..' entry? (y/n) "))) {
 
+				errors_corrected++;
 				entry_ip = fsck_load_inode(sbp, de->de_inum.no_addr);
 				check_inode_eattr(entry_ip, update,
 						  &clear_eattrs);
@@ -403,7 +478,7 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 				*update = 1;
 				return 1;
 			} else {
-				log_err("Duplicate '..' entry remains\n");
+				log_err( _("Duplicate '..' entry remains\n"));
 				/* FIXME: Should we continue on here
 				 * and check the rest of the '..'
 				 * entry? */
@@ -415,11 +490,13 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 		}
 
 		if(q.block_type != gfs2_inode_dir) {
-			log_err("Found '..' entry  in directory %" PRIu64 " (0x%"
-					PRIx64 ") pointing to"
-					" something that's not a directory",
-					ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
-			if(query(&opts, "Clear bad '..' directory entry? (y/n) ")) {
+			log_err( _("Found '..' entry in directory %llu (0x%llx) "
+				"pointing to something that's not a directory"),
+				(unsigned long long)ip->i_di.di_num.no_addr,
+				(unsigned long long)ip->i_di.di_num.no_addr);
+			errors_found++;
+			if(query(&opts, _("Clear bad '..' directory entry? (y/n) "))) {
+				errors_corrected++;
 				entry_ip = fsck_load_inode(sbp, de->de_inum.no_addr);
 				check_inode_eattr(entry_ip, update,
 						  &clear_eattrs);
@@ -429,7 +506,7 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 				*update = 1;
 				return 1;
 			} else {
-				log_err("Bad '..' directory entry remains\n");
+				log_err( _("Bad '..' directory entry remains\n"));
 				increment_link(sbp, de->de_inum.no_addr);
 				(*count)++;
 				ds->entry_count++;
@@ -458,7 +535,7 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 	/* After this point we're only concerned with
 	 * directories */
 	if(q.block_type != gfs2_inode_dir) {
-		log_debug("Found non-dir inode dentry\n");
+		log_debug( _("Found non-dir inode dentry\n"));
 		increment_link(sbp, de->de_inum.no_addr);
 		*update = (opts.no ? not_updated : updated);
 		(*count)++;
@@ -466,21 +543,23 @@ int check_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 		return 0;
 	}
 
-	log_debug("Found plain directory dentry\n");
+	log_debug( _("Found plain directory dentry\n"));
 	error = set_parent_dir(sbp, entryblock, ip->i_di.di_num.no_addr);
 	if(error > 0) {
-		log_err("%s: Hard link to block %" PRIu64" (0x%" PRIx64
-				") detected.\n", filename, entryblock, entryblock);
+		log_err( _("%s: Hard link to block %" PRIu64" (0x%" PRIx64
+				") detected.\n"), filename, entryblock, entryblock);
 
-		if(query(&opts, "Clear hard link to directory? (y/n) ")) {
+		errors_found++;
+		if(query(&opts, _("Clear hard link to directory? (y/n) "))) {
+			errors_corrected++;
 			*update = 1;
 
 			dirent2_del(ip, bh, prev_de, dent);
-			log_warn("Directory entry %s cleared\n", filename);
+			log_warn( _("Directory entry %s cleared\n"), filename);
 
 			return 1;
 		} else {
-			log_err("Hard link to directory remains\n");
+			log_err( _("Hard link to directory remains\n"));
 			(*count)++;
 			ds->entry_count++;
 			return 0;
@@ -512,8 +591,8 @@ struct metawalk_fxns pass2_fxns = {
 
 /* Check system directory inode                                           */
 /* Should work for all system directories: root, master, jindex, per_node */
-int check_system_dir(struct gfs2_inode *sysinode, const char *dirname,
-		     void builder(struct gfs2_sbd *sbp))
+static int check_system_dir(struct gfs2_inode *sysinode, const char *dirname,
+		     int builder(struct gfs2_sbd *sbp))
 {
 	uint64_t iblock = 0;
 	struct dir_status ds = {0};
@@ -524,7 +603,7 @@ int check_system_dir(struct gfs2_inode *sysinode, const char *dirname,
 	enum update_flags update = not_updated;
 	int error = 0;
 
-	log_info("Checking system directory inode '%s'\n", dirname);
+	log_info( _("Checking system directory inode '%s'\n"), dirname);
 
 	if (sysinode) {
 		iblock = sysinode->i_di.di_num.no_addr;
@@ -554,47 +633,57 @@ int check_system_dir(struct gfs2_inode *sysinode, const char *dirname,
 		return -1;
 	}
 	if(!ds.dotdir) {
-		log_err("No '.' entry found for %s directory.\n", dirname);
-		sprintf(tmp_name, ".");
-		filename_len = strlen(tmp_name);  /* no trailing NULL */
-		if(!(filename = malloc(sizeof(char) * filename_len))) {
-			log_err("Unable to allocate name string\n");
-			stack;
-			return -1;
-		}
-		if(!(memset(filename, 0, sizeof(char) * filename_len))) {
-			log_err("Unable to zero name string\n");
-			stack;
-			return -1;
-		}
-		memcpy(filename, tmp_name, filename_len);
-		log_warn("Adding '.' entry\n");
-		dir_add(sysinode, filename, filename_len,
+		log_err( _("No '.' entry found for %s directory.\n"), dirname);
+		errors_found++;
+		if (query(&opts, _("Is it okay to add '.' entry? (y/n) "))) {
+			errors_corrected++;
+			sprintf(tmp_name, ".");
+			filename_len = strlen(tmp_name); /* no trailing NULL */
+			if(!(filename = malloc(sizeof(char) * filename_len))) {
+				log_err( _("Unable to allocate name string\n"));
+				stack;
+				return -1;
+			}
+			if(!(memset(filename, 0, sizeof(char) *
+				    filename_len))) {
+				log_err( _("Unable to zero name string\n"));
+				stack;
+				return -1;
+			}
+			memcpy(filename, tmp_name, filename_len);
+			log_warn( _("Adding '.' entry\n"));
+			dir_add(sysinode, filename, filename_len,
 				&(sysinode->i_di.di_num), DT_DIR);
-		increment_link(sysinode->i_sbd,
-					   sysinode->i_di.di_num.no_addr);
-		ds.entry_count++;
-		free(filename);
-		update = 1;
+			increment_link(sysinode->i_sbd,
+				       sysinode->i_di.di_num.no_addr);
+			ds.entry_count++;
+			free(filename);
+			update = 1;
+		} else
+			log_err( _("The directory was not fixed.\n"));
 	}
 	if(sysinode->i_di.di_entries != ds.entry_count) {
-		log_err("%s inode %" PRIu64 " (0x%" PRIx64
-			"): Entries is %d - should be %d\n", dirname,
-			sysinode->i_di.di_num.no_addr,
-			sysinode->i_di.di_num.no_addr,
+		log_err( _("%s inode %llu (0x%llx"
+			"): Entries is %d - should be %d\n"), dirname,
+			(unsigned long long)sysinode->i_di.di_num.no_addr,
+			(unsigned long long)sysinode->i_di.di_num.no_addr,
 			sysinode->i_di.di_entries, ds.entry_count);
-		if(query(&opts, "Fix entries for %s inode %" PRIu64 " (0x%"
-			 PRIx64 ")? (y/n) ", dirname,
-			 sysinode->i_di.di_num.no_addr,
-			 sysinode->i_di.di_num.no_addr)) {
+		errors_found++;
+		if(query(&opts, _("Fix entries for %s inode %llu (0x%llx"
+			 ")? (y/n) "), dirname,
+			 (unsigned long long)sysinode->i_di.di_num.no_addr,
+			 (unsigned long long)sysinode->i_di.di_num.no_addr)) {
+			errors_corrected++;
 			sysinode->i_di.di_entries = ds.entry_count;
-			log_warn("Entries updated\n");
+			log_warn( _("Entries updated\n"));
 			update = 1;
 		} else {
-			log_err("Entries for inode %" PRIu64 " (0x%" PRIx64
-					") left out of sync\n",
-					sysinode->i_di.di_num.no_addr,
-					sysinode->i_di.di_num.no_addr);
+			log_err( _("Entries for inode %llu (0x%llx"
+				") left out of sync\n"),
+				(unsigned long long)
+				sysinode->i_di.di_num.no_addr,
+				(unsigned long long)
+				sysinode->i_di.di_num.no_addr);
 		}
 	}
 
@@ -637,46 +726,47 @@ int pass2(struct gfs2_sbd *sbp)
 	int filename_len;
 	char tmp_name[256];
 	int error = 0;
+	enum update_flags need_update = NOT_UPDATED;
 
 	/* Check all the system directory inodes. */
 	if (check_system_dir(sbp->md.jiinode, "jindex", build_jindex)) {
 		stack;
-		return -1;
+		return FSCK_ERROR;
 	}
 	if (check_system_dir(sbp->md.pinode, "per_node", build_per_node)) {
 		stack;
-		return -1;
+		return FSCK_ERROR;
 	}
 	if (check_system_dir(sbp->master_dir, "master", build_master)) {
 		stack;
-		return -1;
+		return FSCK_ERROR;
 	}
 	if (check_system_dir(sbp->md.rooti, "root", build_root)) {
 		stack;
-		return -1;
+		return FSCK_ERROR;
 	}
-	log_info("Checking directory inodes.\n");
+	log_info( _("Checking directory inodes.\n"));
 	/* Grab each directory inode, and run checks on it */
 	for(i = 0; i < last_fs_block; i++) {
 		warm_fuzzy_stuff(i);
 		if (skip_this_pass || fsck_abort) /* if asked to skip the rest */
-			return 0;
+			return FSCK_OK;
 
 		/* Skip the system inodes - they're checked above */
 		if (is_system_dir(sbp, i))
 			continue;
 
 		if(gfs2_block_check(sbp, bl, i, &q)) {
-			log_err("Can't get block %"PRIu64 " (0x%" PRIx64
-					") from block list\n", i, i);
-			return -1;
+			log_err( _("Can't get block %"PRIu64 " (0x%" PRIx64
+					") from block list\n"), i, i);
+			return FSCK_ERROR;
 		}
 
 		if(q.block_type != gfs2_inode_dir)
 			continue;
 
-		log_debug("Checking directory inode at block %"PRIu64" (0x%"
-				  PRIx64 ")\n", i, i);
+		log_debug( _("Checking directory inode at block %"PRIu64" (0x%"
+				  PRIx64 ")\n"), i, i);
 
 		memset(&ds, 0, sizeof(ds));
 		pass2_fxns.private = (void *) &ds;
@@ -687,86 +777,108 @@ int pass2(struct gfs2_sbd *sbp)
 			if(check_metatree(ip, &pass2_fxns)) {
 				stack;
 				free(ip);
-				return -1;
+				return FSCK_ERROR;
 			}
 			fsck_inode_put(ip, not_updated);
 		}
 		error = check_dir(sbp, i, &pass2_fxns);
 		if(error < 0) {
 			stack;
-			return -1;
+			return FSCK_ERROR;
 		}
 		if (error > 0) {
 			struct dir_info *di = NULL;
 			error = find_di(sbp, i, &di);
 			if(error < 0) {
 				stack;
-				return -1;
+				return FSCK_ERROR;
 			}
 			if(error == 0) {
 				/* FIXME: factor */
-				if(query(&opts, "Remove directory entry for bad"
+				errors_found++;
+				if(query(&opts, _("Remove directory entry for bad"
 						 " inode %"PRIu64" (0x%" PRIx64 ") in %"PRIu64
-						 " (0x%" PRIx64 ")? (y/n)", i, i, di->treewalk_parent,
+						 " (0x%" PRIx64 ")? (y/n)"), i, i, di->treewalk_parent,
 						 di->treewalk_parent)) {
+					errors_corrected++;
 					error = remove_dentry_from_dir(sbp, di->treewalk_parent,
 												   i);
 					if(error < 0) {
 						stack;
-						return -1;
+						return FSCK_ERROR;
 					}
 					if(error > 0) {
-						log_warn("Unable to find dentry for %"
+						log_warn( _("Unable to find dentry for %"
 								 PRIu64 " (0x%" PRIx64 ") in %" PRIu64
-								 " (0x%" PRIx64 ")\n", i, i,
+								 " (0x%" PRIx64 ")\n"), i, i,
 								 di->treewalk_parent, di->treewalk_parent);
 					}
-					log_warn("Directory entry removed\n");
+					log_warn( _("Directory entry removed\n"));
 				} else
-					log_err("Directory entry to invalid inode remains.\n");
+					log_err( _("Directory entry to invalid inode remains.\n"));
 			}
 			gfs2_block_set(sbp, bl, i, gfs2_meta_inval);
 		}
 		bh = bread(&sbp->buf_list, i);
 		ip = fsck_inode_get(sbp, bh);
 		if(!ds.dotdir) {
-			log_err("No '.' entry found\n");
-			sprintf(tmp_name, ".");
-			filename_len = strlen(tmp_name);  /* no trailing NULL */
-			if(!(filename = malloc(sizeof(char) * filename_len))) {
-				log_err("Unable to allocate name string\n");
-				stack;
-				return -1;
-			}
-			if(!memset(filename, 0, sizeof(char) * filename_len)) {
-				log_err("Unable to zero name string\n");
-				stack;
-				return -1;
-			}
-			memcpy(filename, tmp_name, filename_len);
+			log_err(_("No '.' entry found for directory inode at "
+				  "block %"PRIu64" (0x%" PRIx64 ")\n"), i, i);
 
-			dir_add(ip, filename, filename_len, &(ip->i_di.di_num), DT_DIR);
-			increment_link(ip->i_sbd, ip->i_di.di_num.no_addr);
-			ds.entry_count++;
-			free(filename);
+			errors_found++;
+			if (query(&opts,
+				  _("Is it okay to add '.' entry? (y/n) "))) {
+				errors_corrected++;
+				sprintf(tmp_name, ".");
+				filename_len = strlen(tmp_name); /* no trailing
+								    NULL */
+				if(!(filename = malloc(sizeof(char) *
+						       filename_len))) {
+					log_err(_("Unable to allocate name\n"));
+					stack;
+					return FSCK_ERROR;
+				}
+				if(!memset(filename, 0, sizeof(char) *
+					   filename_len)) {
+					log_err( _("Unable to zero name\n"));
+					stack;
+					return FSCK_ERROR;
+				}
+				memcpy(filename, tmp_name, filename_len);
 
+				dir_add(ip, filename, filename_len,
+					&(ip->i_di.di_num), DT_DIR);
+				increment_link(ip->i_sbd,
+					       ip->i_di.di_num.no_addr);
+				ds.entry_count++;
+				free(filename);
+				log_err( _("The directory was fixed.\n"));
+				need_update = UPDATED;
+			} else {
+				log_err( _("The directory was not fixed.\n"));
+			}
 		}
-		fsck_inode_put(ip, not_updated); /* does a brelse */
 
-		bh = bread(&sbp->buf_list, i);
-		ip = fsck_inode_get(sbp, bh);
 		if(ip->i_di.di_entries != ds.entry_count) {
-			log_err("Entries is %d - should be %d for inode block %" PRIu64
-					" (0x%" PRIx64 ")\n",
-					ip->i_di.di_entries, ds.entry_count,
-					ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
-			ip->i_di.di_entries = ds.entry_count;
-			fsck_inode_put(ip, updated); /* does a gfs2_dinode_out, brelse */
+			log_err( _("Entries is %d - should be %d for inode "
+				"block %llu (0x%llx)\n"),
+				ip->i_di.di_entries, ds.entry_count,
+				(unsigned long long)ip->i_di.di_num.no_addr,
+				(unsigned long long)ip->i_di.di_num.no_addr);
+			errors_found++;
+			if (query(&opts,
+				  _("Fix the entry count? (y/n) "))) {
+				errors_corrected++;
+				ip->i_di.di_entries = ds.entry_count;
+				need_update = UPDATED;
+			} else {
+				log_err( _("The entry count was not fixed.\n"));
+			}
 		}
-		else
-			fsck_inode_put(ip, not_updated); /* does a brelse */
+		fsck_inode_put(ip, need_update); /* does a gfs2_dinode_out,
+						    brelse */
 	}
-	return 0;
+	return FSCK_OK;
 }
 
 
