@@ -20,31 +20,22 @@
 #include <linux/types.h>
 #include <linux/gfs2_ondisk.h>
 #include "gfs_ondisk.h"
-#include "linux_endian.h"
 
 #define die(fmt, args...) \
 do { \
-	fprintf(stderr, "%s: ", prog_name); \
 	fprintf(stderr, fmt, ##args); \
 	exit(EXIT_FAILURE); \
 } while (0)
 
 #define warn(fmt, args...) \
 do { \
-	fprintf(stderr, "%s: " fmt "\n", prog_name, ##args); \
+	fprintf(stderr, fmt "\n", ##args); \
 } while (0)
 
 #define log_debug(fmt, args...) \
 do { \
 	if (verbose) \
-		printf("%s: " fmt "\n", prog_name, ##args); \
-} while (0)
-
-#define do_lseek(fd, off) \
-do { \
-	if (lseek((fd), (off), SEEK_SET) != (off)) \
-		die("bad seek: %s on line %d of file %s\n", \
-		    strerror(errno),__LINE__, __FILE__); \
+		printf(fmt "\n", ##args); \
 } while (0)
 
 #define do_read(fd, buff, len) \
