@@ -1,8 +1,6 @@
 #include "dlm_daemon.h"
 #include "config.h"
 
-uint32_t cpgname_to_crc(const char *data, int len);
-
 struct protocol_version {
 	uint16_t major;
 	uint16_t minor;
@@ -129,7 +127,7 @@ static void ids_in(struct ls_info *li, struct id_info *ids)
 	}
 }
 
-char *msg_name(int type)
+const char *msg_name(int type)
 {
 	switch (type) {
 	case DLM_MSG_PROTOCOL:
@@ -1175,9 +1173,12 @@ void process_lockspace_changes(void)
 }
 
 static int add_change(struct lockspace *ls,
-		      struct cpg_address *member_list, int member_list_entries,
-		      struct cpg_address *left_list, int left_list_entries,
-		      struct cpg_address *joined_list, int joined_list_entries,
+		      const struct cpg_address *member_list,
+		      size_t member_list_entries,
+		      const struct cpg_address *left_list,
+		      size_t left_list_entries,
+		      const struct cpg_address *joined_list,
+		      size_t joined_list_entries,
 		      struct change **cg_out)
 {
 	struct change *cg;
@@ -1274,7 +1275,7 @@ static int add_change(struct lockspace *ls,
 	return error;
 }
 
-static int we_left(struct cpg_address *left_list, int left_list_entries)
+static int we_left(const struct cpg_address *left_list, size_t left_list_entries)
 {
 	int i;
 
@@ -1285,10 +1286,14 @@ static int we_left(struct cpg_address *left_list, int left_list_entries)
 	return 0;
 }
 
-static void confchg_cb(cpg_handle_t handle, struct cpg_name *group_name,
-		       struct cpg_address *member_list, int member_list_entries,
-		       struct cpg_address *left_list, int left_list_entries,
-		       struct cpg_address *joined_list, int joined_list_entries)
+static void confchg_cb(cpg_handle_t handle,
+		       const struct cpg_name *group_name,
+		       const struct cpg_address *member_list,
+		       size_t member_list_entries,
+		       const struct cpg_address *left_list,
+		       size_t left_list_entries,
+		       const struct cpg_address *joined_list,
+		       size_t joined_list_entries)
 {
 	struct lockspace *ls;
 	struct change *cg;
@@ -1349,8 +1354,10 @@ static void dlm_header_in(struct dlm_header *hd)
 	hd->msgdata     = le32_to_cpu(hd->msgdata);
 }
 
-static void deliver_cb(cpg_handle_t handle, struct cpg_name *group_name,
-		       uint32_t nodeid, uint32_t pid, void *data, int len)
+static void deliver_cb(cpg_handle_t handle,
+		       const struct cpg_name *group_name,
+		       uint32_t nodeid, uint32_t pid,
+		       void *data, size_t len)
 {
 	struct lockspace *ls;
 	struct dlm_header *hd;
@@ -1362,7 +1369,7 @@ static void deliver_cb(cpg_handle_t handle, struct cpg_name *group_name,
 	}
 
 	if (len < sizeof(*hd)) {
-		log_error("deliver_cb short message %d", len);
+		log_error("deliver_cb short message %zd", len);
 		return;
 	}
 
@@ -1945,13 +1952,15 @@ int set_protocol(void)
 	return 0;
 }
 
-static void deliver_cb_daemon(cpg_handle_t handle, struct cpg_name *group_name,
-		uint32_t nodeid, uint32_t pid, void *data, int len)
+static void deliver_cb_daemon(cpg_handle_t handle,
+			      const struct cpg_name *group_name,
+			      uint32_t nodeid, uint32_t pid,
+			      void *data, size_t len)
 {
 	struct dlm_header *hd;
 
 	if (len < sizeof(*hd)) {
-		log_error("deliver_cb short message %d", len);
+		log_error("deliver_cb short message %zd", len);
 		return;
 	}
 
@@ -1967,10 +1976,14 @@ static void deliver_cb_daemon(cpg_handle_t handle, struct cpg_name *group_name,
 	}
 }
 
-static void confchg_cb_daemon(cpg_handle_t handle, struct cpg_name *group_name,
-		struct cpg_address *member_list, int member_list_entries,
-		struct cpg_address *left_list, int left_list_entries,
-		struct cpg_address *joined_list, int joined_list_entries)
+static void confchg_cb_daemon(cpg_handle_t handle,
+			      const struct cpg_name *group_name,
+			      const struct cpg_address *member_list,
+			      size_t member_list_entries,
+			      const struct cpg_address *left_list,
+			      size_t left_list_entries,
+			      const struct cpg_address *joined_list,
+			      size_t joined_list_entries)
 {
 	int i;
 
@@ -2051,12 +2064,15 @@ int setup_cpg(void)
 
 void close_cpg(void)
 {
+	struct lockspace *ls;
 	cpg_error_t error;
 	struct cpg_name name;
 	int i = 0;
 
-	if (!daemon_cpg_handle || cluster_down)
+	if (!daemon_cpg_handle)
 		return;
+	if (cluster_down)
+		goto fin;
 
 	memset(&name, 0, sizeof(name));
 	sprintf(name.value, "dlm:controld");
@@ -2072,6 +2088,12 @@ void close_cpg(void)
 	}
 	if (error != CPG_OK)
 		log_error("daemon cpg_leave error %d", error);
+ fin:
+	list_for_each_entry(ls, &lockspaces, list) {
+		if (ls->cpg_handle)
+			cpg_finalize(ls->cpg_handle);
+	}
+	cpg_finalize(daemon_cpg_handle);
 }
 
 /* fs_controld has seen nodedown for nodeid; it's now ok for dlm to do
