@@ -1,25 +1,3 @@
-/******************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) 2004-2005 Red Hat, Inc.  All rights reserved.
-**
-**  This library is free software; you can redistribute it and/or
-**  modify it under the terms of the GNU Lesser General Public
-**  License as published by the Free Software Foundation; either
-**  version 2 of the License, or (at your option) any later version.
-**
-**  This library is distributed in the hope that it will be useful,
-**  but WITHOUT ANY WARRANTY; without even the implied warranty of
-**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-**  Lesser General Public License for more details.
-**
-**  You should have received a copy of the GNU Lesser General Public
-**  License along with this library; if not, write to the Free Software
-**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-**
-*******************************************************************************
-******************************************************************************/
-
 #include <sys/types.h>
 #include <sys/un.h>
 #include <inttypes.h>
@@ -153,7 +131,7 @@ static int do_write(int fd, void *buf, size_t count)
 	int rv, off = 0;
 
  retry:
-	rv = write(fd, buf + off, count);
+	rv = write(fd, (char *)buf + off, count);
 	if (rv == -1 && errno == EINTR)
 		goto retry;
 	if (rv < 0)
@@ -172,7 +150,7 @@ static int do_read(int fd, void *buf, size_t count)
 	int rv, off = 0;
 
 	while (off < count) {
-		rv = read(fd, buf + off, count - off);
+		rv = read(fd, (char *)buf + off, count - off);
 		if (rv == 0)
 			return -1;
 		if (rv == -1 && errno == EINTR)
@@ -194,7 +172,7 @@ struct group_handle
 	char prog_name[32];
 };
 
-static int _joinleave(group_handle_t handle, char *name, char *cmd)
+static int _joinleave(group_handle_t handle, const char *name, const char *cmd)
 {
 	char buf[GROUPD_MSGLEN];
 	struct group_handle *h = (struct group_handle *) handle;
@@ -281,7 +259,7 @@ static int connect_groupd(void)
 	return fd;
 }
 
-group_handle_t group_init(void *private, char *prog_name, int level,
+group_handle_t group_init(void *private, const char *prog_name, int level,
 			  group_callbacks_t *cbs, int timeout)
 {
 	struct group_handle *h;
@@ -519,3 +497,28 @@ int group_get_group(int level, const char *name, group_data_t *data)
 	return rv;
 }
 
+int group_get_version(int *version)
+{
+	char buf[GROUPD_MSGLEN];
+	int fd, rv;
+
+	fd = connect_groupd();
+	if (fd < 0)
+		 return fd;
+
+	memset(buf, 0, sizeof(buf));
+	snprintf(buf, sizeof(buf), "get_version");
+
+	rv = do_write(fd, &buf, GROUPD_MSGLEN);
+	if (rv < 0)
+		 goto out;
+
+	rv = do_read(fd, version, sizeof(int));
+	if (rv < 0)
+		 goto out;
+	rv = 0;
+ out:
+	close(fd);
+	return rv;
+}
+
