@@ -118,21 +118,6 @@ typedef struct _rg_node {
 	int	rn_pad;
 } resource_node_t;
 
-typedef struct _fod_node {
-	list_head();
-	char	*fdn_name;
-	int	fdn_prio;
-	int	fdn_nodeid;
-} fod_node_t;
-
-typedef struct _fod {
-	list_head();
-	char	*fd_name;
-	fod_node_t	*fd_nodes;
-	int	fd_flags;
-	int	_pad_; /* align */
-} fod_t;
-
 
 /*
    Exported Functions
@@ -140,6 +125,7 @@ typedef struct _fod {
 int res_start(resource_node_t **tree, resource_t *res, void *ret);
 int res_stop(resource_node_t **tree, resource_t *res, void *ret);
 int res_status(resource_node_t **tree, resource_t *res, void *ret);
+int res_status_inquiry(resource_node_t **tree, resource_t *res, void *ret);
 int res_condstart(resource_node_t **tree, resource_t *res, void *ret);
 int res_condstop(resource_node_t **tree, resource_t *res, void *ret);
 int res_exec(resource_node_t *node, int op, const char *arg, int depth);
@@ -177,30 +163,23 @@ int build_resource_tree(int ccsfd, resource_node_t **tree,
 void print_resource_tree(resource_node_t **tree);
 void destroy_resource_tree(resource_node_t **tree);
 
-/*
-   Construct/deconstruct failover domains
- */
-int construct_domains(int ccsfd, fod_t **domains);
-void deconstruct_domains(fod_t **domains);
-void print_domains(fod_t **domains);
-int node_should_start(int nodeid, cluster_member_list_t *membership,
-		      char *rg_name, fod_t **domains);
-int node_domain_set(fod_t **domains, char *name, int **ret, int *retlen, int *flags);
-int node_domain_set_safe(char *domainname, int **ret, int *retlen, int *flags);
-
+void *act_dup(resource_act_t *acts);
 
 /*
    Handy functions
  */
-resource_t *find_resource_by_ref(resource_t **reslist, char *type, char *ref);
-resource_t *find_root_by_ref(resource_t **reslist, char *ref);
-resource_rule_t *find_rule_by_type(resource_rule_t **rulelist, char *type);
+resource_t *find_resource_by_ref(resource_t **reslist, const char *type,
+				 const char *ref);
+resource_t *find_root_by_ref(resource_t **reslist, const char *ref);
+resource_rule_t *find_rule_by_type(resource_rule_t **rulelist,
+				   const char *type);
 void res_build_name(char *, size_t, resource_t *);
 
 /*
    Internal functions; shouldn't be needed.
  */
-char *xpath_get_one(xmlDocPtr doc, xmlXPathContextPtr ctx, char *query);
+char *xpath_get_one(xmlDocPtr doc, xmlXPathContextPtr ctx,
+			  const char *query);
 int store_attribute(resource_attr_t **attrsp, char *name, char *value,
 		    int flags);
 
@@ -208,15 +187,10 @@ resource_t *load_resource(int ccsfd, resource_rule_t *rule, char *base);
 int store_resource(resource_t **reslist, resource_t *newres);
 void destroy_resource(resource_t *res);
 
-char *attr_value(resource_node_t *node, char *attrname);
-char *rg_attr_value(resource_node_t *node, char *attrname);
-char *res_attr_value(resource_t *res, char *attrname);
-char *primary_attr_value(resource_t *);
+const char *attr_value(resource_node_t *node, const char *attrname);
+const char *rg_attr_value(resource_node_t *node, const char *attrname);
+const char *res_attr_value(resource_t *res, const char *attrname);
+const char *primary_attr_value(resource_t *);
 int rescmp(resource_t *l, resource_t *r);
 
-#ifdef NO_CCS
-int conf_get(char *query, char **ret);
-int conf_setconfig(char *path);
-#endif
-
 #endif /* _RESLIST_H */
