@@ -1,10 +1,10 @@
 #define _MESSAGE_BUILD
-#include <message.h>
 #include <stdio.h>
 #include <pthread.h>
 #include <libcman.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <message.h>
 #include <string.h>
 #include <sys/types.h>
 #include <time.h>
@@ -339,7 +339,7 @@ cluster_msg_fd_set(msgctx_t *ctx, fd_set *fds, int *max)
 }
 
 
-int
+static int
 cluster_msg_fd_isset(msgctx_t *ctx, fd_set *fds)
 {
 	errno = EINVAL;
@@ -361,7 +361,7 @@ cluster_msg_fd_isset(msgctx_t *ctx, fd_set *fds)
 }
 
 
-int
+static int
 cluster_msg_fd_clr(msgctx_t *ctx, fd_set *fds)
 {
 	errno = EINVAL;
@@ -850,12 +850,13 @@ process_cman_msg(cman_handle_t h, void *priv, char *buf, int len,
 static int
 cluster_msg_accept(msgctx_t *listenctx, msgctx_t *acceptctx)
 {
-	errno = EINVAL;
 	cluster_msg_hdr_t *m;
 	msg_q_t *n;
 	char foo;
 	int err = 0;
 
+	errno = EINVAL;
+
 	if (!listenctx || !acceptctx)
 		return -1;
 	if (listenctx->u.cluster_info.local_ctx != 0)
@@ -1020,7 +1021,7 @@ process_cman_event(cman_handle_t handle, void *private, int reason, int arg)
 		break;
 	}
 
-	argp = ((void *)msg + sizeof(cluster_msg_hdr_t));
+	argp = (int*)((char *)msg + sizeof(cluster_msg_hdr_t));
 	*argp = arg;
 
 	node->len = sizeof(cluster_msg_hdr_t) + sizeof(int);
@@ -1054,7 +1055,7 @@ process_cman_event(cman_handle_t handle, void *private, int reason, int arg)
 
 /* */
 int
-cluster_msg_listen(int me, void *portp, msgctx_t **cluster_ctx)
+cluster_msg_listen(int me, const void *portp, msgctx_t **cluster_ctx)
 {
 	int e;
 	pthread_attr_t attrs;
