@@ -17,6 +17,8 @@
 #include <stdio.h>
 #include <members.h>
 #include <sets.h>
+#include <fo_domain.h>
+#include <groups.h>
 
 
 //#define DEBUG
@@ -34,9 +36,8 @@
      </failoverdomain>
    </failoverdomains>
  */
-int group_property(char *, char *, char *, size_t);
 
-fod_node_t *
+static fod_node_t *
 #ifndef NO_CCS
 fod_get_node(int ccsfd, char *base, int idx, fod_t *domain)
 #else
@@ -56,7 +57,7 @@ fod_get_node(int __attribute__((unused)) ccsfd, char *base, int idx, fod_t *doma
 		if (strcasecmp(ret, fodn->fdn_name))
 			continue;
 
-		log_printf(LOG_ERR, "#30: Node %s defined multiple times in "
+		logt_print(LOG_ERR, "#30: Node %s defined multiple times in "
 		       "domain %s\n", ret, domain->fd_name);
 		free(ret);
 		return NULL;
@@ -75,7 +76,7 @@ fod_get_node(int __attribute__((unused)) ccsfd, char *base, int idx, fod_t *doma
  		 "/cluster/clusternodes/clusternode[@name=\"%s\"]/@nodeid",
  		 ret);
  	if (ccs_get(ccsfd, xpath, &ret) != 0) {
- 		log_printf(LOG_WARNING, "Node %s has no nodeid attribute\n",
+ 		logt_print(LOG_WARNING, "Node %s has no nodeid attribute\n",
  		       fodn->fdn_name);
  		fodn->fdn_nodeid = -1;
  	} else {
@@ -102,7 +103,7 @@ fod_get_node(int __attribute__((unused)) ccsfd, char *base, int idx, fod_t *doma
 }
 
 
-fod_t *
+static fod_t *
 fod_get_domain(int ccsfd, char *base, int idx, fod_t **domains)
 {
 	fod_t *fod;
@@ -120,7 +121,7 @@ fod_get_domain(int ccsfd, char *base, int idx, fod_t **domains)
 		if (strcasecmp(fod->fd_name, ret))
 			continue;
 		    
-		log_printf(LOG_ERR, "#31: Domain %s defined multiple times\n",
+		logt_print(LOG_ERR, "#31: Domain %s defined multiple times\n",
 		       ret);
 		free(ret);
 		return NULL;
@@ -203,22 +204,6 @@ construct_domains(int ccsfd, fod_t **domains)
 }
 
 
-fod_t *
-fod_find_domain(fod_t **domains, char *name)
-{
-	fod_t *dom;
-	
-	list_do(domains, dom) {
-		
-		if (!strcasecmp(dom->fd_name, name))
-			return dom;
-	
-	} while (!list_done(domains,dom));
-	
-	return NULL;
-}
-
-
 void
 deconstruct_domains(fod_t **domains)
 {
@@ -288,7 +273,7 @@ print_domains(fod_t **domains)
  *				2 for Yes, Not lowest-ordered, online member.
  *				3 for Yes, Lowest-ordered, online member.
  */
-int
+static int
 node_in_domain(char *nodename, fod_t *domain,
 	       cluster_member_list_t *membership)
 {
@@ -426,7 +411,7 @@ node_domain_set(fod_t **domains, char *name, int **ret, int *retlen, int *flags)
  */
 int
 node_should_start(int nodeid, cluster_member_list_t *membership,
-		  char *rg_name, fod_t **domains)
+		  const char *rg_name, fod_t **domains)
 {
 	char *nodename = NULL;
 	char domainname[128];
@@ -446,7 +431,7 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 	 */
 	if (!memb_online(membership, nodeid)) {
 #ifdef DEBUG
-		log_printf(LOG_DEBUG,"Member #%d is not online -> NO\n", nodeid);
+		logt_print(LOG_DEBUG,"Member #%d is not online -> NO\n", nodeid);
 #endif
 		return FOD_ILLEGAL;
 	}
@@ -454,15 +439,16 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 	nodename = memb_id_to_name(membership, nodeid);
 
 #ifndef NO_CCS /* XXX Testing only */
-	if (group_property(rg_name, "domain",
+	if (group_property(rg_name, (char *)"domain",
 			    domainname, sizeof(domainname))) {
 		/*
 		 * If no domain is present, then the node in question should
 		 * try to start the service.
 		 */
 #ifdef DEBUG
-		log_printf(LOG_DEBUG,
-		       "Fail-over Domain for service %d nonexistent\n");
+		logt_print(LOG_DEBUG,
+		       "Fail-over Domain for service %s nonexistent\n",
+		       rg_name);
 #endif
 		return FOD_BEST;
 	}
@@ -484,7 +470,7 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 		/*
 		 * Domain doesn't exist!  Weird...
 		 */
-		log_printf(LOG_WARNING, "#66: Domain '%s' specified for resource "
+		logt_print(LOG_WARNING, "#66: Domain '%s' specified for resource "
 		       "group %s nonexistent!\n", domainname, rg_name);
 		return FOD_BEST;
 	}
@@ -507,7 +493,7 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 #ifndef NO_CCS
 	if(nofailback) {
 		if (rg_lock(rg_name, &lockp) != 0) {
-			log_printf(LOG_WARNING, "Error getting a lock\n");
+			logt_print(LOG_WARNING, "Error getting a lock\n");
 			return FOD_BEST;
 		}
                 
@@ -515,7 +501,7 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
                 	/*
 			 * Couldn't get the service state, thats odd
 			 */
-			log_printf(LOG_WARNING, "Problem getting state information for "
+			logt_print(LOG_WARNING, "Problem getting state information for "
 			       "%s\n", rg_name);
 			rg_unlock(&lockp);
 			return FOD_BEST;
@@ -542,17 +528,17 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 		 * domain are online.
 		 */
 #ifdef DEBUG
-		log_printf(LOG_DEBUG, "Member #%d is not a member and no "
+		logt_print(LOG_DEBUG, "Member #%d is not a member and no "
 		       "members are online\n", nodeid);
 #endif
 		if (!restricted) {
 #ifdef DEBUG
-			log_printf(LOG_DEBUG,"Restricted mode off -> BEST\n");
+			logt_print(LOG_DEBUG,"Restricted mode off -> BEST\n");
 #endif
 			return FOD_BEST;
 		}
 #ifdef DEBUG
-		log_printf(LOG_DEBUG,"Restricted mode -> ILLEGAL\n");
+		logt_print(LOG_DEBUG,"Restricted mode -> ILLEGAL\n");
 #endif
 		return FOD_ILLEGAL;
 	case 1:
@@ -562,17 +548,17 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 		 */
 		/* In this case, we can ignore 'restricted' */
 #ifdef DEBUG
-		log_printf(LOG_DEBUG, "Member #%d is not a member of domain %s "
+		logt_print(LOG_DEBUG, "Member #%d is not a member of domain %s "
 		       "and a member is online\n", nodeid, domainname);
 #endif
 		if (!restricted) {
 #ifdef DEBUG
-			log_printf(LOG_DEBUG,"Restricted mode off -> GOOD\n");
+			logt_print(LOG_DEBUG,"Restricted mode off -> GOOD\n");
 #endif
 			return FOD_GOOD;
 		}
 #ifdef DEBUG
-		log_printf(LOG_DEBUG,"Restricted mode -> ILLEGAL\n");
+		logt_print(LOG_DEBUG,"Restricted mode -> ILLEGAL\n");
 #endif
 		return FOD_ILLEGAL;
 	case 2:
@@ -581,7 +567,7 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 		 * lowest-ordered, online member.
 		 */
 #ifdef DEBUG
-		log_printf(LOG_DEBUG, "Member #%d is a member, but is not the "
+		logt_print(LOG_DEBUG, "Member #%d is a member, but is not the "
 		       "lowest-ordered\n", nodeid);
 #endif
 		if (ordered) {
@@ -591,19 +577,19 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 			 */
 			if (nofailback && started && owned_by_node && !no_owner) {
 #ifdef DEBUG
-				log_printf(LOG_DEBUG,"Ordered mode and no "
+				logt_print(LOG_DEBUG,"Ordered mode and no "
 				       "failback -> BEST\n");
 #endif
 				return FOD_BEST;
 			}
 #ifdef DEBUG
-			log_printf(LOG_DEBUG,"Ordered mode -> BETTER\n");
+			logt_print(LOG_DEBUG,"Ordered mode -> BETTER\n");
 #endif
 			return FOD_BETTER;
 		}
 
 #ifdef DEBUG
-		log_printf(LOG_DEBUG,"Not using ordered mode -> BEST\n");
+		logt_print(LOG_DEBUG,"Not using ordered mode -> BEST\n");
 #endif
 		return FOD_BEST;
 	case 3:
@@ -614,7 +600,7 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 
 		if(nofailback && started && !owned_by_node && !no_owner) {
 #ifdef DEBUG
-			log_printf(LOG_DEBUG, "Member #%d is the lowest-ordered "
+			logt_print(LOG_DEBUG, "Member #%d is the lowest-ordered "
 			       "memeber of the domain, but is not the owner "
 			       "-> BETTER\n", nodeid);
 #endif
@@ -623,13 +609,13 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
  
 		/* In this case, we can ignore 'ordered' */
 #ifdef DEBUG
-		log_printf(LOG_DEBUG, "Member #%d is the lowest-ordered member "
+		logt_print(LOG_DEBUG, "Member #%d is the lowest-ordered member "
 		       "of the domain -> BEST\n", nodeid);
 #endif
 		return FOD_BEST;
 	default:
 		/* Do what? */
-		log_printf(LOG_ERR, "#32: Code path error: "
+		logt_print(LOG_ERR, "#32: Code path error: "
 		       "Invalid return from node_in_domain()\n");
 		return FOD_ILLEGAL;
 	}
