@@ -15,11 +15,9 @@
 #ifndef NO_CCS
 #include <logging.h>
 #endif
+#include <groups.h>
 
 
-char *attr_value(resource_node_t *node, char *attrname);
-char *rg_attr_value(resource_node_t *node, char *attrname);
-
 void
 res_build_name(char *buf, size_t buflen, resource_t *res)
 {
@@ -34,8 +32,8 @@ res_build_name(char *buf, size_t buflen, resource_t *res)
    @param attrname	Attribute to retrieve.
    @return 		value of attribute or NULL if not found
  */
-char *
-res_attr_value(resource_t *res, char *attrname)
+const char *
+res_attr_value(resource_t *res, const char *attrname)
 {
 	resource_attr_t *ra;
 	int x;
@@ -67,7 +65,7 @@ res_attr_value(resource_t *res, char *attrname)
    @return 		value of attribute or NULL if not found
  */
 static char *
-_attr_value(resource_node_t *node, char *attrname, char *ptype)
+_attr_value(resource_node_t *node, const char *attrname, const char *ptype)
 {
 	resource_t *res;
 	resource_attr_t *ra;
@@ -118,8 +116,8 @@ _attr_value(resource_node_t *node, char *attrname, char *ptype)
 }
 
 
-char *
-attr_value(resource_node_t *node, char *attrname)
+const char *
+attr_value(resource_node_t *node, const char *attrname)
 {
 	return _attr_value(node, attrname, NULL);
 }
@@ -129,15 +127,15 @@ attr_value(resource_node_t *node, char *attrname)
   Run to the top of the tree.  Used to determine certain attributes of the
   resource group in-line, during resource tree operations.
  */
-char *
-rg_attr_value(resource_node_t *node, char *attrname)
+const char *
+rg_attr_value(resource_node_t *node, const char *attrname)
 {
 	for (; node->rn_parent; node = node->rn_parent);
 	return res_attr_value(node->rn_resource, attrname);
 }
 
 
-char *
+const char *
 primary_attr_value(resource_t *res)
 {
 	int x;
@@ -156,7 +154,6 @@ primary_attr_value(resource_t *res)
 }
 
 
-
 /**
    Compare two resources.
 
@@ -257,7 +254,7 @@ rescmp(resource_t *left, resource_t *right)
    @return		Resource matching type/ref or NULL if none.
  */   
 resource_t *
-find_resource_by_ref(resource_t **reslist, char *type, char *ref)
+find_resource_by_ref(resource_t **reslist, const char *type, const char *ref)
 {
 	resource_t *curr;
 	int x;
@@ -295,12 +292,12 @@ find_resource_by_ref(resource_t **reslist, char *type, char *ref)
    @return		Resource matching type/ref or NULL if none.
  */   
 resource_t *
-find_root_by_ref(resource_t **reslist, char *ref)
+find_root_by_ref(resource_t **reslist, const char *ref)
 {
 	resource_t *curr;
 	char ref_buf[128];
 	char *type;
-	char *name = ref;
+	char *name = (char *)ref;
 	int x;
 
 	snprintf(ref_buf, sizeof(ref_buf), "%s", ref);
@@ -311,8 +308,8 @@ find_root_by_ref(resource_t **reslist, char *ref)
 		name++;
 	} else {
 		/* Default type */
-		type = "service";
-		name = ref;
+		type = (char *)"service";
+		name = (char *)ref;
 	}
 
 	list_do(reslist, curr) {
@@ -399,7 +396,7 @@ store_resource(resource_t **reslist, resource_t *newres)
 					       newres->r_attrs[x].ra_value
 					       );
 #else 
-					log_printf(LOG_ERR,
+					logt_print(LOG_ERR,
                                                "%s attribute collision. "
                                                "type=%s attr=%s value=%s\n",
 					       (newres->r_attrs[x].ra_flags&
@@ -433,7 +430,7 @@ store_resource(resource_t **reslist, resource_t *newres)
  */
 char *
 xpath_get_one(xmlDocPtr __attribute__ ((unused)) doc,
-	      xmlXPathContextPtr ctx, char *query)
+	      xmlXPathContextPtr ctx, const char *query)
 {
 	char *val = NULL, *ret = NULL;
 	xmlXPathObjectPtr obj;
@@ -627,7 +624,7 @@ act_dup(resource_act_t *acts)
 
 
 /* Copied from resrules.c -- _get_actions */
-void
+static void
 #ifndef NO_CCS
 _get_actions_ccs(int ccsfd, char *base, resource_t *res)
 #else
@@ -720,7 +717,7 @@ _get_actions_ccs(int __attribute__((unused)) ccsfd, char *base, resource_t *res)
 resource_t *
 load_resource(int ccsfd, resource_rule_t *rule, char *base)
 {
-	resource_t *res;
+	resource_t *res = NULL;
 	char ccspath[1024];
 	char *attrname, *attr;
 	int x, found = 0, flags;
@@ -835,7 +832,7 @@ int
 load_resources(int ccsfd, resource_t **reslist, resource_rule_t **rulelist)
 {
 	int resID = 0;
-	resource_t *newres;
+	resource_t *newres = NULL;
 	resource_rule_t *currule;
 	char tok[256];
 
@@ -854,7 +851,7 @@ load_resources(int ccsfd, resource_t **reslist, resource_rule_t **rulelist)
 	       		       printf("Error storing %s resource\n",
 				      newres->r_rule->rr_type);
 #else
-	       		       log_printf(LOG_ERR,
+	       		       logt_print(LOG_ERR,
 				      "Error storing %s resource\n",
 				      newres->r_rule->rr_type);
 #endif
