@@ -15,32 +15,17 @@
 #include <logging.h>
 #include <assert.h>
 
-#ifdef INTERNAL_MALLOC
-void malloc_zap_mutex(void);
-#endif
-
 /* XXX from resrules.c */
-int store_childtype(resource_child_t **childp, char *name, int start,
-		    int stop, int forbid, int flags);
 int _res_op(resource_node_t **tree, resource_t *first, char *type,
 	    void * __attribute__((unused))ret, int op);
 static inline int
 _res_op_internal(resource_node_t **tree, resource_t *first,
 		 char *type, void *__attribute__((unused))ret, int realop,
 		 resource_node_t *node);
-void print_env(char **env);
 static inline int _res_op_internal(resource_node_t **tree, resource_t *first,
 		 char *type, void *__attribute__((unused))ret, int realop,
 		 resource_node_t *node);
 
-/* XXX from reslist.c */
-void * act_dup(resource_act_t *acts);
-
-
-/* XXX from reslist.c */
-void * act_dup(resource_act_t *acts);
-
-
 /* XXX from reslist.c */
 void * act_dup(resource_act_t *acts);
 time_t get_time(char *action, int depth, resource_node_t *node);
@@ -63,6 +48,7 @@ const char *ocf_errors[] = {
 /* XXX MEGA HACK */
 #ifdef NO_CCS
 static int _no_op_mode_ = 0;
+void _no_op_mode(int arg);
 void
 _no_op_mode(int arg)
 {
@@ -74,7 +60,7 @@ _no_op_mode(int arg)
 /**
    ocf_strerror
  */
-const char *
+static const char *
 ocf_strerror(int ret)
 {
 	if (ret >= 0 && ret < OCF_RA_MAX)
@@ -125,7 +111,7 @@ add_ocf_stuff(resource_t *res, char **env, int depth, int refcnt)
 		*minor = 0;
 		minor++;
 	} else
-		minor = "0";
+		minor = (char *)"0";
 		
 	/*
 	   Store the OCF major version
@@ -247,7 +233,7 @@ build_env(resource_node_t *node, int depth, int refcnt)
 	attrs = 0;
 	for (x = 0; res->r_attrs && res->r_attrs[x].ra_name; x++) {
 
-		val = attr_value(node, res->r_attrs[x].ra_name);
+		val = (char *)attr_value(node, res->r_attrs[x].ra_name);
 		if (!val)
 			continue;
 
@@ -279,27 +265,11 @@ build_env(resource_node_t *node, int depth, int refcnt)
 }
 
 
-/**
-   Print info in an environment array
-
-   @param env		Environment to print.
- */
-void
-print_env(char **env)
-{
-	int x;
-
-	for (x = 0; env[x]; x++) {
-		printf("%s\n", env[x]);
-	}
-}
-
-
 /**
    Set all signal handlers to default for exec of a script.
    ONLY do this after a fork().
  */
-void
+static void
 restore_signals(void)
 {
 	sigset_t set;
@@ -314,7 +284,7 @@ restore_signals(void)
 
 
 /** Find the index for a given operation / depth in a resource node */
-int
+static int
 res_act_index(resource_node_t *node, const char *op_str, int depth)
 {
 	int x = 0;
@@ -388,9 +358,6 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 
 	if (!childpid) {
 		/* Child */ 
-#ifdef INTERNAL_MALLOC
-		malloc_zap_mutex();
-#endif
 #if 0
 		printf("Exec of script %s, action %s type %s\n",
 			res->r_rule->rr_agent, agent_op_str(op),
@@ -441,7 +408,7 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 
 		if (pid != childpid && sleeptime == 0) {
 
-			log_printf(LOG_ERR,
+			logt_print(LOG_ERR,
 			       "%s on %s:%s timed out after %d seconds\n",
 			       op_str, res->r_rule->rr_type,
 			       res->r_attrs->ra_value,
@@ -453,7 +420,7 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 			sleep(1);
 			pid = waitpid(childpid, &ret, WNOHANG);
 			if (pid == 0) {
-				log_printf(LOG_ERR,
+				logt_print(LOG_ERR,
 				       "Task %s PID %d did not exit "
 				       "after SIGKILL\n",
 				       op_str, childpid);
@@ -481,7 +448,7 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 #else
 		if (ret) {
 #endif
-			log_printf(LOG_NOTICE,
+			logt_print(LOG_NOTICE,
 			       "%s on %s \"%s\" returned %d (%s)\n",
 			       op_str, res->r_rule->rr_type,
 			       res->r_attrs->ra_value, ret,
@@ -513,16 +480,16 @@ assign_restart_policy(resource_t *curres, resource_node_t *parent,
 	if (parent) /* Non-parents don't get one for now */
 		return;
 
-	val = res_attr_value(curres, "max_restarts");
+	val = (char *)res_attr_value(curres, "max_restarts");
 	if (!val)
 		return;
 	max_restarts = atoi(val);
 	if (max_restarts <= 0)
 		return;
-	val = res_attr_value(curres, "restart_expire_time");
+	val = (char *)res_attr_value(curres, "restart_expire_time");
 	if (val) {
 		restart_expire_time = (time_t)expand_time(val);
-		if (!restart_expire_time)
+		if ((int64_t)restart_expire_time < 0)
 			return;
 	}
 
@@ -671,7 +638,7 @@ build_tree(int ccsfd, resource_node_t **tree,
 {
 	char tok[512];
 	resource_rule_t *childrule;
-	resource_node_t *node;
+	resource_node_t *node = NULL;
 	char *ref;
 	char *tmp;
 	int ccount = 0, x = 0, y = 0, flags = 0;
@@ -872,7 +839,7 @@ build_resource_tree(int ccsfd, resource_node_t **tree,
 void
 destroy_resource_tree(resource_node_t **tree)
 {
-	resource_node_t *node;
+	resource_node_t *node = NULL;
 
 	while ((node = *tree)) {
 		if ((*tree)->rn_child)
@@ -892,7 +859,7 @@ destroy_resource_tree(resource_node_t **tree)
 }
 
 
-void
+static void
 _print_resource_tree(resource_node_t **tree, int level)
 {
 	resource_node_t *node;
@@ -1059,7 +1026,7 @@ _do_child_default_level(resource_node_t **tree, resource_t *first,
 			in the subtree).
    @see			_res_op res_exec
  */
-int
+static int
 _res_op_by_level(resource_node_t **tree, resource_t *first, void *ret,
 		 int op)
 {
@@ -1089,19 +1056,21 @@ _res_op_by_level(resource_node_t **tree, resource_t *first, void *ret,
 }
 
 
-void
-mark_nodes(resource_node_t *node, int state, int flags)
+static void
+mark_nodes(resource_node_t *node, int state, int setflags, int clearflags)
 {
 	int x;
 	resource_node_t *child;
 
 	list_for(&node->rn_child, child, x) {
 		if (child->rn_child)
-			mark_nodes(child->rn_child, state, flags);
+			mark_nodes(child->rn_child, state, setflags,
+				   clearflags);
 	}
 
 	node->rn_state = state;
-	node->rn_flags |= (RF_NEEDSTART | RF_NEEDSTOP);
+	node->rn_flags |= (setflags);
+	node->rn_flags &= ~(clearflags);
 }
 
 
@@ -1110,7 +1079,7 @@ mark_nodes(resource_node_t *node, int state, int flags)
    status operation was run and selects the highest possible resource depth
    to use given the elapsed time.
   */
-int
+static int
 do_status(resource_node_t *node)
 {
 	int x = 0, idx = -1;
@@ -1151,15 +1120,19 @@ do_status(resource_node_t *node)
 	}
 
 	/* No check levels ready at the moment. */
-	if (idx == -1) {
+	/* Cap status check children if configured to do so */
+	if (idx == -1 || rg_inc_children() < 0) {
 		if (node->rn_checked)
 			return node->rn_last_status;
 		return 0;
 	}
 
-
-	node->rn_actions[idx].ra_last = now;
 	x = res_exec(node, RS_STATUS, NULL, node->rn_actions[idx].ra_depth);
+	rg_dec_children();
+
+	/* Record status check result *after* the status check has
+	 * completed. */
+	node->rn_actions[idx].ra_last = time(NULL);
 
 	node->rn_last_status = x;
 	node->rn_last_depth = node->rn_actions[idx].ra_depth;
@@ -1179,8 +1152,8 @@ do_status(resource_node_t *node)
 }
 
 
-void
-set_time(char *action, int depth, resource_node_t *node)
+static void
+set_time(const char *action, int depth, resource_node_t *node)
 {
 	time_t now;
 	int x = 0;
@@ -1217,7 +1190,7 @@ get_time(char *action, int depth, resource_node_t *node)
 }
 
 
-void
+static void
 clear_checks(resource_node_t *node)
 {
 	time_t now;
@@ -1337,7 +1310,19 @@ _res_op_internal(resource_node_t __attribute__ ((unused)) **tree,
 			++node->rn_resource->r_incarnations;
 			node->rn_state = RES_STARTED;
 		}
-	} else if (me && (op == RS_STATUS)) {
+	} else if (me && (op == RS_STATUS || op == RS_STATUS_INQUIRY)) {
+
+		/* Special quick-check for status inquiry */
+		if (op == RS_STATUS_INQUIRY) {
+			if (res_exec(node, RS_STATUS, NULL, 0) != 0)
+				return SFL_FAILURE;
+
+			/* XXX: A migratable service (the only place this
+			 * check can be used) cannot have child dependencies
+			 * anyway, so this is a short-circuit. */
+			return 0;
+		}
+
 		/* Check status before children*/
 		rv = do_status(node);
 		if (rv != 0) {
@@ -1347,7 +1332,7 @@ _res_op_internal(resource_node_t __attribute__ ((unused)) **tree,
 			   node is independent or not.
 			 */
 			mark_nodes(node, RES_FAILED,
-				   RF_NEEDSTART | RF_NEEDSTOP);
+				   RF_NEEDSTART | RF_NEEDSTOP, 0);
 
 			/* If we're an independent subtree, return a flag
 			   stating that this section is recoverable apart
@@ -1360,6 +1345,7 @@ _res_op_internal(resource_node_t __attribute__ ((unused)) **tree,
 			return SFL_FAILURE;
 		}
 
+		mark_nodes(node, RES_STARTED, 0, RF_NEEDSTOP);
 	}
 
        if (node->rn_child) {
@@ -1374,7 +1360,7 @@ _res_op_internal(resource_node_t __attribute__ ((unused)) **tree,
 		if (op == RS_STATUS && (rv & SFL_FAILURE) &&
 		    (node->rn_flags & RF_INDEPENDENT)) {
 			mark_nodes(node, RES_FAILED,
-				   RF_NEEDSTART | RF_NEEDSTOP);
+				   RF_NEEDSTART | RF_NEEDSTOP, 0);
 			rv = SFL_RECOVERABLE;
 		}
 	}
@@ -1523,16 +1509,16 @@ res_status(resource_node_t **tree, resource_t *res, void *ret)
 
 
 /**
-   Grab resource info for all occurrences of a resource in a tree
+   Check status of all occurrences of a resource in a tree
 
    @param tree		Tree to search for our resource.
    @param res		Resource to start/stop
    @param ret		Unused
  */
 int
-res_resinfo(resource_node_t **tree, resource_t *res, void *ret)
+res_status_inquiry(resource_node_t **tree, resource_t *res, void *ret)
 {
-	return _res_op(tree, res, NULL, ret, RS_RESINFO);
+	return _res_op(tree, res, NULL, ret, RS_STATUS_INQUIRY);
 }
 
 
