@@ -1,6 +1,9 @@
-//#define DEBUG
 #include <assert.h>
 #include <platform.h>
+#include <list.h>
+#include <time.h>
+#include <restart_counter.h>
+#include <reslist.h>
 #include <message.h>
 #include <members.h>
 #ifdef OPENAIS
@@ -19,74 +22,21 @@
 #include <msgsimple.h>
 #include <res-ocf.h>
 #include <event.h>
+#include <groups.h>
+#include <fo_domain.h>
 
 /* XXX - copied :( */
 #define cn_svccount cn_address.cna_address[0] /* Theses are uint8_t size */
 #define cn_svcexcl  cn_address.cna_address[1]
 
-int node_should_start_safe(uint32_t, cluster_member_list_t *, char *);
+static int msvc_check_cluster(const char *svcName);
+static int handle_started_status(const char *svcName, int ret, rg_state_t *svcStatus);
+static int handle_migrate_status(const char *svcName, int ret, rg_state_t *svcStatus);
 
-int next_node_id(cluster_member_list_t *membership, int me);
+static int _svc_stop_finish(const char *svcName, int failed, uint32_t newstate);
 
-int rg_exec_script(char *rgname, char *script, char *action);
-static int _svc_stop_finish(char *svcName, int failed, uint32_t newstate);
-
-int set_rg_state(char *servicename, rg_state_t *svcblk);
-int get_rg_state(char *servicename, rg_state_t *svcblk);
-void get_recovery_policy(char *rg_name, char *buf, size_t buflen);
-int check_depend_safe(char *servicename);
-int group_migratory(char *servicename, int lock);
-int have_exclusive_resources(void);
-int check_exclusive_resources(cluster_member_list_t *membership, char *svcName);
-static int msvc_check_cluster(char *svcName);
-static inline int handle_started_status(char *svcName, int ret, rg_state_t *svcStatus);
-static inline int handle_migrate_status(char *svcName, int ret, rg_state_t *svcStatus);
-int count_resource_groups_local(cman_node_t *mp);
-int is_exclusive(char *svcName);
-
-
-int 
-next_node_id(cluster_member_list_t *membership, int me)
-{
-	int low = (int)(-1);
-	int next = me, curr;
-	int x;
-
-	for (x = 0; x < membership->cml_count; x++) {
-		curr = membership->cml_members[x].cn_nodeid;
-		if (curr < low)
-			low = curr;
-
-		if ((curr > me) && ((next == me) || (curr < next)))
-			next = curr;
-	}
-
-	/* I am highest ID; go to lowest */
-	if (next == me)
-		next = low;
-
-	return next;
-}
-
-
-char *
-c_name(char *svcName)
-{
-	char *ptr, *ret = svcName;
-
-	ptr = strchr(svcName,':');
-	if (!ptr)
-		return ret;
-	if ((int)(ptr - svcName) == 7 &&
-	    !memcmp(svcName, "service", 7)) /* strlen("service") */
-		ret = ptr + 1;
-
-	return ret;
-}
-
-
-void
-broadcast_event(char *svcName, uint32_t state, int owner, int last)
+static void
+broadcast_event(const char *svcName, uint32_t state, int owner, int last)
 {
 	SmMessageSt msgp;
 	msgctx_t everyone;
@@ -113,7 +63,7 @@ broadcast_event(char *svcName, uint32_t state, int owner, int last)
 
 
 int
-svc_report_failure(char *svcName)
+svc_report_failure(const char *svcName)
 {
 	struct dlm_lksb lockp;
 	rg_state_t svcStatus;
@@ -121,13 +71,13 @@ svc_report_failure(char *svcName)
 	cluster_member_list_t *membership;
 
 	if (rg_lock(svcName, &lockp) == -1) {
-		log_printf(LOG_ERR, "#41: Couldn't obtain lock for RG %s: %s\n",
+		logt_print(LOG_ERR, "#41: Couldn't obtain lock for RG %s: %s\n",
 		       svcName, strerror(errno));
 		return -1;
 	}
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
-		log_printf(LOG_ERR, "#42: Couldn't obtain status for RG %s\n",
+		logt_print(LOG_ERR, "#42: Couldn't obtain status for RG %s\n",
 		       svcName);
 		rg_unlock(&lockp);
 		return -1;
@@ -137,17 +87,17 @@ svc_report_failure(char *svcName)
 	membership = member_list();
 	nodeName = memb_id_to_name(membership, svcStatus.rs_last_owner);
 	if (nodeName) {
-		log_printf(LOG_ALERT, "#2: Service %s returned failure "
+		logt_print(LOG_ALERT, "#2: Service %s returned failure "
 		       "code.  Last Owner: %s\n", svcName, nodeName);
 	} else {
-		log_printf(LOG_ALERT, "#3: Service %s returned failure "
+		logt_print(LOG_ALERT, "#3: Service %s returned failure "
 		       "code.  Last Owner: %d\n",
 		       svcName, (int)svcStatus.rs_last_owner);
 	}
 
 	free_member_list(membership);
 
-	log_printf(LOG_ALERT,
+	logt_print(LOG_ALERT,
 	       "#4: Administrator intervention required.\n");
 
 	return 0;
@@ -156,9 +106,9 @@ svc_report_failure(char *svcName)
 
 int
 #ifdef DEBUG
-_rg_lock(char *name, struct dlm_lksb *p)
+_rg_lock(const char *name, struct dlm_lksb *p)
 #else
-rg_lock(char *name, struct dlm_lksb *p)
+rg_lock(const char *name, struct dlm_lksb *p)
 #endif
 {
 	char res[256];
@@ -170,7 +120,7 @@ rg_lock(char *name, struct dlm_lksb *p)
 
 #ifdef DEBUG
 int
-_rg_lock_dbg(char *name, struct dlm_lksb *p, char *file, int line)
+_rg_lock_dbg(const char *name, struct dlm_lksb *p, const char *file, int line)
 {
 	dbg_printf("rg_lock(%s) @ %s:%d\n", name, file, line);
 	return _rg_lock(name, p);
@@ -191,7 +141,7 @@ rg_unlock(struct dlm_lksb *p)
 
 #ifdef DEBUG
 int
-_rg_unlock_dbg(struct dlm_lksb *p, char *file, int line)
+_rg_unlock_dbg(struct dlm_lksb *p, const char *file, int line)
 {
 	dbg_printf("rg_unlock() @ %s:%d\n", file, line);
 	return _rg_unlock(p);
@@ -248,7 +198,7 @@ send_response(int ret, int nodeid, request_t *req)
 
 
 int
-set_rg_state(char *name, rg_state_t *svcblk)
+set_rg_state(const char *name, rg_state_t *svcblk)
 {
 	char res[256];
 	rg_state_t svcblk_store;
@@ -286,7 +236,7 @@ set_rg_state(char *name, rg_state_t *svcblk)
 
 
 static int
-init_rg(char *name, rg_state_t *svcblk)
+init_rg(const char *name, rg_state_t *svcblk)
 {
 	svcblk->rs_owner = 0;
 	svcblk->rs_last_owner = 0;
@@ -301,7 +251,7 @@ init_rg(char *name, rg_state_t *svcblk)
 
 
 int
-get_rg_state(char *name, rg_state_t *svcblk)
+get_rg_state(const char *name, rg_state_t *svcblk)
 {
 	char res[256];
 	int ret;
@@ -396,9 +346,8 @@ get_rg_state(char *name, rg_state_t *svcblk)
 }
 
 
-int vf_read_local(char *, uint64_t *, void *, uint32_t *);
 int
-get_rg_state_local(char *name, rg_state_t *svcblk)
+get_rg_state_local(const char *name, rg_state_t *svcblk)
 {
 	char res[256];
 	int ret;
@@ -407,7 +356,7 @@ get_rg_state_local(char *name, rg_state_t *svcblk)
 	int datalen;
 #else
 	void *data = NULL;
-	uint64_t viewno;
+	uint64_t viewno = 0;
 	uint32_t datalen;
 #endif
 
@@ -465,14 +414,14 @@ get_rg_state_local(char *name, rg_state_t *svcblk)
  *			6 = DO NOT stop service, mark stopped and return
  *			    RG_SUCCESS (0)
  */
-int
-svc_advise_stop(rg_state_t *svcStatus, char *svcName, int req)
+static int
+svc_advise_stop(rg_state_t *svcStatus, const char *svcName, int req)
 {
 	cluster_member_list_t *membership = member_list();
 	int ret = 0;
 	
 	if (svcStatus->rs_flags & RG_FLAG_FROZEN) {
-		log_printf(LOG_DEBUG, "Service %s frozen.\n", svcName);
+		logt_print(LOG_DEBUG, "Service %s frozen.\n", svcName);
 		return 5;
 	}
 
@@ -508,7 +457,7 @@ svc_advise_stop(rg_state_t *svcStatus, char *svcName, int req)
 			}
 
 			/* Disable or relocate request here. */
-			log_printf(LOG_DEBUG, "Forwarding req. to %s.\n",
+			logt_print(LOG_DEBUG, "Forwarding req. to %s.\n",
 			       memb_id_to_name(membership,
 					       svcStatus->rs_owner));
 			ret = 3;
@@ -540,7 +489,7 @@ svc_advise_stop(rg_state_t *svcStatus, char *svcName, int req)
 			ret = 1;
 			break;
 		}
-		log_printf(LOG_DEBUG,
+		logt_print(LOG_DEBUG,
 		       "Not stopping %s: service is failed\n",
 		       svcName);
 		ret = 0;
@@ -557,7 +506,7 @@ svc_advise_stop(rg_state_t *svcStatus, char *svcName, int req)
 	case RG_STATE_DISABLED:
 	case RG_STATE_UNINITIALIZED:
 		if (req == RG_DISABLE) {
-			log_printf(LOG_NOTICE,
+			logt_print(LOG_NOTICE,
 			       "Disabling disabled service %s\n",
 			       svcName);
 			ret = 1;
@@ -565,12 +514,12 @@ svc_advise_stop(rg_state_t *svcStatus, char *svcName, int req)
 		}
 
 		ret = 2;
-		log_printf(LOG_DEBUG, "Not stopping disabled service %s\n",
+		logt_print(LOG_DEBUG, "Not stopping disabled service %s\n",
 		       svcName);
 		break;
 
 	default:
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#42: Cannot stop RG %s: Invalid State %d\n",
 		       svcName, svcStatus->rs_state);
 		break;
@@ -596,20 +545,20 @@ svc_advise_stop(rg_state_t *svcStatus, char *svcName, int req)
  *			4 = DO NOT start service, return RG_ERUN
  *			5 = DO NOT start service, return RG_EFROZEN
  */
-int
-svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
+static int
+svc_advise_start(rg_state_t *svcStatus, const char *svcName, int req)
 {
 	cluster_member_list_t *membership = member_list();
 	int ret = 0;
 	
 	if (svcStatus->rs_flags & RG_FLAG_FROZEN) {
-		log_printf(LOG_DEBUG, "Service %s frozen.\n", svcName);
+		logt_print(LOG_DEBUG, "Service %s frozen.\n", svcName);
 		return 5;
 	}
 
 	switch(svcStatus->rs_state) {
 	case RG_STATE_FAILED:
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#43: Service %s has failed; can not start.\n",
 		       svcName);
 		break;
@@ -625,7 +574,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 		if (svcStatus->rs_owner == (uint32_t)my_id()) {
 		    	/*
 			 * Service is already running locally
-			log_printf(LOG_DEBUG,
+			logt_print(LOG_DEBUG,
 			       "RG %s is already running locally\n", svcName);
 			 */
 			ret = 4;
@@ -636,7 +585,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 		    memb_online(membership, svcStatus->rs_owner)) {
 			/*
 			 * Service is running and the owner is online!
-			log_printf(LOG_DEBUG, "RG %s is running on member %s.\n",
+			logt_print(LOG_DEBUG, "RG %s is running on member %s.\n",
 			       svcName,
 			       memb_id_to_name(membership,svcStatus->rs_owner));
 			 */
@@ -652,7 +601,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 				break;
 			}
 
-			log_printf(LOG_NOTICE,
+			logt_print(LOG_NOTICE,
 			       "Starting stopped service%s\n",
 			       svcName);
 			ret = 1;
@@ -660,7 +609,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 		}
 
 		if (rg_locked()) {
-			log_printf(LOG_WARNING, "Not initiating failover of %s: "
+			logt_print(LOG_WARNING, "Not initiating failover of %s: "
 			       "Resource groups locked!\n", svcName);
 			ret = 3;
 			break;
@@ -669,7 +618,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 		/*
 		 * Service is running but owner is down -> RG_EFAILOVER
 		 */
-		log_printf(LOG_NOTICE,
+		logt_print(LOG_NOTICE,
 		       "Taking over service %s from down member %s\n",
 		       svcName, memb_id_to_name(membership,
 						svcStatus->rs_owner));
@@ -681,17 +630,16 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 		 * Starting failed service...
 		 */
 		if (req == RG_START_RECOVER || central_events_enabled()) {
-			log_printf(LOG_NOTICE,
+			logt_print(LOG_NOTICE,
 			       "Recovering failed service %s\n",
 			       svcName);
-			svcStatus->rs_state = RG_STATE_STOPPED;
 			/* Start! */
 			ret = 1;
 			break;
 		}
 
 		/* Don't start, but return success. */
-		log_printf(LOG_DEBUG,
+		logt_print(LOG_DEBUG,
 		       "Not starting %s: recovery state\n",
 		       svcName);
 		ret = 2;
@@ -705,7 +653,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 			break;
 		}
 
-		log_printf(LOG_NOTICE, "Starting stopped service %s\n",
+		logt_print(LOG_NOTICE, "Starting stopped service %s\n",
 		       svcName);
 		ret = 1;
 		break;
@@ -719,7 +667,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 				break;
 			}
 
-			log_printf(LOG_NOTICE,
+			logt_print(LOG_NOTICE,
 			       "Starting disabled service %s\n",
 			       svcName);
 			ret = 1;
@@ -730,12 +678,12 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 			break;
 		}
 
-		log_printf(LOG_DEBUG, "Not starting disabled RG %s\n",
+		logt_print(LOG_DEBUG, "Not starting disabled RG %s\n",
 		       svcName);
 		break;
 
 	default:
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#44: Cannot start RG %s: Invalid State %d\n",
 		       svcName, svcStatus->rs_state);
 		break;
@@ -755,21 +703,21 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
  * @return		FAIL, 0
  */
 int
-svc_start(char *svcName, int req)
+svc_start(const char *svcName, int req)
 {
 	struct dlm_lksb lockp;
 	int ret;
 	rg_state_t svcStatus;
 
 	if (rg_lock(svcName, &lockp) < 0) {
-		log_printf(LOG_ERR, "#45: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#45: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return RG_EFAIL;
 	}
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#46: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#46: Failed getting status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
@@ -798,16 +746,19 @@ svc_start(char *svcName, int req)
 	/* LOCK HELD if we get here */
 
 	svcStatus.rs_owner = my_id();
-	svcStatus.rs_state = RG_STATE_STARTING;
 	svcStatus.rs_transition = (uint64_t)time(NULL);
 
-	if (req == RG_START_RECOVER)
+	if (svcStatus.rs_state == RG_STATE_RECOVER) {
+		add_restart(svcName);
 		svcStatus.rs_restarts++;
-	else
+	} else {
 		svcStatus.rs_restarts = 0;
+	}
+
+	svcStatus.rs_state = RG_STATE_STARTING;
 
 	if (set_rg_state(svcName, &svcStatus) < 0) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#47: Failed changing service status\n");
 		rg_unlock(&lockp);
 		return RG_EFAIL;
@@ -820,14 +771,14 @@ svc_start(char *svcName, int req)
 			cute values scripts might return */
 
 	if (rg_lock(svcName, &lockp) < 0) {
-		log_printf(LOG_ERR, "#74: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#74: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return RG_EFAIL;
 	}
 
 	svcStatus.rs_state = RG_STATE_STARTED;
 	if (set_rg_state(svcName, &svcStatus) != 0) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#75: Failed changing service status\n");
 		rg_unlock(&lockp);
 		return RG_EFAIL;
@@ -835,14 +786,14 @@ svc_start(char *svcName, int req)
 	rg_unlock(&lockp);
        
 	if (ret == 0) {
-		log_printf(LOG_NOTICE,
+		logt_print(LOG_NOTICE,
 		       "Service %s started\n",
 		       svcName);
 
 		broadcast_event(svcName, RG_STATE_STARTED, svcStatus.rs_owner,
 				svcStatus.rs_last_owner);
 	} else {
-		log_printf(LOG_WARNING,
+		logt_print(LOG_WARNING,
 		       "#68: Failed to start %s; return value: %d\n",
 		       svcName, ret);
 	}
@@ -855,7 +806,7 @@ svc_start(char *svcName, int req)
  * Migrate a service to another node.
  */
 int
-svc_migrate(char *svcName, int target)
+svc_migrate(const char *svcName, int target)
 {
 	struct dlm_lksb lockp;
 	rg_state_t svcStatus;
@@ -882,14 +833,14 @@ svc_migrate(char *svcName, int target)
 	free_member_list(membership);
 
 	if (rg_lock(svcName, &lockp) < 0) {
-		log_printf(LOG_ERR, "#45: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#45: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return RG_EFAIL;
 	}
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#46: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#46: Failed getting status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
@@ -927,7 +878,7 @@ svc_migrate(char *svcName, int target)
 	svcStatus.rs_transition = (uint64_t)time(NULL);
 
 	if (set_rg_state(svcName, &svcStatus) != 0) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#75: Failed changing service status\n");
 		rg_unlock(&lockp);
 		return RG_EFAIL;
@@ -965,14 +916,14 @@ svc_migrate(char *svcName, int target)
 	   services are ... well, migratable, we can just flip the state
 	   back to 'started' and error checking will fix it later. */
 	if (rg_lock(svcName, &lockp) < 0) {
-		log_printf(LOG_ERR, "#45: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#45: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return ret;
 	}
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#46: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#46: Failed getting status for RG %s\n",
 		       svcName);
 		return ret;
 	}
@@ -1002,8 +953,8 @@ svc_migrate(char *svcName, int target)
  * Returns the node ID of the new owner, if any.  -1 if no one in the
  * cluster has seen the service.
  */
-int
-get_new_owner(char *svcName)
+static int
+get_new_owner(const char *svcName)
 {
 	SmMessageSt msgp, response;
 	msgctx_t ctx;
@@ -1034,7 +985,7 @@ get_new_owner(char *svcName)
 			continue;
 
 		if (msg_open(MSG_CLUSTER, membership->cml_members[x].cn_nodeid,
-			     RG_PORT, &ctx, 2) < 0) {
+			     RG_PORT, &ctx, 10) < 0) {
 			/* failed to open: better to claim false successful
 			   status rather than claim a failure and possibly
 			   end up with a service on >1 node */
@@ -1042,8 +993,9 @@ get_new_owner(char *svcName)
 		}
 
 		msg_send(&ctx, &msgp, sizeof(msgp));
-		if (msg_receive(&ctx, &response, sizeof (response), 5) != sizeof(response))
-			goto cont;;
+		if (msg_receive(&ctx, &response, sizeof (response), 10) !=
+		    		sizeof(response))
+			goto cont;
 
 		swab_SmMessageSt(&response);
 		if (response.sm_data.d_ret == RG_SUCCESS)
@@ -1071,7 +1023,7 @@ out:
    using the standard management tools for the service/virtual machine.
  */
 static int
-msvc_check_cluster(char *svcName)
+msvc_check_cluster(const char *svcName)
 {
 	struct dlm_lksb lockp;
 	int newowner;
@@ -1082,24 +1034,24 @@ msvc_check_cluster(char *svcName)
 
 	newowner = get_new_owner(svcName);
 	if (newowner < 0) {
-		log_printf(LOG_DEBUG, "No other nodes have seen %s\n", svcName);
+		logt_print(LOG_DEBUG, "No other nodes have seen %s\n", svcName);
 		return -1;
 	}
 
 	/* New owner found */
-	log_printf(LOG_NOTICE, "Migration: %s is running on %d\n", svcName, newowner);
+	logt_print(LOG_NOTICE, "Migration: %s is running on %d\n", svcName, newowner);
 
 	/* If the check succeeds (returns 0), then flip the state back to
 	   'started' - with a new owner */
 	if (rg_lock(svcName, &lockp) < 0) {
-		log_printf(LOG_ERR, "#451: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#451: Unable to obtain cluster lock: %s\n",
 			strerror(errno));
 		return -1;
 	}
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#452: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#452: Failed getting status for RG %s\n",
 		       svcName);
 		return -1;
 	}
@@ -1109,7 +1061,7 @@ msvc_check_cluster(char *svcName)
 
 	if (set_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#453: Failed setting status for RG %s\n",
+		logt_print(LOG_ERR, "#453: Failed setting status for RG %s\n",
 		       svcName);
 		return -1;
 	}
@@ -1126,21 +1078,21 @@ msvc_check_cluster(char *svcName)
  * @return		RG_EFORWARD, RG_EFAIL, 0
  */
 int
-svc_status(char *svcName)
+svc_status(const char *svcName)
 {
 	struct dlm_lksb lockp;
 	rg_state_t svcStatus;
 	int ret;
 
 	if (rg_lock(svcName, &lockp) < 0) {
-		log_printf(LOG_ERR, "#48: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#48: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return RG_EFAIL;
 	}
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#49: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#49: Failed getting status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
@@ -1154,12 +1106,20 @@ svc_status(char *svcName)
 		/* Don't check status for anything not owned */
 		return 0;
 
-	if (svcStatus.rs_state != RG_STATE_STARTED &&
-	    svcStatus.rs_state != RG_STATE_MIGRATE)
+	if (svcStatus.rs_state == RG_STATE_STARTED) {
+		/* Running locally and not migrating = normal status
+		 * check
+		 */
+		ret = group_op(svcName, RG_STATUS);
+	} else if (svcStatus.rs_state == RG_STATE_MIGRATE) {
+		/* Migrating resources need an inquiry check to avoid
+		 * setting NEEDSTOP/NEEDSTART in the resource tree.
+		 */
+		ret = group_op(svcName, RG_STATUS_INQUIRY);
+	} else {
 		/* Not-running RGs should not be checked either. */
 		return 0;
-
-	ret = group_op(svcName, RG_STATUS);
+	}
 
 	/* For running services, if the return code is 0, we're done*/
 	if (svcStatus.rs_state == RG_STATE_STARTED)
@@ -1169,8 +1129,8 @@ svc_status(char *svcName)
 }
 
 
-static inline int
-handle_started_status(char *svcName, int ret,
+static int
+handle_started_status(const char *svcName, int ret,
 		      rg_state_t __attribute__((unused)) *svcStatus)
 {
 	int newowner;
@@ -1186,7 +1146,7 @@ handle_started_status(char *svcName, int ret,
 	   inline recovery */
 	if (ret & SFL_RECOVERABLE) {
 
-		log_printf(LOG_WARNING, "Some independent resources in %s failed; "
+		logt_print(LOG_WARNING, "Some independent resources in %s failed; "
 		       "Attempting inline recovery\n", svcName);
 
 		ret = group_op(svcName, RG_CONDSTOP);
@@ -1195,10 +1155,10 @@ handle_started_status(char *svcName, int ret,
 		}
 
 		if (ret) {
-			log_printf(LOG_WARNING, "Inline recovery of %s failed\n",
+			logt_print(LOG_WARNING, "Inline recovery of %s failed\n",
 			       svcName);
 		} else {
-			log_printf(LOG_NOTICE,
+			logt_print(LOG_NOTICE,
 			       "Inline recovery of %s succeeded\n",
 			       svcName);
 			return 0;
@@ -1209,8 +1169,8 @@ handle_started_status(char *svcName, int ret,
 }
 
 
-static inline int
-handle_migrate_status(char *svcName, int ret, rg_state_t *svcStatus)
+static int
+handle_migrate_status(const char *svcName, int ret, rg_state_t *svcStatus)
 {
 	struct dlm_lksb lockp;
 	/* For service(s) migrating to the local node, ignore invalid
@@ -1222,7 +1182,7 @@ handle_migrate_status(char *svcName, int ret, rg_state_t *svcStatus)
 	/* If the check succeeds (returns 0), then flip the state back to
 	   'started' - we now own the service */
 	if (rg_lock(svcName, &lockp) < 0) {
-		log_printf(LOG_ERR, "#45: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#45: Unable to obtain cluster lock: %s\n",
 			strerror(errno));
 		return RG_EFAIL;
 	}
@@ -1230,25 +1190,25 @@ handle_migrate_status(char *svcName, int ret, rg_state_t *svcStatus)
 	svcStatus->rs_state = RG_STATE_STARTED;
 	if (set_rg_state(svcName, svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#46: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#46: Failed getting status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
 	rg_unlock(&lockp);
 
-	log_printf(LOG_NOTICE, "%s is now running locally\n", svcName);
+	logt_print(LOG_NOTICE, "%s is now running locally\n", svcName);
 
 	return 0;
 }
 
 
 int
-svc_status_inquiry(char *svcName)
+svc_status_inquiry(const char *svcName)
 {
 	rg_state_t svcStatus;
 
 	if (get_rg_state_local(svcName, &svcStatus) != 0) {
-		log_printf(LOG_ERR, "Failed getting local status for RG %s\n",
+		logt_print(LOG_ERR, "Failed getting local status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
@@ -1256,7 +1216,7 @@ svc_status_inquiry(char *svcName)
 	if (svcStatus.rs_flags & RG_FLAG_FROZEN)
 		return 0;
 	
-	return group_op(svcName, RG_STATUS);
+	return group_op(svcName, RG_STATUS_INQUIRY);
 }
 
 
@@ -1269,28 +1229,28 @@ svc_status_inquiry(char *svcName)
  * @return		FAIL, 0
  */
 static int
-_svc_stop(char *svcName, int req, int recover, uint32_t newstate)
+_svc_stop(const char *svcName, int req, int recover, uint32_t newstate)
 {
 	struct dlm_lksb lockp;
 	rg_state_t svcStatus;
-	int ret;
+	int ret = 0;
 	int old_state;
 
 	if (!rg_quorate()) {
-		log_printf(LOG_WARNING, "#69: Unclean %s of %s\n", 
+		logt_print(LOG_WARNING, "#69: Unclean %s of %s\n", 
 		       rg_req_str(req), svcName);
 		return group_op(svcName, RG_STOP);
 	}
 
 	if (rg_lock(svcName, &lockp) == RG_EFAIL) {
-		log_printf(LOG_ERR, "#50: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#50: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return RG_EFAIL;
 	}
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#51: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#51: Failed getting status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
@@ -1298,7 +1258,7 @@ _svc_stop(char *svcName, int req, int recover, uint32_t newstate)
 	switch (svc_advise_stop(&svcStatus, svcName, req)) {
 	case 0:
 		rg_unlock(&lockp);
-		log_printf(LOG_DEBUG, "Unable to stop RG %s in %s state\n",
+		logt_print(LOG_DEBUG, "Unable to stop RG %s in %s state\n",
 		       svcName, rg_state_str(svcStatus.rs_state));
 		return RG_EFAIL;
 	case 6:
@@ -1329,7 +1289,19 @@ _svc_stop(char *svcName, int req, int recover, uint32_t newstate)
 
 	old_state = svcStatus.rs_state;
 
-	log_printf(LOG_NOTICE, "Stopping service %s\n", svcName);
+	if (old_state == RG_STATE_RECOVER) {
+		logt_print(LOG_DEBUG, "%s is clean; skipping double-stop\n",
+		       svcName);
+		svcStatus.rs_state = newstate;
+
+		if (set_rg_state(svcName, &svcStatus) != 0) {
+			rg_unlock(&lockp);
+			logt_print(LOG_ERR, "#52: Failed changing RG status\n");
+			return RG_EFAIL;
+		}
+	} 
+
+	logt_print(LOG_NOTICE, "Stopping service %s\n", svcName);
 
 	if (recover) 
 	       	svcStatus.rs_state = RG_STATE_ERROR;
@@ -1341,7 +1313,7 @@ _svc_stop(char *svcName, int req, int recover, uint32_t newstate)
 
 	if (set_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#52: Failed changing RG status\n");
+		logt_print(LOG_ERR, "#52: Failed changing RG status\n");
 		return RG_EFAIL;
 	}
 	rg_unlock(&lockp);
@@ -1350,7 +1322,7 @@ _svc_stop(char *svcName, int req, int recover, uint32_t newstate)
 
 	if (old_state == RG_STATE_FAILED && newstate == RG_STATE_DISABLED) {
 		if (ret)
-			log_printf(LOG_ALERT, "Marking %s as 'disabled', "
+			logt_print(LOG_ALERT, "Marking %s as 'disabled', "
 			       "but some resources may still be allocated!\n",
 			       svcName);
 		_svc_stop_finish(svcName, 0, newstate);
@@ -1363,20 +1335,20 @@ _svc_stop(char *svcName, int req, int recover, uint32_t newstate)
 
 
 static int
-_svc_stop_finish(char *svcName, int failed, uint32_t newstate)
+_svc_stop_finish(const char *svcName, int failed, uint32_t newstate)
 {
 	rg_state_t svcStatus;
 	struct dlm_lksb lockp;
 
 	if (rg_lock(svcName, &lockp) == RG_EFAIL) {
-		log_printf(LOG_ERR, "#53: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#53: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return RG_EFAIL;
 	}
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#54: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#54: Failed getting status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
@@ -1392,7 +1364,7 @@ _svc_stop_finish(char *svcName, int failed, uint32_t newstate)
 	svcStatus.rs_owner = 0;
 
 	if (failed) {
-		log_printf(LOG_CRIT, "#12: RG %s failed to stop; intervention "
+		logt_print(LOG_CRIT, "#12: RG %s failed to stop; intervention "
 		       "required\n", svcName);
 		newstate = RG_STATE_FAILED;
 	} else if (svcStatus.rs_state == RG_STATE_ERROR) {
@@ -1402,14 +1374,14 @@ _svc_stop_finish(char *svcName, int failed, uint32_t newstate)
 
 	svcStatus.rs_state = newstate;
 
-	log_printf(LOG_NOTICE, "Service %s is %s\n", svcName,
+	logt_print(LOG_NOTICE, "Service %s is %s\n", svcName,
 	       rg_state_str(svcStatus.rs_state));
 	//printf("rg state = %s\n", rg_state_str(svcStatus.rs_state));
 
 	svcStatus.rs_transition = (uint64_t)time(NULL);
 	if (set_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#55: Failed changing RG status\n");
+		logt_print(LOG_ERR, "#55: Failed changing RG status\n");
 		return RG_EFAIL;
 	}
 	rg_unlock(&lockp);
@@ -1429,14 +1401,14 @@ _svc_stop_finish(char *svcName, int failed, uint32_t newstate)
  * @return		FAIL, 0
  */
 int
-svc_disable(char *svcName)
+svc_disable(const char *svcName)
 {
 	return _svc_stop(svcName, RG_DISABLE, 0, RG_STATE_DISABLED);
 }
 
 
 int
-svc_stop(char *svcName, int req)
+svc_stop(const char *svcName, int req)
 {
 	return _svc_stop(svcName, req, (req == RG_STOP_RECOVER),
 			 RG_STATE_STOPPED);
@@ -1450,22 +1422,22 @@ svc_stop(char *svcName, int req)
  * @return		FAIL, 0
  */
 int
-svc_fail(char *svcName)
+svc_fail(const char *svcName)
 {
 	struct dlm_lksb lockp;
 	rg_state_t svcStatus;
 
 	if (rg_lock(svcName, &lockp) == RG_EFAIL) {
-		log_printf(LOG_ERR, "#55: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#55: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return RG_EFAIL;
 	}
 
-	log_printf(LOG_DEBUG, "Handling failure request for RG %s\n", svcName);
+	logt_print(LOG_DEBUG, "Handling failure request for RG %s\n", svcName);
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#56: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#56: Failed getting status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
@@ -1473,7 +1445,7 @@ svc_fail(char *svcName)
 	if ((svcStatus.rs_state == RG_STATE_STARTED) &&
 	    (svcStatus.rs_owner != (uint32_t)my_id())) {
 		rg_unlock(&lockp);
-		log_printf(LOG_DEBUG, "Unable to disable RG %s in %s state\n",
+		logt_print(LOG_DEBUG, "Unable to disable RG %s in %s state\n",
 		       svcName, rg_state_str(svcStatus.rs_state));
 		return RG_EFAIL;
 	}
@@ -1490,7 +1462,7 @@ svc_fail(char *svcName)
 	svcStatus.rs_restarts = 0;
 	if (set_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#57: Failed changing RG status\n");
+		logt_print(LOG_ERR, "#57: Failed changing RG status\n");
 		return RG_EFAIL;
 	}
 	rg_unlock(&lockp);
@@ -1507,24 +1479,24 @@ svc_fail(char *svcName)
  * @param svcName	Service ID to flag/unflag as frozen.
  * @return		FAIL, 0
  */
-int
-_svc_freeze(char *svcName, int enabled)
+static int
+_svc_freeze(const char *svcName, int enabled)
 {
 	struct dlm_lksb lockp;
 	rg_state_t svcStatus;
 
 	if (rg_lock(svcName, &lockp) == RG_EFAIL) {
-		log_printf(LOG_ERR, "#55: Unable to obtain cluster lock: %s\n",
+		logt_print(LOG_ERR, "#55: Unable to obtain cluster lock: %s\n",
 		       strerror(errno));
 		return RG_EFAIL;
 	}
 
-	log_printf(LOG_DEBUG, "Handling %s request for RG %s\n", enabled?"freeze":"unfreeze",
+	logt_print(LOG_DEBUG, "Handling %s request for RG %s\n", enabled?"freeze":"unfreeze",
 	       svcName);
 
 	if (get_rg_state(svcName, &svcStatus) != 0) {
 		rg_unlock(&lockp);
-		log_printf(LOG_ERR, "#56: Failed getting status for RG %s\n",
+		logt_print(LOG_ERR, "#56: Failed getting status for RG %s\n",
 		       svcName);
 		return RG_EFAIL;
 	}
@@ -1535,16 +1507,16 @@ _svc_freeze(char *svcName, int enabled)
 	case RG_STATE_DISABLED:
 
 		if (enabled == 1) {
-			log_printf(LOG_DEBUG, "Freezing RG %s\n", svcName);
+			logt_print(LOG_DEBUG, "Freezing RG %s\n", svcName);
 			svcStatus.rs_flags |= RG_FLAG_FROZEN;
 		} else {
-			log_printf(LOG_DEBUG, "Unfreezing RG %s\n", svcName);
+			logt_print(LOG_DEBUG, "Unfreezing RG %s\n", svcName);
 			svcStatus.rs_flags &= ~RG_FLAG_FROZEN;
 		}
 
 		if (set_rg_state(svcName, &svcStatus) != 0) {
 			rg_unlock(&lockp);
-			log_printf(LOG_ERR, "#57: Failed changing RG status\n");
+			logt_print(LOG_ERR, "#57: Failed changing RG status\n");
 			return RG_EFAIL;
 		}
 		break;
@@ -1561,13 +1533,13 @@ _svc_freeze(char *svcName, int enabled)
 }
 
 int
-svc_freeze(char *svcName)
+svc_freeze(const char *svcName)
 {
 	return _svc_freeze(svcName, 1);
 }
 
 int
-svc_unfreeze(char *svcName)
+svc_unfreeze(const char *svcName)
 {
 	return _svc_freeze(svcName, 0);
 }
@@ -1577,7 +1549,7 @@ svc_unfreeze(char *svcName)
  * Send a message to the target node to start the service.
  */
 int
-svc_start_remote(char *svcName, int request, uint32_t target)
+svc_start_remote(const char *svcName, int request, uint32_t target)
 {
 	SmMessageSt msg_relo;
 	int msg_ret;
@@ -1598,7 +1570,7 @@ svc_start_remote(char *svcName, int request, uint32_t target)
 	/* Open a connection to the other node */
 
 	if (msg_open(MSG_CLUSTER, target, RG_PORT, &ctx, 2)< 0) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#58: Failed opening connection to member #%d\n",
 		       target);
 		return -1;
@@ -1610,14 +1582,14 @@ svc_start_remote(char *svcName, int request, uint32_t target)
 	/* Send relocate message to the other node */
 	if (msg_send(&ctx, &msg_relo, sizeof (SmMessageSt)) < 
 	    (int)sizeof (SmMessageSt)) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "#59: Error sending remote-start request to member #%d\n",
 		       target);
 		msg_close(&ctx);
 		return -1;
 	}
 
-	log_printf(LOG_DEBUG, "Sent remote-start request to %d\n", (int)target);
+	logt_print(LOG_DEBUG, "Sent remote-start request to %d\n", (int)target);
 
 	/* Check the response */
 	do {
@@ -1632,7 +1604,7 @@ svc_start_remote(char *svcName, int request, uint32_t target)
 		/* Check to see if resource groups are locked for local
 		   shutdown */
 		if (rg_locked()) {
-			log_printf(LOG_WARNING,
+			logt_print(LOG_WARNING,
 			       "#XX: Cancelling relocation: Shutting down\n");
 			msg_close(&ctx);
 			return RG_NO;
@@ -1644,7 +1616,7 @@ svc_start_remote(char *svcName, int request, uint32_t target)
 			free_member_list(ml);
 			continue;
 		}
-		log_printf(LOG_WARNING,
+		logt_print(LOG_WARNING,
 		       "#XX: Cancelling relocation: Target node down\n");
 		free_member_list(ml);
 		msg_close(&ctx);
@@ -1656,7 +1628,7 @@ svc_start_remote(char *svcName, int request, uint32_t target)
 		 * In this case, we don't restart the service, because the 
 		 * service state is actually unknown to us at this time.
 		 */
-		log_printf(LOG_ERR, "#60: Mangled reply from member #%d during RG "
+		logt_print(LOG_ERR, "#60: Mangled reply from member #%d during RG "
 		       "relocate\n", target);
 		msg_close(&ctx);
 		return 0;	/* XXX really UNKNOWN */
@@ -1843,19 +1815,19 @@ handle_relocate_req(char *svcName, int request, int preferred_target,
 			return RG_EFAIL;
 		default:
 			/* deliberate fallthrough */
-			log_printf(LOG_ERR,
+			logt_print(LOG_ERR,
 			       "#61: Invalid reply from member %d during"
 			       " relocate operation!\n", target);
 		case RG_NO:
 			/* state uncertain */
 			free_member_list(allowed_nodes);
-			log_printf(LOG_CRIT, "State Uncertain: svc:%s "
+			logt_print(LOG_CRIT, "State Uncertain: svc:%s "
 			       "nid:%d req:%s ret:%d\n", svcName,
 			       target, rg_req_str(request), ret);
 			return 0;
 		case 0:
 			*new_owner = target;
-			log_printf(LOG_NOTICE, "Service %s is now running "
+			logt_print(LOG_NOTICE, "Service %s is now running "
 			       "on member %d\n", svcName, (int)target);
 			free_member_list(allowed_nodes);
 			return 0;
@@ -1878,7 +1850,7 @@ handle_relocate_req(char *svcName, int request, int preferred_target,
 	 */
 exhausted:
 	if (!rg_locked()) {
-		log_printf(LOG_WARNING,
+		logt_print(LOG_WARNING,
 		       "#70: Failed to relocate %s; restarting locally\n",
 		       svcName);
 		if (svc_start(svcName, RG_START_RECOVER) == 0) {
@@ -1989,9 +1961,9 @@ handle_start_req(char *svcName, int req, int *new_owner)
 	 * Keep the state open so the other nodes don't try to start
 	 * it.  This allows us to be the 'root' of a given service.
 	 */
-	log_printf(LOG_DEBUG, "Stopping failed service %s\n", svcName);
+	logt_print(LOG_DEBUG, "Stopping failed service %s\n", svcName);
 	if (svc_stop(svcName, RG_STOP_RECOVER) != 0) {
-		log_printf(LOG_CRIT,
+		logt_print(LOG_CRIT,
 		       "#13: Service %s failed to stop cleanly\n",
 		       svcName);
 		(void) svc_fail(svcName);
@@ -2011,7 +1983,7 @@ relocate:
 	 * we should relocate the service.
 	 */
 	if (actual_failure)
-		log_printf(LOG_WARNING, "#71: Relocating failed service %s\n",
+		logt_print(LOG_WARNING, "#71: Relocating failed service %s\n",
 	       	       svcName);
 	ret = handle_relocate_req(svcName, RG_START_RECOVER, -1, new_owner);
 
@@ -2085,7 +2057,7 @@ handle_start_remote_req(char *svcName, int req)
 
 	x = svc_start(svcName, req);
 
-	if ((x == 0) || (x == RG_ERUN)) {
+	if ((x == 0) || (x == RG_ERUN) || (x == RG_EFROZEN)) {
 		if (need_check)
 			pthread_mutex_unlock(&exclusive_mutex);
 		return x;
@@ -2120,7 +2092,7 @@ handle_recover_req(char *svcName, int *new_owner)
 
 	/* Check restart counter/timer for this resource */
 	if (check_restart(svcName) > 0) {
-		log_printf(LOG_NOTICE, "Restart threshold for %s exceeded; "
+		logt_print(LOG_NOTICE, "Restart threshold for %s exceeded; "
 		       "attempting to relocate\n", svcName);
 		return handle_relocate_req(svcName, RG_START_RECOVER, -1,
 					   new_owner);
@@ -2152,13 +2124,16 @@ handle_fd_start_req(char *svcName, int request, int *new_owner)
 		target = best_target_node(allowed_nodes, 0,
 	 				  svcName, 1);
 	  	if (target == me) {
-	   		ret = handle_start_remote_req(svcName, request);
+ 		      	ret = handle_start_remote_req(svcName,
+				(request==RG_ENABLE?RG_START_RECOVER:request));
 			if (ret == RG_EAGAIN)
 				goto out;
-	    	} else if (target < 0) {
+		} else if (!(target > 0)) {
 			goto out;
 	       	} else {
-			ret = svc_start_remote(svcName, RG_START_REMOTE,
+			ret = svc_start_remote(svcName,
+					       (request==RG_ENABLE?
+					    RG_START_RECOVER:RG_START_REMOTE),
 					       target);
 		}
 
@@ -2177,7 +2152,7 @@ handle_fd_start_req(char *svcName, int request, int *new_owner)
 			ret = RG_EFAIL;
 			goto out;
       		default:
-			log_printf(LOG_ERR,
+			logt_print(LOG_ERR,
 	 		       "#6X: Invalid reply [%d] from member %d during"
 	  		       " relocate operation!\n", ret, target);
 	   	}
